// Chess.java (requires also the files ChessDataClient2.java, ChessMemoryClient2.java,
// and RememberedPosition.java)
//
// Compile with the -o (optimizing) switch:
// javac -O Chess.java ChessDataClient2.java ChessMemoryClient2.java RememberedPosition.java
//
// Version 6.06, last updated May 17 26, 2016, by Rick Wagner. See version history at end
// of file. This file was originally created February 2, 2001, by Rick Wagner.
// Versions 6.xx to undo SQL support by Hannes Mayer to upgrade the system for MySQL chess data service.
// See version history at the bottom of index.html (which also sets the applet parameters).
//
// Applet for a computer player chess game.
// Copyright 2001-2016 by Rick Wagner, all rights reserved.
//
// If performance were the driving concern I wouldn't have written it in Java. This
// is a fully object oriented (OO) chess applet. Written for the JDK 1.1 it will
// run in any browser on any Web-aware computer. Strict OO discipline was applied throughout,
// regardless of performance considerations (except where noted).
//
// Restrictions:
//
// Use of this source code is authorized for educational purposes only. No use without
// proper attribution to Rick Wagner (http://rjwagner49.com/
// e-mail: Richard.J.Wagner@gmail.com). Distribution of this code is prohibited without
// expressed permission. Source and executable of derived products must carry this notice.
//
// Naming Conventions:
//
// Classes use the initial capital convention, e.g., "ChessPiece" and "Square."
// Functions use the initial lowercase convention, e.g., "setupBoard()" and "constructPieces()."
// Accessors use the standard "get" naming convention and mutators use the standard "set"
// naming convention, e.g., "setType()" and getType()."
//
// The object prefix naming convention used here is a modified Hungarian notation.
// "s" is string, "sf" is single precision floating point, "df" is double precision
// floating point, "i" is integer, "b" is boolean, "d" is dimension, and "c" is color.
// Other prefixes are derived from the capital letters in the class names.
//
// The predicate function naming convention of a "P" suffix, standard in the Lisp language,
// is used here: any function returning a boolean value has a suffix "P" for "predicate" (disclaimer:
// not used consistently in this program).

import java.applet.*;                                       // For the class Applet.
import java.awt.*;                                          // Abstract windowing toolkit; all the GUI stuff.
import java.util.*;                                         // For Vector and Date.

public final class Chess extends Applet implements Runnable
{
  // Applet constants:
  private final String sVerNum = "6.06";                    // Current version.
  private final String sCompiledDate = "May 18, 2016";      // Compiled date.
  private final String sCompilerVersion = "1.7.0_09";       // JDK "javac" version.
  private final String sNumCodeLines = "7197";              // Length of this file in lines.

  // Applet instance variables and objects (private where possible):
  private GridBagLayout gblLayout = null;                   // GridBagLayout is used for the applet GUI layout.
  private boolean bLongSounds = true;                       // Flag to enable long sounds.
  private AudioClip acStartup = null;                       // Startup sound.
  private AudioClip acHALChess = null;                      // Alternate sound (HAL 9000 from 2001 Space Odyssey).
  private AudioClip acReset = null;                         // Reset sound.
  private AudioClip acNotify = null;                        // Ding sound after computer move.
  private AudioClip acCheck = null;                         // Sound for check.
  private AudioClip acVictory = null;                       // Laughter sound on computer win.
  private AudioClip acDefeat = null;                        // Scream sound on computer loss.
  private AudioClip acMoveIllegal = null;                   // Sound on illegal user move.
  private AudioClip acMoveIntoCheck = null;                 // Vomit sound if user tries to move into check.
  private AudioClip acResign = null;                        // Sigh sound if comuter resigns.
  private AudioClip acPromote = null;                       // Fanfare sound on pawn promotion.
  private AudioClip acDraw = null;                          // Yawn sound on a drawn game.

  protected Button btnReset = null;                         // Reset button. Accessed by ChessDataClient2 (cdc).
  private Button btnListing = null;                         // Game listing button.
  private Button btnSound = null;                           // Long sounds button.
  protected Button btnSave = null;                          // Save game button. Accessed by cdc.
  protected Button btnOpen = null;                          // Open game button. Accessed by cdc.
  private Panel pButtonRow;                                 // Panel for the row of buttons.
  protected Label lblMessage = null;                        // Talk to the user here. Accessed by cdc.

  private Image imOffScreen = null;                         // Offscreen image for double buffering.
  private Graphics grOffScreen = null;                      // Offscreen graphics for double buffering.

  private Dimension dApplet = null;                         // The applet panel size (set in calling html).
  private int iSquareWidth = 0;                             // Region boundaries for dividing the game grid.
  private int iSquareHeight = 0;                            //

  private Thread tCompute = null;                           // Do the thinking in a thread.

  protected Square sHumanFrom = null;                       // Square the human user is moving from.  Accessed by cdc.
  protected GameNode gnGameRoot = null;                     // Root game node for computing the game tree, accessed by cdc.
  protected int iRecursionLimit = 0;                        // Initial look-ahead, modified later, accessed by cdc.
  protected int iParamRecurLimit = 0;                       // Initial look-ahead, read from applet parameter.

  protected boolean bThinking = false;                      // Don't take user inputs while thinking.
  protected boolean bGameOver = false;                      // Flag for the end of the game.

  private boolean bNeedPromoting = false;                   // Data for pawn promotion.
  private Square sPromoteSquare = null;                     //
  private ChessPiece cpPromote = null;                      // Chesspiece for promotion.

  private Polygon pgPawn = null;                            // Polygons for drawing the chess pieces:
  private Polygon pgKKnight = null;                         // We don't want to burden the ChessPiece
  private Polygon pgQKnight = null;                         // class with these data, so we keep them
  private Polygon pgBishop = null;                          // at Applet level.
  private Polygon pgRook = null;                            //
  private Polygon pgQueen = null;                           // This is a concession to performance considerations.
  private Polygon pgKing = null;                            //

  private long lPositionsCounter = 0;                       // Count the number of positions considered.
  private int iNumImmediateChoices = 0;                     // Number of immediate moves to consider.

  private boolean bHighlightFlag[][] = new boolean[8][8];   // For square highlighting on mouseMove().

  protected TextArea taGameListing = null;                  // The game listing text area.  Accessed by cdc.
  protected int iNumMoves = 0;                              // The move number in the game listing, accessed by cdc.
  protected java.awt.List lGamesList = null;                // List for games to open.  Accessed by cdc.

  private FontMetrics fmApplet = null;                      // For finding the width of text in paint().

  private int iMaxTime = 200;                               // Maximum thinking time, if exceeded decrease depth.
  private int iMinTime = 30;                                // Minimum thinking time, if less increase depth.
  private int iPreviousUsedTime1 = 0;                       // For time use averaging, version 3.10.
  private int iPreviousUsedTime2 = 0;                       //
  private int iPreviousUsedTime3 = 0;                       //
  private boolean bDepthChanged = false;                    // Flag to stabilize recursion depth.

  private int iMoveCounter = 0;                             // How many of the possible computer moves have been considered.
  private String strBest = null;                            // Best move found so far.
  private String strBestMetric = null;                      // Best move's metric.
  protected Vector vHistory = new Vector();                 // Vector of ChessBoards for draw testing (accessed by cdc).
  protected Vector vDuplicate = new Vector();               // Vector of duplicate chessboards for draw testing (accessed by cdc).

  protected int iPortNumber = 1997;                         // Default port.
  private ChessDataClient2 cdc = null;                      // For server communication for save and open.
  private String sHeader = null;                            // Client data.
  protected String sSessionID = null;                       // Session identifier.
  protected int iOpenGameNumber = 0;                        // Game number identifier.
  private boolean bDirty = false;                           // See if the user actually tries to play.
  private boolean bAutoPlay = false;                        // Flag for auto-play.
  protected boolean bVerbose = false;                       // Flag for verbose mode.
  private boolean bFixedDepth = false;                      // Flag for using a fixed search depth (non-dynamic).
  private boolean bRotated = false;                         // Flag for board orientation.
  private boolean bNewby = true;                            // Flag for move highlighting for newby assistance.
  private float sfDeepMetric = 0;                           // Metric of the best future position after min-max computation.

  // For server memory feature:
  protected ChessMemoryClient2 cmcPutter = null;            // For putting moves to the server, accessed from tMovePutter.
  private ChessMemoryClient2 cmc = null;                    // For other chess memory functions.
  private boolean bMemorySend = true;                       // Flag for sending memory to the server.
  private boolean bMemoryUse = true;                        // Flag for the use of memory.
  private boolean bMemoryUseDisabled = false;               //
  private int iRecurrences = 0;                             // Timer counter.
  private Timer tMoveGetter = null;                         // Timer (extends Thread) for callback waiting.
  protected String sMoveFromServer = null;                  // cmc communication string.
  private boolean bWaitingForServer = false;                // Flag for callback waiting.
  private boolean bHumanHuman = false;                      // Flag for human-human play.
  private boolean bMoveSendInhibited = false;               // Inhibit sending moves if contact with the server is broken.

  // For timed play feature:
  private boolean bTimedGame = false;                       // Flag for timed game.
  private long lMaxCumTime = 90 * 60;                       // 90 minute (per side) games.
  private int iNumMovesTimeControl = 40;                    // Number of moves until the time control.
  private long lMaxExtendedTime = 60 * 60;                  // 60 minutes (per side) added after 40 moves.
  private long lBlackCumTime = 0;                           // Black game time accumulator.
  private long lWhiteCumTime = 0;                           // White game time accumulator.
  private long lBlackStartTime = 0;                         // Black start time.
  private long lWhiteStartTime = 0;                         // White start time.
  private boolean bTimeLossNotified = false;                // Flag for having notified the loser on time.

  // For piece color and contrast feature:
  private Color cWhitePiece = null;                         // Color of the white pieces.
  private Color cBlackPiece = null;                         // Color of the black pieces.
  private int iPieceColorScheme = 0;                        // The user-selected color scheme, default is coffee-cream.

  // Applet member functions:
  // To allow browsers to get information about the applet (not yet implemented in Netscape nor in MSIE):
  public String getAppletInfo()
  {
    return "Chess applet, version " + sVerNum +
           ", by Rick Wagner, copyright 2001-2016,\nall rights reserved. " +
           "Compiled " + sCompiledDate + ", using JDK version " + sCompilerVersion + ".\n" +
           "There are " + sNumCodeLines + " lines of code in the primary Java source file.\n" +
           "See http://rjwagner49.com/ for author information.\n";

  } // End of getAppletInfo().

  // Initialize the applet
  public void init()
  {
    int i = 0;
    Dimension dMessage = null;
    GridBagConstraints gbc = null;
    String sSend = null;
    String sTemp = null;

    // Set some applet instance variables:
    bThinking = false;                                       // Make sure no thinking continues (version 3.88).
    bNewby = true;                                           // Assume the user is a novice (version 4.02).
    iPieceColorScheme = 0;                                   // Default coffee and cream piece colors.

    if (bVerbose)
    {
      System.out.println("Initializing chess applet:\nCodebase is " + getCodeBase() +
                       "\nDocumentbase is " + getDocumentBase());
    }

    // Piece default colors:
    cWhitePiece = new Color(255, 255, 223);
    cBlackPiece = new Color(48, 48, 0);

    // Applet parameters are set in the calling HTML code.:
    sTemp = getParameter("MinTime");
    if (sTemp != null)
    {
      iMinTime = Integer.parseInt(sTemp);
    }
    sTemp = getParameter("MaxTime");
    if (sTemp != null)
    {
      iMaxTime = Integer.parseInt(sTemp);
    }
    sTemp = getParameter("FixedDepth");
    if (sTemp != null)
    {
      i = Integer.parseInt(sTemp);
      if (i == 1)
      {
        bFixedDepth = true;
        bMemoryUse = false;
      }
    }
    sTemp = getParameter("RecursionLimit");
    if (sTemp != null)
    {
      iParamRecurLimit = Integer.parseInt(sTemp);
      iRecursionLimit = iParamRecurLimit;
    }
    else
    {
      if (iMinTime > 60)
      {
        iRecursionLimit = 4;                                   // Initial look-ahead half moves.
        iParamRecurLimit = 4;
      }
      else
      {
        if (iMinTime < 5)                                      // For speedchess mode.
        {
          iRecursionLimit = 2;                                 // Initial look-ahead half moves.
          iParamRecurLimit = 2;
        }
        else
        {
          iRecursionLimit = 3;
          iParamRecurLimit = 3;
        }
      }
    }
    sTemp = getParameter("PortNumber");                        // For changing from the default port number 1997.
    if (sTemp != null)
    {
      iPortNumber = Integer.parseInt(sTemp);
    }

    setBackground(Color.lightGray);                            // The gray-on-gray color scheme.
    dApplet = size();                                          // Size is set in the calling HTML code.
    fmApplet = getFontMetrics(getFont());

    // Get the audio clips (students and researchers, download the sound files from the Chess Applet Website if necessary):
    acStartup = getAudioClip(getDocumentBase(), "play_a_game.au");
    acReset = getAudioClip(getDocumentBase(), "chess.au");
    acHALChess = getAudioClip(getDocumentBase(), "HALchess.au");
    acNotify = getAudioClip(getDocumentBase(), "dingIN.au");
    acCheck = getAudioClip(getDocumentBase(), "check.au");
    acMoveIllegal = getAudioClip(getDocumentBase(), "illegal.au");
    acMoveIntoCheck = getAudioClip(getDocumentBase(), "heaveIN.au");
    acVictory = getAudioClip(getDocumentBase(), "victoryIN.au");
    acDefeat = getAudioClip(getDocumentBase(), "defeatIN.au");
    acResign = getAudioClip(getDocumentBase(), "resign.au");
    acPromote = getAudioClip(getDocumentBase(), "fanfare.au");
    acDraw = getAudioClip(getDocumentBase(), "yawn.au");

    // Set the region boundaries:
    iSquareWidth = dApplet.width / 10;                         // Buffer around 8 x 8 board.
    iSquareHeight = dApplet.height / 10;

    // Set up the GUI layout:
    gblLayout = new GridBagLayout();
    
    String osname = System.getProperty("os.name","");
    if (!osname.startsWith("Windows")) {
	setFont(new Font("Arial", Font.PLAIN, 9));
    }
    setLayout(gblLayout);
    gbc = new GridBagConstraints();
    gbc.weightx = 0.0;
    gbc.weighty = 0.0;

    // Add the buttons panel;
    pButtonRow = new Panel();
    pButtonRow.setLayout(new GridLayout(1, 5));

    // Add the reset button:
    btnReset = new Button("Reset Game");                        // The reset button.
    btnReset.setForeground(Color.black);
    btnReset.setBackground(Color.lightGray);
    btnReset.disable();

    btnListing = new Button("Show Listing");                    // Game listing button.
    btnListing.setForeground(Color.black);
    btnListing.setBackground(Color.lightGray);
    btnListing.enable();

    btnSound = new Button("Disable Sound");                     // Long sounds button.
    btnSound.setForeground(Color.black);
    btnSound.setBackground(Color.lightGray);
    btnSound.enable();

    btnSave = new Button("Save Game");                          // Save game button.
    btnSave.setForeground(Color.black);
    btnSave.setBackground(Color.lightGray);
    btnSave.disable();

    btnOpen = new Button("Open Game");                          // Open game button.
    btnOpen.setForeground(Color.black);
    btnOpen.setBackground(Color.lightGray);
    btnOpen.disable();

    pButtonRow.add(btnReset);
    pButtonRow.add(btnListing);
    pButtonRow.add(btnSound);
    pButtonRow.add(btnSave);
    pButtonRow.add(btnOpen);

    gbc.insets = new Insets(15, 0, 220, 0);                     // top, left, bottom, right.
    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.fill = GridBagConstraints.NONE;
    gblLayout.setConstraints(pButtonRow, gbc);
    add(pButtonRow);

    // Add the games list:
    lGamesList = new java.awt.List(24, false);                  // 24 rows for listing game files.
    gbc.insets = new Insets(0, 0, 0, 0);                        // top, left, bottom, right.
    gbc.gridx = 0;
    gbc.gridy = 1;
    gblLayout.setConstraints(lGamesList, gbc);
    add(lGamesList);

    // Add the game listing:
    taGameListing = new TextArea("\tWhite\tBlack", 24, 28);     // 24 rows and 28 columns.
    taGameListing.setEditable(true);
    gblLayout.setConstraints(taGameListing, gbc);
    add(taGameListing);


    // Add the message label:
    lblMessage = new Label("Click the white piece you want to move, " +
                           "then click the move-to square.", Label.CENTER);
    dMessage = new Dimension(dApplet.width - 2, 23);
    lblMessage.setMinimumSize(dMessage);
    gbc.insets = new Insets(220, 0, 5, 0);                      // top, left, bottom, right.
    gbc.gridx = 0;
    gbc.gridy = 2;
    gblLayout.setConstraints(lblMessage, gbc);
    add(lblMessage);
    
    constructPieces();                                          // Construct the piece shapes for painting.
    bVerbose = true;
    setupBoard();                                               // Set up the chess board.
    bVerbose = false;

    // Create the chess data and memory clients:
    cdc = new ChessDataClient2(this);                            // Session ID is requested in start().
    cmc = new ChessMemoryClient2(this);                          // For getting moves.
    cmcPutter = new ChessMemoryClient2(this);                    // For putting moves.

    if (bLongSounds && !bFixedDepth)
    {
      if (Math.random() < 0.5)
      {
        acStartup.play();
      }
      else
      {
        acHALChess.play();
      }
    }
  } // End of init().

  private void setPerfBoost()                                   // Boost the look-ahead based on system performance.
  {
    int i = 0;
    long lPerfTestStart = 0;
    double dfPerfTestDuration = 0;
    double dfPerfBoost = 0;
    ChessBoard cbPerfTest[] = new ChessBoard[1200];             // Set to 1200 from 1000 in version 4.10.

    lPerfTestStart = System.currentTimeMillis();
    for (i = 0; i < 1200; i++)
    {
      cbPerfTest[i] = new ChessBoard();                         // How fast can the system create objects
      cbPerfTest[i].setUpBoard();                               // and call functions?
    }
    for (i = 0; i < 1200; i++)
    {
      cbPerfTest[i] = null;
    }
    cbPerfTest = null;
    dfPerfTestDuration = (double) (System.currentTimeMillis() - lPerfTestStart);
    dfPerfBoost = Math.log(1600.0 / dfPerfTestDuration) / 1.791;  // 1600 is the duration on my baseline machine.
    iRecursionLimit += (int) (dfPerfBoost + 0.5);                 // 1.791 is the natural log of 6 (sqr(BranchFactor)).
    if (bVerbose)
    {
      System.out.println("Test duration = " + ((int) dfPerfTestDuration) +
                         " ms, boost factor = " + ((float) dfPerfBoost) +
                         ", recursion limit = " + iRecursionLimit + ".");
    }
  } // End of setPerfBoost().

  private void constructPieces()
  {
    constructPawn();
    constructKKnight();
    constructQKnight();
    constructBishop();
    constructRook();
    constructQueen();
    constructKing();

  } // End of constructPieces().

  private void constructPawn()
  {
    int i;
    int iNumPoints = 9;
    int x[] = new int[iNumPoints];
    int y[] = new int[iNumPoints];

    // Here we define the shape of the pawn polygon:
    x[0] = 2; y[0] = 8;
    x[1] = 8; y[1] = 8;
    x[2] = 6; y[2] = 7;
    x[3] = 5; y[3] = 4;
    x[4] = 6; y[4] = 4;
    x[5] = 5; y[5] = 3;
    x[6] = 4; y[6] = 4;
    x[7] = 5; y[7] = 4;
    x[8] = 4; y[8] = 7;

    // Now we scale the polygon to the size we want:
    for (i = 0; i < iNumPoints; i++)
    {
      x[i] = iSquareWidth * x[i] / 10;
      y[i] = iSquareHeight * y[i] / 10;
    }

    pgPawn = new Polygon(x, y, iNumPoints);                   // Polygon object constructor

  } // End of constructPawn().

  private void constructKKnight()
  {
    int i;
    int iNumPoints = 14;
    int x[] = new int[iNumPoints];
    int y[] = new int[iNumPoints];

    // Here we define the shape of the polygon:
    x[0] = 2; y[0] = 9;
    x[1] = 8; y[1] = 9;
    x[2] = 8; y[2] = 8;
    x[3] = 7; y[3] = 7;
    x[4] = 7; y[4] = 3;
    x[5] = 6; y[5] = 2;
    x[6] = 5; y[6] = 2;
    x[7] = 4; y[7] = 1;
    x[8] = 4; y[8] = 2;
    x[9] = 1; y[9] = 4;
    x[10] = 2; y[10] = 5;
    x[11] = 4; y[11] = 4;
    x[12] = 3; y[12] = 7;
    x[13] = 2; y[13] = 8;

    // Now we scale the polygon to the size we want:
    for (i = 0; i < iNumPoints; i++)
    {
      x[i] = iSquareWidth * x[i] / 10;
      y[i] = iSquareHeight * y[i] / 10;
    }
    pgKKnight = new Polygon(x, y, iNumPoints);                   // Polygon object constructor

  } // End of constructKKnight().

  private void constructQKnight()
  {
    int i;
    int iNumPoints = 14;
    int x[] = new int[iNumPoints];
    int y[] = new int[iNumPoints];

    // Here we define the shape of the polygon:
    x[0] = 2; y[0] = 9;
    x[1] = 8; y[1] = 9;
    x[2] = 8; y[2] = 8;
    x[3] = 7; y[3] = 7;
    x[4] = 6; y[4] = 4;
    x[5] = 8; y[5] = 5;
    x[6] = 9; y[6] = 4;
    x[7] = 6; y[7] = 2;
    x[8] = 6; y[8] = 1;
    x[9] = 5; y[9] = 2;
    x[10] = 4; y[10] = 2;
    x[11] = 3; y[11] = 3;
    x[12] = 3; y[12] = 7;
    x[13] = 2; y[13] = 8;

    // Now we scale the polygon to the size we want:
    for (i = 0; i < iNumPoints; i++)
    {
      x[i] = iSquareWidth * x[i] / 10;
      y[i] = iSquareHeight * y[i] / 10;
    }
    pgQKnight = new Polygon(x, y, iNumPoints);                   // Polygon object constructor

  } // End of constructQKnight().

  private void constructBishop()
  {
    int i;
    int iNumPoints = 11;
    int x[] = new int[iNumPoints];
    int y[] = new int[iNumPoints];

    // Here we define the shape of the polygon:
    x[0] = 2; y[0] = 9;
    x[1] = 8; y[1] = 9;
    x[2] = 8; y[2] = 8;
    x[3] = 7; y[3] = 7;
    x[4] = 6; y[4] = 3;
    x[5] = 7; y[5] = 3;
    x[6] = 5; y[6] = 2;
    x[7] = 3; y[7] = 3;
    x[8] = 4; y[8] = 3;
    x[9] = 3; y[9] = 7;
    x[10] = 2; y[10] = 8;

    // Now we scale the polygon to the size we want:
    for (i = 0; i < iNumPoints; i++)
    {
      x[i] = iSquareWidth * x[i] / 10;
      y[i] = iSquareHeight * y[i] / 10;
    }
    pgBishop = new Polygon(x, y, iNumPoints);                   // Polygon object constructor

  } // End of constructBishop().

  private void constructRook()
  {
    int i;
    int iNumPoints = 20;
    int x[] = new int[iNumPoints];
    int y[] = new int[iNumPoints];

    // Here we define the shape of the polygon:
    x[0] = 2; y[0] = 9;
    x[1] = 8; y[1] = 9;
    x[2] = 8; y[2] = 8;
    x[3] = 7; y[3] = 7;
    x[4] = 7; y[4] = 4;
    x[5] = 8; y[5] = 3;
    x[6] = 8; y[6] = 2;
    x[7] = 7; y[7] = 2;
    x[8] = 7; y[8] = 3;
    x[9] = 6; y[9] = 3;
    x[10] = 6; y[10] = 2;
    x[11] = 4; y[11] = 2;
    x[12] = 4; y[12] = 3;
    x[13] = 3; y[13] = 3;
    x[14] = 3; y[14] = 2;
    x[15] = 2; y[15] = 2;
    x[16] = 2; y[16] = 3;
    x[17] = 3; y[17] = 4;
    x[18] = 3; y[18] = 7;
    x[19] = 2; y[19] = 8;

    // Now we scale the polygon to the size we want:
    for (i = 0; i < iNumPoints; i++)
    {
      x[i] = iSquareWidth * x[i] / 10;
      y[i] = iSquareHeight * y[i] / 10;
    }
    pgRook = new Polygon(x, y, iNumPoints);                   // Polygon object constructor

  } // End of constructRook().

  private void constructQueen()
  {
    int i;
    int iNumPoints = 15;
    int x[] = new int[iNumPoints];
    int y[] = new int[iNumPoints];

    // Here we define the shape of the polygon:
    x[0] = 2; y[0] = 9;
    x[1] = 8; y[1] = 9;
    x[2] = 8; y[2] = 8;
    x[3] = 7; y[3] = 7;
    x[4] = 6; y[4] = 4;
    x[5] = 8; y[5] = 2;
    x[6] = 6; y[6] = 3;
    x[7] = 6; y[7] = 1;
    x[8] = 5; y[8] = 3;
    x[9] = 4; y[9] = 1;
    x[10] = 4; y[10] = 3;
    x[11] = 2; y[11] = 2;
    x[12] = 4; y[12] = 4;
    x[13] = 3; y[13] = 7;
    x[14] = 2; y[14] = 8;

    // Now we scale the polygon to the size we want:
    for (i = 0; i < iNumPoints; i++)
    {
      x[i] = iSquareWidth * x[i] / 10;
      y[i] = iSquareHeight * y[i] / 10;
    }
    pgQueen = new Polygon(x, y, iNumPoints);                   // Polygon object constructor

  } // End of constructQueen();

  private void constructKing()
  {
    int i;
    int iNumPoints = 20;
    int x[] = new int[iNumPoints];
    int y[] = new int[iNumPoints];

    // Here we define the shape of the polygon:
    x[0] = 2; y[0] = 9;
    x[1] = 8; y[1] = 9;
    x[2] = 8; y[2] = 8;
    x[3] = 7; y[3] = 7;
    x[4] = 7; y[4] = 5;
    x[5] = 6; y[5] = 4;
    x[6] = 6; y[6] = 3;
    x[7] = 7; y[7] = 3;
    x[8] = 7; y[8] = 2;
    x[9] = 6; y[9] = 2;
    x[10] = 6; y[10] = 1;
    x[11] = 4; y[11] = 1;
    x[12] = 4; y[12] = 2;
    x[13] = 3; y[13] = 2;
    x[14] = 3; y[14] = 3;
    x[15] = 4; y[15] = 3;
    x[16] = 4; y[16] = 4;
    x[17] = 3; y[17] = 5;
    x[18] = 3; y[18] = 7;
    x[19] = 2; y[19] = 8;

    // Now we scale the polygon to the size we want:
    for (i = 0; i < iNumPoints; i++)
    {
      x[i] = iSquareWidth * x[i] / 10;
      y[i] = iSquareHeight * y[i] / 10;
    }
    pgKing = new Polygon(x, y, iNumPoints);                   // Polygon object constructor

  } // End of constructKing().

  private void setupBoard()
  {
    lWhiteCumTime = 0;
    lBlackCumTime = 0;
    lWhiteStartTime =  System.currentTimeMillis();
    lBlackStartTime =  lWhiteStartTime;
    iPreviousUsedTime1 = iMaxTime;
    iPreviousUsedTime2 = iMaxTime;
    iPreviousUsedTime3 = iMaxTime;
    bTimeLossNotified = false;

    gnGameRoot = new GameNode();                             // New game root node.
    gnGameRoot.setBoard(new ChessBoard());                   // New chess board.

    gnGameRoot.getBoard().setUpBoard();                      // Virgin chess board.
    //gnGameRoot.getBoard().TestSetUpBoard();                  // Test chess board (comment out for regular use).
    //gnGameRoot.getBoard().TestSetUpBoard2();                  // Test chess board (comment out for regular use).
    //gnGameRoot.getBoard().TestSetUpBoard3();                  // Test chess board (comment out for regular use).

    gnGameRoot.getBoard().setMove(true);                     // White to move.

    vHistory = new Vector();
    vDuplicate = new Vector();

    if (!bFixedDepth) setPerfBoost();                        // Don't want to boost depth if fixed depth.

  } // End of setupBoard().

  // Execute this code after initialization (and runs each time the browser tells it to):
  public void start()
  {
    String sSend = null;
    Runtime rt = null;
    long lFreeMemory = 0;
    long lTotalMemory = 0;
    Date date = new Date();

    if (bVerbose) System.out.println("start()");
    taGameListing.hide();                                    // Make the text area invisible initially.
    lGamesList.hide();

    rt = Runtime.getRuntime();
    lFreeMemory = rt.freeMemory();
    lTotalMemory = rt.totalMemory();
    System.out.println(lFreeMemory + " bytes of free memory, " + lTotalMemory + " bytes of total memory.\n");
    System.out.println(getAppletInfo());                     // Identify self to the Java console-aware user.
    rt = null;                                               // Garbage collect the runtime object.
    requestFocus();

    // ChessDataClient2 communication with ChessDataServer:
    sSessionID = null;
    sHeader = date.toString() + ", " +
              System.getProperty("os.arch") + ", " +
              System.getProperty("os.name") + ", " +
              System.getProperty("os.version") + ", Recursion limit = " +
              iRecursionLimit;
    System.out.println("\n" + sHeader);
    sSend = "Request SessionID, " + sHeader;
    cdc.listen();
    cdc.send(sSend);
  } // End of start().

  public void stop()
  {
    if (bVerbose) System.out.println("stop()");
    if (bThinking)
    {
      System.out.println("Chess applet thinking continues until the\nmove is made or the applet is destroyed.");
    }
    if (cdc != null)
    {
      cdc.close();
    }
    if (cmc != null)
    {
      cmc.close();
    }
  }

  public void destroy()
  {
    if (bVerbose) System.out.println("destroy()");
    bMemorySend = false;
    bThinking = false;                                       // Halts thinking.
    if (bDirty)
    {
      System.out.println("Goodbye. Thanks for playing chess.");
    }
    else
    {
      System.out.println("Goodbye. Maybe you'll try the game next time.");
    }
    if (cdc != null)
    {
      cdc.close();
    }
    if (cmc != null)
    {
      cmc.close();
    }
  } // End of destroy().

  public void run()                                             // Thread computation.
  {
    ChessBoard cbCurrent = null;

    if (bAutoPlay && !bGameOver)
    {
      while(bAutoPlay && !bGameOver)
      {
        if (!bWaitingForServer)
        {
          if (bGameOver)
          {
            break;                                               // Break out of while.
          }
          else
          {
            cbCurrent = gnGameRoot.getBoard();
            gnGameRoot = new GameNode();                         // Eliminate any old children.
            gnGameRoot.setMetric(cbCurrent.evaluate());          // Evaluate the board.
            gnGameRoot.setBoard(cbCurrent);                      // Set the board for the game root.
            if (cbCurrent.moveP())
            {
              if (bVerbose) System.out.println("White to move.");
            }
            else
            {
              if (bVerbose) System.out.println("Black to move.");
            }
            computerMove();
          }
        }
      }
    }
    else
    {
      if (!bGameOver)
      {
        computerMove();
      }
    }

    if (bAutoPlay)
    {
      bAutoPlay = false;
      if (bVerbose) System.out.println("Auto-play set to off.");
    }
  } // End of run().

  private void computerMove()
  {
    long lStartTime = System.currentTimeMillis();                // Start time in miliseconds since 1970.
    String sTemp = null;
    ChessBoard cbCurrent = null;

    cbCurrent = gnGameRoot.getBoard();
    bDirty = true;                                           // The user has used the game.
    bThinking = true;                                        // No user input while thinking.
    btnReset.disable();
    btnSave.disable();
    btnOpen.disable();
    lPositionsCounter = 0;
    iMoveCounter = 0;
    requestFocus();                                          // Be receptive to keystrokes.
    bMoveSendInhibited = false;
    Date dNow = null;

    // See if the memory server has it:
    if (sSessionID != null && bMemoryUse)
    {
      sMoveFromServer = null;
      sTemp = cbCurrent.toString();
      sTemp = "query " + sTemp;
      cmc.close();
      cmc.reconnect();
      cmc.listen();
      cmc.send(sTemp);                                       // Send the query to the server.
      if (bVerbose)
      {
        dNow = new Date();
        System.out.println("Move query sent to server " + dNow.toString() + ".");
      }
      // Launch the callback waiting thread:
      lblMessage.setText("Connecting to server to look up the position.");
      tMoveGetter = new Timer(this, 100, lStartTime);
      iRecurrences = 0;
      bWaitingForServer = true;
      tMoveGetter.start();
    }
    else
    {
      localMove(cbCurrent, lStartTime);
    }

    if (bMemoryUseDisabled)
    {
      bMemoryUse = true;
      bMemoryUseDisabled = false;
    }

  } // End of computerMove()

  private boolean playerResigns(boolean bMove)
  {
    boolean bResigns = false;
    String strMessage = null;

    if (bMove)
    {
      if (sfDeepMetric < -100)
      {
        strMessage = "PM = " + String.valueOf(sfDeepMetric) + ", white resigns.";
        taGameListing.appendText(" White resigns");
        bResigns = true;
      }
    }
    else
    {
      if (sfDeepMetric > 100)
      {
        strMessage = "PM = " + String.valueOf(sfDeepMetric) + ", black resigns.";
        taGameListing.appendText(" Black resigns");
        bResigns = true;
      }
    }
    if (bResigns)
    {
      acResign.play();
      lblMessage.setText(strMessage);
      System.out.println(strMessage + '\n');
      bGameOver = true;
      bThinking = false;
      sfDeepMetric = 0;
      btnReset.enable();
      if (sSessionID != null)
      {
        btnSave.enable();
        btnOpen.enable();
      }
    }
    return bResigns;
  } // End of playerResigns().

  private void localMove(ChessBoard cbCurrent, long lStartTime)
  {
    String sTemp = null;
    RememberedPosition rpTemp = null;                        // For learning.
    ChessBoard cbBest = null;

    sTemp = cbCurrent.toString();

    // Start the recursive search:
    strBest = null;
    strBestMetric = null;
    iNumImmediateChoices = 0;
    computeMoveAlphaBeta(gnGameRoot, 0, -5000, 5000);
    System.runFinalization();                                         // Finalize and
    System.gc();                                                      // garbage collect.
    sfDeepMetric = gnGameRoot.getMetric();
    if (!playerResigns(cbCurrent.moveP()))
    {
      // Set the ChessBoard:
      cbBest = gnGameRoot.getBest();
      cbBest.setMove(!cbCurrent.moveP());
      rpTemp = new RememberedPosition(sTemp, cbBest.toString(), iRecursionLimit);

      if (sSessionID != null && bMemorySend  && bMemoryUse && !bMoveSendInhibited)
      {
        // Send the position to the server to remember:
        sTemp = "remember " + rpTemp.toString();
        cmcPutter.close();
        cmcPutter.reconnect();
        cmcPutter.listen();
        cmcPutter.send(sTemp);                                // Send the computed position to the server.
        if (bVerbose) System.out.println("Computed move sent to server.");
      }
      bMemorySend = true;
      computerMoveAfterActions(cbCurrent, cbBest, lStartTime, false);
    }
  } // End of localMove().

  private void computerMoveAfterActions(ChessBoard cbCurrent, ChessBoard cbBest,
                                        long lStartTime, boolean bRememberedPosition)
  {
    int i = 0;
    int iAveragedUsedTime = 0;
    char cPiece = 0;
    boolean bDraw = false;                                   // Flag for draw indication.
    String strMessage = null;
    long lElTime = 0;
    long lEndTime = 0;
    Square sMove = null;
    String strMove = null;
    String strMoveNum = null;
    ChessPiece cpAttacker[] = new ChessPiece[2];
    boolean bMove = cbCurrent.moveP();
    boolean bTemp = false;

    cpAttacker[0] = null;
    cpAttacker[1] = null;

    if (cbBest != null)
    {
      gnGameRoot.setBoard(cbBest);
      repaint();
      sMove = cbBest.getMove();
      if (sMove == null)
      {
        sMove = inferMoveToSquare(cbCurrent.toString(), cbBest.toString());
      }
      if (sMove != null)                                    // Should never be null, defensive programming.
      {
        cPiece = cbBest.getPiece(sMove).getType();          // Irreversible move if cPiece is a pawn.

        // Update the game listing text area:
        strMove = algebraicNotation(cbCurrent, cbBest, sMove, bMove);  // Player's move is last param.
        if (bMove)
        {
          iNumMoves++;
          strMoveNum = Integer.toString(iNumMoves) + ".";
          taGameListing.appendText("\n" + strMoveNum + "\t" + strMove); // New line before white's move.
        }
        else
        {
          taGameListing.appendText("\t" + strMove); // No move number or new line before black's move.
        }
      }
      else
      {
        System.out.println("Null move square.");
      }
      
      lEndTime = System.currentTimeMillis();
      lElTime = lEndTime - lStartTime;
      lElTime /= 1000;
      
      // Did player make a mating move?
      if (cbBest.numChecking(cbBest.getKing(!bMove), cpAttacker, !bMove) > 0)
      {
        if (!bFixedDepth) acCheck.play();
        taGameListing.appendText("+");
      }
      if (cbBest.checkmateP(!bMove))
      {
        if (bLongSounds) acVictory.play();

        strMessage = strMove + ": PM = " +
                     String.valueOf(sfDeepMetric) + ", ET = " +
                     String.valueOf(lElTime) + ". ";
        if (bMove)
        {
          strMessage = strMessage + "Checkmate. White won.";
        }
        else
        {
          strMessage = strMessage + "Checkmate. Black won.";
        }
        lblMessage.setText(strMessage);
        System.out.println(strMessage + '\n');
        bGameOver = true;
        taGameListing.appendText(" Mate");
        if (sSessionID != null)
        {
          btnSave.enable();
        }
      }
      else
      {
        // Is it a draw?
        bDraw = false;
      
        // It's a draw if player's number of legal moves is zero (stalemate).
        if (cbBest.stalemateP(!bMove))
        {
          bDraw = true;
          strMessage = strMove + ": PM = " +
                       String.valueOf(sfDeepMetric) + ", ET = " +
                       String.valueOf(lElTime) + ". Draw by stalemate.";
        }
        
        // It's a draw if the same position with the same player to move has been repeated three times.
        for (i = 0; i < vDuplicate.size(); i++)
        {
          if (((ChessBoard) vDuplicate.elementAt(i)).equalsP(cbBest))
          {
            bDraw = true;
            strMessage = strMove + ": PM = " +
                         String.valueOf(sfDeepMetric) + ", ET = " +
                         String.valueOf(lElTime) + ". Draw by repetition.";
            break;
          }
        }
        if (!bDraw)
        {
          for (i = 0; i < vHistory.size(); i++)
          {
            if (((ChessBoard) vHistory.elementAt(i)).equalsP(cbBest))
            {
              vDuplicate.addElement(new ChessBoard(cbBest));
              if (bMemoryUse)
              {
                if (bVerbose) System.out.println("Duplicate position, adding to vDuplicate and disabling the use of memory.");
                bMemoryUse = false;
                bMemoryUseDisabled = true;
              }
              else
              {
                if (bVerbose) System.out.println("Duplicate position, adding to vDuplicate.");
              }

              // Also increment the search depth to try to get a better move if computer is winning:
              if (sfDeepMetric > 0.5 && bMove)
              {
                iRecursionLimit++;
                if (bVerbose) System.out.println("Duplicate position, white is better, increasing recursion limit to " + iRecursionLimit + ".");
              }
              if (sfDeepMetric < -0.5 && !bMove)
              {
                iRecursionLimit++;
                if (bVerbose) System.out.println("Duplicate position, black is better, increasing recursion limit to " + iRecursionLimit + ".");
              }
              break;
            }
          }
        }
        
        // Update the history Vector:
        if (cbBest.captureP() || cPiece == 'P')               // It's an irreversible move.
        {
          vHistory = new Vector();                            // Wipe out the history.
          vDuplicate = new Vector();
        }
        vHistory.addElement(new ChessBoard(cbBest));

        // It's a draw if there've been no irreversible moves for 50 turns.
        if (vHistory.size() > 100)
        {
          bDraw = true;              // 100 half moves.
          strMessage = "Draw by lack of progress.";
        }
         
        if (bDraw)
        {         
          lblMessage.setText(strMessage);
          System.out.println(strMessage + '\n');
          bGameOver = true;
          taGameListing.appendText(" Draw");
          btnSave.enable();
          acDraw.play();
        }
        else
        {
          strMessage = strMove + ": position metric = " +
                       String.valueOf(sfDeepMetric) + ", elapsed time = " +
                       String.valueOf(lElTime) + " seconds.";
          lblMessage.setText(strMessage);
          strMessage = strMove + ": PM = " +
                       String.valueOf(sfDeepMetric) + ", ET = " +
                       String.valueOf(lElTime) + " seconds, BF = " + iNumImmediateChoices + ".";
          if (bFixedDepth)
          {
            System.out.println(iNumImmediateChoices);
          }
          else
          {
            System.out.println(strMessage);
          }
          timeLoss(bMove);
          if (bRememberedPosition)
          {
            lElTime = iMaxTime;
          }
          iAveragedUsedTime = (((int) lElTime + iPreviousUsedTime1) * 4 + iPreviousUsedTime2 + iPreviousUsedTime3 + 9) / 10;
          if (bVerbose)
          {
            System.out.println("Weighted average time for the last four moves = " + iAveragedUsedTime + " seconds." + '\n');
          }
          if (iPreviousUsedTime1 > iMaxTime && lElTime > iMaxTime)
          {
            lElTime = iMaxTime;
            if (bVerbose)
            {
              System.out.println("Truncating elapsed time to max time.");
            }
          }
          iPreviousUsedTime3 = iPreviousUsedTime2;
          iPreviousUsedTime2 = iPreviousUsedTime1;
          iPreviousUsedTime1 = (int) lElTime;
          bTemp = bDepthChanged;                               // Disallow two depth changes in succession.
          bDepthChanged = false;
          if (iAveragedUsedTime > iMaxTime)                    // The computer should take more than iMinTime
          {                                                    // but less than iMaxTime to move.
            if (iRecursionLimit > 3 || iMinTime < 5)           // Less than 3 only for speedchess mode.
            {
              if (iRecursionLimit > 2)
              {                                                // Never look only 1 tempo ahead, it's pointless.
                if (!bFixedDepth)
                {
                  if (!bTemp)
                  {
                    iRecursionLimit--;
                    bDepthChanged = true;
                    if (bVerbose)
                    {
                      System.out.println("Decrementing the recursion limit to " + iRecursionLimit + ".");
                    }
                  }
                  else
                  {
                    if (bVerbose)
                    {
                      System.out.println("Recursion limit decrementing inhibited.");
                    }
                  }
                }
              }
            }
          }
          else
          {
            if (iAveragedUsedTime < iMinTime)
            {
              if (!bRememberedPosition)                       // Don't bump up the time if it's a remembered position.
              {
                if (!bFixedDepth)
                {
                  if (!bTemp)
                  {
                    iRecursionLimit++;
                    bDepthChanged = true;
                    if (bVerbose)
                    {
                      System.out.println("Incrementing the recursion limit to " + iRecursionLimit + ".");
                    }
                  }
                  else
                  {
                    if (bVerbose)
                    {
                      System.out.println("Recursion limit incrementing inhibited.");
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    else
    {
      // This should never occur in normal play (it means all the pieces are gone, even the kings):
      bGameOver = true;                                      // Defensive programming.
      lblMessage.setText("Game over.");                      // For test and debugging purposes only.
    } // End if (cbBest != null).
    
    // Notify the user:
    if (!bFixedDepth) acNotify.play();
    btnReset.enable();
    if (sSessionID != null)
    {
      if (!bMove) btnSave.enable();                          // Only save non-ended games with white to move.
      btnOpen.enable();
    }
    bThinking = false;                                       // Enable the user to input his move.

  } // End of computerMoveAfterActions().

  private void timeLoss(boolean bMove)
  {
    long lWhiteStopTime = 0;
    long lBlackStopTime = 0;

    if (bMove)
    {
      lWhiteStopTime = System.currentTimeMillis();
      lWhiteCumTime += ((lWhiteStopTime - lWhiteStartTime) / 1000);
      if (bTimedGame && !bTimeLossNotified)
      {
        if (iNumMoves < iNumMovesTimeControl)                               // 40 is the current international standard.
        {
          if (lWhiteCumTime > lMaxCumTime)
          {
            System.out.println("White loses on time (exceeded " + (lMaxCumTime / 60) + " minutes)."); 
            taGameListing.appendText(" White loses on time");
            bTimeLossNotified = true;
          }
        }
        else
        {
          if (lWhiteCumTime > (lMaxCumTime + lMaxExtendedTime))
          {
            System.out.println("White loses on time (exceeded " + ((lMaxCumTime + lMaxExtendedTime) / 60) + " minutes)."); 
            taGameListing.appendText(" White loses on time");
            bTimeLossNotified = true;
          }
        }
      }
      lBlackStartTime = System.currentTimeMillis();
    }
    else
    {
      lBlackStopTime = System.currentTimeMillis();
      lBlackCumTime += ((lBlackStopTime - lBlackStartTime) / 1000);
      if (bTimedGame && !bTimeLossNotified)
      {
        if (iNumMoves < iNumMovesTimeControl)                               // 40 is the current international standard.
        {
          if (lBlackCumTime > lMaxCumTime)
          {
            System.out.println("Black loses on time (exceeded " + (lMaxCumTime / 60) + " minutes)."); 
            taGameListing.appendText(" Black loses on time");
            bTimeLossNotified = true;
          }
        }
        else
        {
          if (lBlackCumTime > (lMaxCumTime + lMaxExtendedTime))
          {
            System.out.println("Black loses on time (exceeded " + ((lMaxCumTime + lMaxExtendedTime) / 60) + " minutes)."); 
            taGameListing.appendText(" Black loses on time");
            bTimeLossNotified = true;
          }
        }
      }
      lWhiteStartTime = System.currentTimeMillis();
    }
    if (bVerbose & bTimedGame)
    { 
      if (bMove)
      {
        System.out.println("White has used " + (lWhiteCumTime / 60) +
                           " minutes and " + (lWhiteCumTime % 60) + " seconds.");
      }
      else
      {
        System.out.println("Black has used " + (lBlackCumTime / 60) +
                           " minutes and " + (lBlackCumTime % 60) + " seconds.");
      }
    }
  } // End of timeLoss()

  private Square inferMoveToSquare(String sChessBoardWas, String sChessBoardIs)
  {
    Square sR = null;
    int i = 0;
    int iToRank[] = new int[2];
    int iToFile[] = new int[2];
    int iNumMovedPieces = 0;
    char cPiece[] = new char[2];

    cPiece[0] = 0;
    cPiece[1] = 0;
    iToRank[0] = 0;
    iToRank[1] = 0;
    iToFile[0] = 0;
    iToFile[1] = 0;

    for (i = 0; i < 64; i++)
    {
      if (sChessBoardWas.charAt(i) != sChessBoardIs.charAt(i) && sChessBoardIs.charAt(i) != '_')
      {
        cPiece[iNumMovedPieces] = sChessBoardIs.charAt(i);
        iToRank[iNumMovedPieces] = i / 8;
        iToFile[iNumMovedPieces] = i % 8;
        iNumMovedPieces++;
      }
    }

    if (iNumMovedPieces == 2)
    {
      if (cPiece[1] == 'S' || cPiece[1] == 's')              // A king has moved during castling.
      {
        iToRank[0] = iToRank[1];
        iToFile[0] = iToFile[1];
      }
    }

    sR = new Square(iToRank[0], iToFile[0]);
    return sR;
  }

  private String algebraicNotation(ChessBoard cbWas, ChessBoard cbIs, Square sMove, boolean bWhite)
  {
    String strR = null;                                      // Return value: algebraic notation string.

    int i = 0;                                               // Rank index.
    int j = 0;                                               // File index.
    int k = 0;                                               // Legal move index.
    int iRank = sMove.getRank();
    int iFile = sMove.getFile();
    String strRank = Integer.toString(8 - iRank);
    String strFile = String.valueOf((char) (iFile + 97));
    ChessPiece cp = cbIs.getPiece(sMove);
    char cPiece = cp.getType();
    ChessPiece cpTemp = null;
    Square sTemp = null;
    int iNumPieces = 0;
    Square sFrom[] = null;
    sFrom = new Square[10];                                  // Could have pawns promoted to piece type.
    String strDisAmbRank = "";
    String strDisAmbFile = "";
    String strCapture = "";
    float sfWhitePower0 = 0;
    float sfBlackPower0 = 0;
    float sfWhitePower1 = 0;
    float sfBlackPower1 = 0;

    if (cbIs.captureP())
    {
      strCapture = "x";
    }

    // Disambiguate if necessary: the move is ambiguous if two of the kind of piece could have gotten
    // to the sMove square. Search the cbWas to count how many of the piece type had a legal move
    // to the sMove square:

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        cpTemp = cbWas.getPiece(i, j);
        if (cpTemp != null)
        {
          if ((cpTemp.whiteP() == bWhite) &&                          // My color.
              (cpTemp.getType() == cPiece))                           // Same type of piece.
          {
            for (k = 0; k < cpTemp.numLegal(); k ++)
            {
              sTemp = cpTemp.getLegalMove(k);
              if (sTemp.equals(sMove))
              {
                sFrom[iNumPieces] = new Square(cpTemp.getSquare());
                iNumPieces++;
              }
            }
          }
          if (cpTemp.whiteP())
          {
            sfWhitePower0 += cpTemp.getValue();
          }
          else
          {
            sfBlackPower0 += cpTemp.getValue();
          }
        }
        cpTemp = cbIs.getPiece(i, j);
        if (cpTemp != null)
        {
          if (cpTemp.whiteP())
          {
            sfWhitePower1 += cpTemp.getValue();
          }
          else
          {
            sfBlackPower1 += cpTemp.getValue();
          }
        }
      } 
    }

    if (iNumPieces > 1)                                              // Ambiguous move.
    {
      for (i = 0; i < iNumPieces; i++)
      {
        if (cbIs.getPiece(sFrom[i]) == null)
        {
          // We know where the piece moved from.
          strDisAmbRank = Integer.toString(8 - sFrom[i].getRank());
          strDisAmbFile = String.valueOf((char) (sFrom[i].getFile() + 97));
        }
      }
    }
    else
    {
      if (iNumPieces == 1 && cPiece == 'P' && cbIs.captureP())      // Non-mbiguous move of a pawn capture.
      {
        if (cbIs.getPiece(sFrom[0]) == null)
        {
          // We know where the pawn moved from, list only the file (algebraic notation convention).
          strDisAmbFile = String.valueOf((char) (sFrom[0].getFile() + 97));
        }
      }
    }

    // Assign the notation:
    if (cPiece == 'P')
    {
      strR = new String(strDisAmbFile + strDisAmbRank + strCapture + strFile + strRank);
    }
    else
    {
      strR = new String(String.valueOf(cPiece) + strDisAmbFile + strDisAmbRank + strCapture + strFile + strRank);
    }

    // Castling?
    if (iNumPieces == 1 && cPiece == 'K' && !cbIs.captureP())      // Maybe the king has castled.
    {
      if (cbIs.getPiece(sFrom[0]) == null)
      {
        // We know where the pawn moved from, list only the file (algebraic notation convention).
        if (sFrom[0].getFile() == 4)
        {
          if (sMove.getFile() == 6)
          {
            strR = "O-O";                                          // Castles king side.
          }
          else
          {
            if (sMove.getFile() == 2)
            {
              strR = "O-O-O";                                      // Castles long (queen side).
            }
          }
        }
      }
    }

    // Pawn promotion?
    if ((sfWhitePower1 > sfWhitePower0 + 1) ||
        (sfBlackPower1 > sfBlackPower0 + 1))                       // Power has increased, only promotion can do this.
    {
      if (cbIs.captureP())
      {
        // Need to find the square the pawn came from:
        for (i = 0; i < 8; i++)
        {
          if (cbWas.getPiece(1, i) != null && cbIs.getPiece(1, i) == null)
          {
            strDisAmbFile = String.valueOf((char) (i + 97));
            break;
          }
          if (cbWas.getPiece(6, i) != null && cbIs.getPiece(6, i) == null)
          {
            strDisAmbFile = String.valueOf((char) (i + 97));
            break;
          }
        }
        strR = strDisAmbFile + "x" + strFile + strRank + "=" + String.valueOf(cPiece);
      }
      else
      {
        strR = strFile + strRank + "=" + String.valueOf(cPiece);
      }
    }

    return strR;

  } // End of algebraicNotation().

  public void sortGames()
  {
    int i = 0;
    int j = 0;
    int iNumGames = lGamesList.countItems();
    String strSwapSpace = null;
    Vector vGames = new Vector();

    // Can't use List.replaceItem() due to a spurious error message. Instead, read the strings into
    // a Vector, sort them there, and read them back.

    for (i = 0; i < iNumGames; i++)
    {
      vGames.addElement(lGamesList.getItem(i));
    }
    lGamesList.clear();

    for (i = 0; i < iNumGames - 1; i++)                                          // Bubble sort.
    {
      for (j = i + 1; j < iNumGames; j++)
      {
        if (gameNumber((String) vGames.elementAt(i)) > gameNumber((String) vGames.elementAt(j)))
        {
          strSwapSpace = (String) vGames.elementAt(i);
          vGames.setElementAt(vGames.elementAt(j), i);
          vGames.setElementAt(strSwapSpace, j);
        }
      }
    }

    for (i = 0; i < iNumGames; i++)
    {
      lGamesList.addItem((String) vGames.elementAt(i));
    }
  }

  public int gameNumber(String s)
  {
    String sGameNum = null;
    int iGameNum = 0;

    sGameNum = s.substring(1);
    sGameNum = sGameNum.substring(0, sGameNum.length() - 4);
    try
    {
      iGameNum = Integer.parseInt(sGameNum);
    }
    catch (NumberFormatException e9)
    {
    }
    return iGameNum;
  }

  public boolean action(Event e, Object o)
  {
    String sSend = null;
    String sGameNum = null;

    if (e.target == btnReset)
    {
      sSend = "Request SessionID, " + sHeader;
      if (bHumanHuman)
      {
        sSend = sSend + ", human vs. human mode";
      }
      cdc.close();
      cdc.reconnect();
      cdc.listen();
      cdc.send(sSend);
      if (bLongSounds)
      {
        if (Math.random() < 0.5)
        {
          acReset.play();
        }
        else
        {
          acHALChess.play();
        }
      }
      iRecursionLimit = iParamRecurLimit;                           // Reset the recursion limit.
      lblMessage.setText("Click the white piece you want to move, " +
                         "then click the move-to square.");
      System.out.println("\nNew game.");
      setupBoard();                                                 // Calls setPerfBoost().
      sHumanFrom = null;
      bGameOver = false;
      repaint();
      iNumMoves = 0;
      taGameListing.setText("\tWhite\tBlack");
      btnReset.disable();
      btnSave.disable();
    }

    if (e.target == btnListing)
    {
      if (taGameListing.isShowing())
      {
        taGameListing.hide();
        btnListing.setLabel("Show Listing");
      }
      else
      {
        taGameListing.show();
        btnListing.setLabel("Hide Listing");
      }
    }
 
    if (e.target == btnSound)
    {
      bLongSounds = !bLongSounds;
      if (bLongSounds)
      {
        lblMessage.setText("Long sounds enabled.");
        btnSound.setLabel("Disable Sound");
      }
      else
      {
        lblMessage.setText("Long sounds disabled.");
        btnSound.setLabel("Enable Sound");
      }
    }

    if (e.target == btnSave)
    {
      try
      {
        if (sSessionID == null)
        {
          lblMessage.setText("Game not saved, no connection to the chess data server.");
          btnOpen.disable();
        }
        else
        {

	  //moved by captain
	  lblMessage.setText("Saving game " + Integer.parseInt(sSessionID) + ".");
          
	  cdc.close();
          cdc.reconnect();
          cdc.listen();

          sSend = sSessionID;
          cdc.send(sSend);

          sSend = gnGameRoot.getBoard().toString();
          cdc.send(sSend);

          sSend = taGameListing.getText();
          cdc.send(sSend);

          sSend = "end";
          cdc.send(sSend);
        }
      }
      catch (NullPointerException e8)
      {
        System.out.println("Caught error in action() cdc: " + e8);
      }
      btnSave.disable();
    }

    if (e.target == btnOpen)
    {
      lblMessage.setText("Retrieving game list, please wait.");
      cdc.close();
      cdc.reconnect();
      cdc.listen();
      cdc.send("directory request");
      btnOpen.disable();
      bThinking = true;
    }

    if (e.target == lGamesList)
    {
      sSend = lGamesList.getSelectedItem();
      iOpenGameNumber = gameNumber(sSend);
      sGameNum = String.valueOf(iOpenGameNumber);
      sSend = "open " + sSend;
      try
      {
        if (sSessionID == null)
        {
          lblMessage.setText("Game " + sGameNum + " not opened, no connection to the chess data server.");
          btnSave.disable();
        }
        else
        {
          if (iOpenGameNumber == 0)
          {
            lblMessage.setText("Open game canceled.");
            bThinking = false;
            btnOpen.enable();
          }
          else
          {
            lblMessage.setText("Opening game " + sGameNum + ", please wait.");
            cdc.close();
            cdc.reconnect();
            cdc.listen();
            cdc.send(sSend);
            lWhiteCumTime = 0;
            lBlackCumTime = 0;
            lWhiteStartTime =  System.currentTimeMillis();
            lBlackStartTime =  System.currentTimeMillis();
            bTimeLossNotified = false;
          }
        }
      }
      catch (NullPointerException e8)
      {
        System.out.println("Caught error in action() cdc: " + e8);
      }
      lGamesList.hide();
      System.runFinalization();
      System.gc();
    }

    requestFocus();
    return true;                                              // Absorb the event;

  } // End of action().

  public boolean keyDown(Event e, int k)
  {
    char cPromoteTo = 0;
    ChessBoard cbCurrent = null;
    ChessBoard cbTemp = null;
    int iRank = 0;
    int iFile = 0;
    String strMoveNum = null;
    String strMove = null;
    String strMessage = null;
    ChessPiece cpAttacker[] = new ChessPiece[2];
    cpAttacker[0] = null;
    cpAttacker[1] = null;
    boolean bMove = false;
    int iBlackColor = 0;
    int iWhiteColor = 0;

    if (k == 91)
    {
      k = 27;
    }

    if (k == 93)
    {
      k = 29;
    }

    if (e.target == taGameListing)
    {
      return false;
    }
    else
    {
      switch (k)
      {
        case 1:                                                    // Control-A. Switch on auto-play
        {
          if (!bThinking || bAutoPlay)
          {
            if (true)
            {
              bAutoPlay = !bAutoPlay;
              if (bHumanHuman)
              {
                bHumanHuman = false;
                System.out.println("Human-human play mode deactivated.");
              }
              if (bAutoPlay)
              {
                System.out.println("Auto-play on.");
                tCompute = new Thread(this);                         // Create a new thread.
                tCompute.start();                                    // Kick off the run() function.
              }
              else
              {
                System.out.println("Auto-play off.");
              }
            }
            else
            {
              System.out.println("Can't auto-play when in server move retrieve mode.");
            }
          }
          else
          {
            System.out.println("Can't set auto-play on while thinking.");
          }
          break;
        }        
        case 2:                                                    // Control-B.
        {
          bMemoryUse = !bMemoryUse;                                // Toggles memory use.
          if (bMemoryUse)
          {
            System.out.println("Memory use enabled.");
          }
          else
          {
            System.out.println("Memory use disabled.");
          }
          break;
        }
        case 3:                                                    // Control-C. 
        {                                                          // Toggles computer's side.
          if (!bThinking && !bAutoPlay)
          {
            if (bHumanHuman)
            {
              bHumanHuman = false;
              System.out.println("Human-human play mode deactivated.");
            }
            cbCurrent = gnGameRoot.getBoard();
            gnGameRoot = new GameNode();                           // Eliminate any old children.
            gnGameRoot.setMetric(cbCurrent.evaluate());            // Evaluate the board.
            gnGameRoot.setBoard(cbCurrent);                        // Set the board for the game root.
            if (cbCurrent.moveP())
            {
              if (bVerbose) System.out.println("White to move.");
            }
            else
            {
              if (bVerbose) System.out.println("Black to move.");
            }
            tCompute = new Thread(this);                           // Create a new thread.
            tCompute.start();                                      // Kick off the run() function.
          }
          else
          {
            System.out.println("Can't switch the computer's side while thinking.");
          }
          break;
        }
        case 4:                                                    // Control-D.
        {
          if (!bThinking && !bAutoPlay)                            // Decrements search depth.
          {
            if (iRecursionLimit > 2)
            {
              iRecursionLimit--;
              System.out.println("Search depth decremented to " + iRecursionLimit + ".");
            }
            else
            {
              System.out.println("Can't decrement the search depth to less than two.");
            }
          }
          else
          {
            System.out.println("Can't decrement the search depth while thinking.");
          }
          break;
        }
        case 8:                                                    // Control-H.
        {
          bHumanHuman = !bHumanHuman;                              // Toggles human-human mode.
          if (bHumanHuman)
          {
            System.out.println("Human-human play mode activated.");
            strMessage = "Request SessionID, " + sHeader + ", human vs. human mode";
            cdc.close();
            cdc.reconnect();
            cdc.listen();
            cdc.send(strMessage);
          }
          else
          {
            System.out.println("Human-human play mode deactivated.");
          }
          break;
        }
        case 9:                                                    // Control-I.
        {
          if (!bThinking && !bAutoPlay)                            // Increments search depth.
          {
            iRecursionLimit++;
            System.out.println("Search depth incremented to " + iRecursionLimit + ".");
          }
          else
          {
            System.out.println("Can't increment the search depth while thinking.");
          }
          break;
        }
        case 13:                                                   // Control-M.
        {
          bThinking = false;                                       // Stop thinking and move now.
          bMemorySend = false;                                     // Don't store a forced move.
          break;
        }
        case 14:                                                   // Control-N.
        {
          bNewby = !bNewby;                                        // Toggle the newby mode.
          if (bVerbose)
          {
            System.out.println("Newby mode toggled to " + bNewby + ".");
          }
          break;
        }
        case 16:                                                   // Control-P.
        {
          System.out.println(gnGameRoot.getBoard().toString());    // Print the current board.
          break;
        }
        case 18:                                                   // Control-R.
        {
          bRotated = !bRotated;                                    // Toggle board orientation.
          if (bRotated)
          {
            System.out.println("Rotated to black below.");
          }
          else
          {
            System.out.println("Rotated to white below.");
          }
          repaint();
          break;
        }
        case 19:                                                   // Control-S.
        {
          System.out.println("Sending server request to save backup of memory.");
          strMessage = "Save backup";
          cmc.close();
          cmc.reconnect();
          cmc.listen();
          cmc.send(strMessage);
          break;
        }
        case 20:                                                   // Control-T.
        {
          bTimedGame = !bTimedGame;                                // Toggle timed game mode.
          if (bTimedGame)
          {
            System.out.println("Timed game mode on.");
          }
          else
          {
            System.out.println("Timed game mode off.");
          }
          break;
        }
        case 22:                                                   // Control-V.
        {
          bVerbose = !bVerbose;                                    // Toggle verbose mode.
          if (bVerbose)
          {
            System.out.println("Verbose mode on.");
            System.out.println("Minimum time = " + iMinTime + " seconds.");
            System.out.println("Maximum time = " + iMaxTime + " seconds.");
            System.out.println("Fixed depth = " + bFixedDepth + ".");
            System.out.println("Recursion limit = " + iRecursionLimit + ".");
          }
          else
          {
            System.out.println("Verbose mode off.");
          }
          break;
        }
        case 27:                                                   // Control-[ (left bracket), decrease contrast.
        {
          switch (iPieceColorScheme)
          {
            case 0:                                                // Default coffee and cream.
            {
              // Decrease piece color contrast:
              iBlackColor = cBlackPiece.getRed();
              iWhiteColor = cWhitePiece.getBlue();

              iBlackColor += 12;
              iWhiteColor -= 8;

              if (iBlackColor <= 96)
              {
                cBlackPiece = new Color(iBlackColor, iBlackColor, 0);
                cWhitePiece = new Color(255, 255, iWhiteColor);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is minimal.");
              }
              break;
            }
            case 1:                                                // Greenish.
            {
              // Decrease piece color contrast:
              iBlackColor = cBlackPiece.getGreen();
              iWhiteColor = cWhitePiece.getRed();

              iBlackColor += 16;
              iWhiteColor -= 4;

              if (iBlackColor <= 128)
              {
                cBlackPiece = new Color(0, iBlackColor, 0);
                cWhitePiece = new Color(iWhiteColor, 255, iWhiteColor);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is minimal.");
              }
              break;
            }
            case 2:                                                // Bluish.
            {
              // Decrease piece color contrast:
              iBlackColor = cBlackPiece.getBlue();
              iWhiteColor = cWhitePiece.getRed();

              iBlackColor += 12;
              iWhiteColor -= 8;

              if (iBlackColor <= 96)
              {
                cBlackPiece = new Color(0, iBlackColor, iBlackColor);
                cWhitePiece = new Color(iWhiteColor, 255, 255);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is minimal.");
              }
              break;
            }
            case 3:                                                // Purplish.
            {
              // Decrease piece color contrast:
              iBlackColor = cBlackPiece.getBlue();
              iWhiteColor = cWhitePiece.getRed();

              iBlackColor += 16;
              iWhiteColor -= 4;

              if (iBlackColor <= 128)
              {
                cBlackPiece = new Color(0, 0, iBlackColor);
                cWhitePiece = new Color(iWhiteColor, iWhiteColor, 255);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is minimal.");
              }
              break;
            }
            case 4:                                                // Pinkish.
            {
              // Decrease piece color contrast:
              iBlackColor = cBlackPiece.getRed();
              iWhiteColor = cWhitePiece.getGreen();

              iBlackColor += 12;
              iWhiteColor -= 8;

              if (iBlackColor <= 96)
              {
                cBlackPiece = new Color(iBlackColor, 0, iBlackColor);
                cWhitePiece = new Color(255, iWhiteColor, 255);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is minimal.");
              }
              break;
            }
            case 5:                                                // Orangish.
            {
              // Decrease piece color contrast:
              iBlackColor = cBlackPiece.getRed();
              iWhiteColor = cWhitePiece.getGreen();

              iBlackColor += 16;
              iWhiteColor -= 4;

              if (iBlackColor <= 128)
              {
                cBlackPiece = new Color(iBlackColor, 0, 0);
                cWhitePiece = new Color(255, iWhiteColor, iWhiteColor);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is minimal.");
              }
              break;
            }
          }
          break;
        }
        case 29:                                                   // Control-] (right bracket).
        {
          switch (iPieceColorScheme)
          {
            case 0:                                                // Default coffee and cream.
            {
              // Increase piece color contrast:
              iBlackColor = cBlackPiece.getRed();
              iWhiteColor = cWhitePiece.getBlue();

              iBlackColor -= 12;
              iWhiteColor += 8;

              if (iWhiteColor <= 255)
              {
                cBlackPiece = new Color(iBlackColor, iBlackColor, 0);
                cWhitePiece = new Color(255, 255, iWhiteColor);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is maximal.");
              }
              break;
            }
            case 1:                                                // Greenish.
            {
              // Increase piece color contrast:
              iBlackColor = cBlackPiece.getGreen();
              iWhiteColor = cWhitePiece.getRed();

              iBlackColor -= 16;
              iWhiteColor += 4;

              if (iWhiteColor <= 255)
              {
                cBlackPiece = new Color(0, iBlackColor, 0);
                cWhitePiece = new Color(iWhiteColor, 255, iWhiteColor);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is maximal.");
              }
              break;
            }
            case 2:                                                // Bluish.
            {
              // Increase piece color contrast:
              iBlackColor = cBlackPiece.getBlue();
              iWhiteColor = cWhitePiece.getRed();

              iBlackColor -= 12;
              iWhiteColor += 8;

              if (iWhiteColor <= 255)
              {
                cBlackPiece = new Color(0, iBlackColor, iBlackColor);
                cWhitePiece = new Color(iWhiteColor, 255, 255);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is maximal.");
              }
              break;
            }
            case 3:                                                // Purplish.
            {
              // Increase piece color contrast:
              iBlackColor = cBlackPiece.getBlue();
              iWhiteColor = cWhitePiece.getRed();

              iBlackColor -= 16;
              iWhiteColor += 4;

              if (iWhiteColor <= 255)
              {
                cBlackPiece = new Color(0, 0, iBlackColor);
                cWhitePiece = new Color(iWhiteColor, iWhiteColor, 255);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is maximal.");
              }
              break;
            }
            case 4:                                                // Bluish.
            {
              // Increase piece color contrast:
              iBlackColor = cBlackPiece.getRed();
              iWhiteColor = cWhitePiece.getGreen();

              iBlackColor -= 12;
              iWhiteColor += 8;

              if (iWhiteColor <= 255)
              {
                cBlackPiece = new Color(iBlackColor, 0, iBlackColor);
                cWhitePiece = new Color(255, iWhiteColor, 255);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is maximal.");
              }
              break;
            }
            case 5:                                                // Orangish.
            {
              // Increase piece color contrast:
              iBlackColor = cBlackPiece.getRed();
              iWhiteColor = cWhitePiece.getGreen();

              iBlackColor -= 16;
              iWhiteColor += 4;

              if (iWhiteColor <= 255)
              {
                cBlackPiece = new Color(iBlackColor, 0, 0);
                cWhitePiece = new Color(255, iWhiteColor, iWhiteColor);
                repaint();
              }
              else
              {
                System.out.println("Piece color contrast is maximal.");
              }
              break;
            }
          }
          break;
        }
        case 49:                                                   // "1", ivory piece color scheme assignment.
        {
          iPieceColorScheme = 0;
          cBlackPiece = new Color(48, 48, 0);
          cWhitePiece = new Color(255, 255, 223);
          repaint();
          break;
        }
        case 50:                                                   // "2", greenish piece color scheme assignment.
        {
          iPieceColorScheme = 1;
          cBlackPiece = new Color(0, 64, 0);
          cWhitePiece = new Color(239, 255, 239);
          repaint();
          break;
        }
        case 51:                                                   // "3", bluish piece color scheme assignment.
        {
          iPieceColorScheme = 2;
          cBlackPiece = new Color(0, 48, 48);
          cWhitePiece = new Color(223, 255, 255);
          repaint();
          break;
        }
        case 52:                                                   // "4", purplish piece color scheme assignment.
        {
          iPieceColorScheme = 3;
          cBlackPiece = new Color(0, 0, 64);
          cWhitePiece = new Color(239, 239, 255);
          repaint();
          break;
        }
        case 53:                                                   // "5", pinkish piece color scheme assignment.
        {
          iPieceColorScheme = 4;
          cBlackPiece = new Color(48, 0, 48);
          cWhitePiece = new Color(255, 223, 255);
          repaint();
          break;
        }
        case 54:                                                   // "6", orangish piece color scheme assignment.
        {
          iPieceColorScheme = 5;
          cBlackPiece = new Color(64, 0, 0);
          cWhitePiece = new Color(255, 239, 239);
          repaint();
          break;
        }
        default:                                                      // See if it's a pawn promotion key input:
        {
          if (k == 'n') k = 'N';
          if (k == 'b') k = 'B';
          if (k == 'r') k = 'R';
          if (k == 'q') k = 'Q';
          if (k == 'N' || k == 'B' || k == 'R' || k == 'Q')
          {
            cPromoteTo = (char) k;
            if (sPromoteSquare != null && bNeedPromoting)
            {
              cbCurrent = gnGameRoot.getBoard();
              bMove = cbCurrent.moveP();
              cbTemp = new ChessBoard(cbCurrent);                     // Copy the current chessboard.
              cpPromote.setType(cPromoteTo);
              cbTemp.setPiece(cpPromote, sPromoteSquare);
              
              // Refresh legal moves for all the pieces on the board:
              cbTemp.setMove(!bMove);                                 // Opponent has the move.
              cbTemp.refreshLegal();
              bNeedPromoting = false;
              gnGameRoot = new GameNode();                            // Eliminate any old children.
              gnGameRoot.setBoard(cbTemp);
              repaint();
              
              // Update the game listing text area:
              iRank = sPromoteSquare.getRank();
              iFile = sPromoteSquare.getFile();
              strMove = algebraicNotation(cbCurrent, cbTemp, sPromoteSquare, bMove);
              if (bMove)
              {
                strMoveNum = Integer.toString(iNumMoves) + ".";
                taGameListing.appendText("\n" + strMoveNum + "\t" + strMove); // New line before white's move.
              }
              else
              {
                taGameListing.appendText("\t" + strMove); // No move number or new line before black's move.
              }
             
              // Did the player make a mating move?
              if (cbTemp.numChecking(cbTemp.getKing(!bMove), cpAttacker, !bMove) > 0)
              {
                if (!bFixedDepth) acCheck.play();
                taGameListing.appendText("+");
              }
              if (cbTemp.checkmateP(!bMove))
              {
                if (bLongSounds) acDefeat.play();
                lblMessage.setText("Checkmate.");
                bGameOver = true;
                taGameListing.appendText(" Mate");
                btnReset.enable();
                if (sSessionID != null)
                {
                  btnSave.enable();
                  btnOpen.enable();
                }
              }
              else
              {
                // Is it stalemate?
                if (cbTemp.stalemateP(!bMove))
                {
                  strMessage = "Draw by stalemate.";
                  lblMessage.setText(strMessage);
                  System.out.println(strMessage + '\n');
                  bGameOver = true;
                  taGameListing.appendText(" Stalemate");
                  btnReset.enable();
                  if (sSessionID != null)
                  {
                    btnSave.enable();
                    btnOpen.enable();
                  }
                }
                else
                {
                  timeLoss(bMove);
                  if (bHumanHuman)
                  {
                    btnReset.enable();
                    if (sSessionID != null)
                    {
                      btnSave.enable();
                      btnOpen.enable();
                    }
                    if (bMove)
                    {
                      lblMessage.setText("Click the black piece you want to move, " +
                                        "then click the move-to square.");
                    }
                    else
                    {
                      lblMessage.setText("Click the white piece you want to move, " +
                                        "then click the move-to square.");
                    }
                  }
                  else
                  {
                    // Compute opponent's move:
                    vHistory = new Vector();                             // Wipe out the history (irreversible move).
                    vDuplicate = new Vector();
                    vHistory.addElement(new ChessBoard(cbTemp));
                    tCompute = new Thread(this);
                    tCompute.start();
                  }
                }
              }
            }
          }
          else
          {
            cPromoteTo = 0;
            if (k >= 32 && k <= 127 && bNeedPromoting)
            {
              acMoveIllegal.play();
            }
          }
          return false;
        }
      }
      return true;
    }
  } // End of keyDown().

  public boolean mouseDown(Event e, int x, int y)            // Handle the down part of the click event.
  {
    int i = 0;
    Square sHumanTo = null;
    boolean bLegal = false;
    int iCountLegal = 0;
    ChessPiece cpTemp = null;
    ChessBoard cbCurrent = null;
    ChessBoard cbCurrentCopy = null;
    ChessBoard cbTestCheck = null;
    ChessPiece cpKing = null;
    ChessPiece cpTestMove = null;
    ChessPiece cpAttacker[] = new ChessPiece[2];
    cpAttacker[0] = null;
    cpAttacker[1] = null;
    int iNumChecking = 0;
    String strMove = null;
    String strMoveNum = null;
    String strMessage = null;
    char cPiece = 0;
    boolean bEPC = false;
    boolean bDraw = false;                                   // Flag for draw indication.
    boolean bMove = false;
    boolean bEPCapture = false;                              // May 2016 bug fix for EP capture when king is in check

    requestFocus();

    if (!bThinking && !bNeedPromoting)                       // Ignore the user if I'm thinking
    {                                                        // or need a keystroke for promotion.
      cbCurrent = gnGameRoot.getBoard();
      bMove = cbCurrent.moveP();
      cbCurrentCopy = new ChessBoard(cbCurrent);
      cbTestCheck = new ChessBoard(cbCurrent);
      if (sHumanFrom == null)
      {
        // Get the from square.
        sHumanFrom = getSquare(x, y);
        if (sHumanFrom != null)
        {
          // Make sure there's a player's piece on the from square, it has legal moves, and the game isn't over:
          cpTemp = cbCurrent.getPiece(sHumanFrom);
          if (cpTemp != null &&
              cpTemp.whiteP() == bMove &&
              cpTemp.numLegal() > 0 &&
              !bGameOver)
          {
            // Highlight the from square:
            repaint();
          }
          else
          {
            sHumanFrom = null;
          }
        }
      }
      else
      {
        // We have the from square, get the to square:
        sHumanTo = getSquare(x, y);
        if (sHumanTo != null)
        {
          // Make sure it's a legal move:
          cpTemp = cbCurrent.getPiece(sHumanFrom);
          if(cpTemp.legalP(sHumanTo))
          {
            // Verify it's not moving into nor discovering check:
            // Make the move on the test check chessboard:
            cpTestMove = new ChessPiece(cpTemp);
            cbTestCheck.setPiece(cpTestMove, sHumanTo);
            cbTestCheck.refreshLegal();
             
            // Get the king:
            cpKing = cbTestCheck.getKing(bMove);
            
            // Is the king in check?
            iNumChecking = cbTestCheck.numChecking(cpKing, cpAttacker, bMove);

            // Boolean test to fix May 2016 bug for EP capture when king is in check.                       
            if (cpAttacker[0] != null)
            {
              if (sHumanTo.getFile() ==  cpAttacker[0].getSquare().getFile() &&
                  cpTemp.getType() == 'P' &&
                  cpAttacker[0].getType() == 'P' &&
                  cpTemp.EPCP(sHumanTo))
              {
                bEPCapture = true;
              }
            }

            if (iNumChecking == 0 || bEPCapture)
            {
              if (bMove)
              {
                iNumMoves++;                                           // Increment game list move counter.
              }
              // See if the move is a pawn promotion:              
              if (cpTemp.getType() == 'P' && (sHumanTo.getRank() == 0 || sHumanTo.getRank() == 7))
              {
                // It's a pawn promotion
                sPromoteSquare = sHumanTo;
                if (cbCurrent.getPiece(sPromoteSquare) != null)
                {
                  sPromoteSquare.setCapture(true);
                }
                cpPromote = new ChessPiece(cpTemp);
                
                bNeedPromoting = true;
                lblMessage.setText("Hit N, B, R, or Q key to indicate piece for pawn promotion.");
                acPromote.play();
                requestFocus();
              }
              else
              {
                // Move the piece:
                if (cpTemp.EPCP(sHumanTo))
                {
                  sHumanTo.setEPC(true);
                }
                if (cbCurrent.getPiece(sHumanTo) != null)
                {
                  sHumanTo.setCapture(true);
                }
                cbCurrent.setPiece(cpTemp, sHumanTo);
                cPiece = cpTemp.getType();
                
                // Update the game listing text area:
                strMove = algebraicNotation(cbCurrentCopy, cbCurrent, sHumanTo, bMove);
                if (bMove)
                {
                  strMoveNum = Integer.toString(iNumMoves) + ".";
                  taGameListing.appendText("\n" + strMoveNum + "\t" + strMove); // New line before white's move.
                }
                else
                {
                  taGameListing.appendText("\t" + strMove); // No line number line before black's move.
                }
                      
                // Refresh legal moves for all the pieces on the board:
                cbCurrent.setMove(!bMove);
                cbCurrent.refreshLegal();
                   
                // Did the player make a mating move?
                if (cbCurrent.numChecking(cbCurrent.getKing(!bMove), cpAttacker, !bMove) > 0)
                {
                  if (!bFixedDepth) acCheck.play();
                  taGameListing.appendText("+");
                }
                if (cbCurrent.checkmateP(!bMove))
                {
                  if (bLongSounds) acDefeat.play();
                  if (bMove)
                  {
                    lblMessage.setText("Checkmate. White won.");
                  }
                  else
                  {
                    lblMessage.setText("Checkmate. Black won.");
                  }
                  bGameOver = true;
                  taGameListing.appendText(" Mate");
                  btnReset.enable();
                  if (sSessionID != null)
                  {
                    btnSave.enable();
                    btnOpen.enable();
                  }
                }
                else
                {
                  // Did the player make a drawing move?
                  bDraw = false;
                  if (cbCurrent.stalemateP(!bMove))
                  {
                    bDraw = true;
                    strMessage = "Draw by stalemate.";
                    taGameListing.appendText(" Stalemate");
                  }
                   
                  // It's a draw if the same position with the same player to move has been repeated three times.
                  for (i = 0; i < vDuplicate.size(); i++)
                  {
                    if (((ChessBoard) vDuplicate.elementAt(i)).equalsP(cbCurrent))
                    {
                      bDraw = true;
                      strMessage = "Draw by repetition.";
                      break;
                    }
                  }
                  if (!bDraw)
                  {
                    for (i = 0; i < vHistory.size(); i++)
                    {
                      if (((ChessBoard) vHistory.elementAt(i)).equalsP(cbCurrent))
                      {
                        vDuplicate.addElement(new ChessBoard(cbCurrent));
                      }
                    }
                  }
                     
                  // Update the history Vector:
                  if (cbCurrent.captureP() || cPiece == 'P')               // It's an irreversible move.
                  {
                    vHistory = new Vector();                               // Wipe out the history.
                    vDuplicate = new Vector();
                  }
                  vHistory.addElement(new ChessBoard(cbCurrent));
                      
                  // It's a draw if there've been no irreversible moves for 50 turns.
                  if (vHistory.size() > 100)
                  {
                    bDraw = true;                  // 100 half moves.
                    strMessage = "Draw by lack of progress.";
                  }
                  if (bDraw)
                  {
                    lblMessage.setText(strMessage);
                    System.out.println(strMessage + '\n');
                    bGameOver = true;
                    taGameListing.appendText(" Draw");
                    btnReset.enable();
                    if (sSessionID != null)
                    {
                      btnSave.enable();
                      btnOpen.enable();
                    }
                  }
                  else
                  {
                    timeLoss(bMove);
                    if (bHumanHuman)
                    {
                      btnReset.enable();
                      if (sSessionID != null)
                      {
                        if (!bMove)
                        {
                          btnSave.enable();
                        }
                        else
                        {
                          btnSave.disable();
                        }
                        btnOpen.enable();
                      }
                      if (bMove)
                      {
                        lblMessage.setText("Click the black piece you want to move, " +
                                          "then click the move-to square.");
                      }
                      else
                      {
                        lblMessage.setText("Click the white piece you want to move, " +
                                            "then click the move-to square.");
                      }
                    }
                    else
                    {
                      // Compute opponent's move:
                      if (cbCurrent.moveP())
                      {
                        if (bVerbose) System.out.println("White to move.");
                      }
                      else
                      {
                        if (bVerbose) System.out.println("Black to move.");
                      }
                      gnGameRoot = new GameNode();                         // Eliminate any old children.
                      gnGameRoot.setMetric(cbCurrent.evaluate());          // Evaluate the board.
                      gnGameRoot.setBoard(cbCurrent);                      // Set the board for the game root.
                      tCompute = new Thread(this);                         // Create a new thread.
                      tCompute.start();                                    // Kick off the run() function.
                    }
                  }
                  repaint();
                }
              }
            }
            else
            {
              // The king is in check
              // Play the vomit sound:
              
              acMoveIntoCheck.play();
              lblMessage.setText("Can't move into check.");
            }
          }
          else
          {
            if (sHumanFrom.equals(sHumanTo))
            {
              lblMessage.setText("Move canceled.");
            }
            else
            {
              acMoveIllegal.play();
              lblMessage.setText("That's not a legal move.");
            }
          }
          sHumanFrom = null;
          repaint();
        }
      }
    }
    return true;                                             // Absorb the event.
          
  } // End of mouseDown().

  public boolean mouseMove(Event e, int x, int y)            // Handle the mouse-over event.
  {
    int i = 0;
    int j = 0;
    int k = 0;
    Square sOver = null;
    ChessBoard cbCurrent = gnGameRoot.getBoard();
    ChessPiece cpTemp = null;
    boolean bMove = cbCurrent.moveP();

    requestFocus();                                          // Mousing over sets focus for keyboard events. May 18, 2016
                                                             // requestFocus in mouseMove() seems not to work in MSIE, perhaps other browsers
    if (!bThinking &&                                        // Ignore the user mouse-move if I'm thinking,
        !bNeedPromoting &&                                   // or getting keyboard input, or have selected
        sHumanFrom == null &&                                // a sqaure, or the game is over, or I'm not.
        !bGameOver &&                                        // a newby.
        bNewby)
    {                                                        
      sOver = getSquare(x, y);
      if (sOver != null)
      {
        for (i = 0; i < 8; i++)
        {
          if (sOver.getRank() == i)
          {
            for (j = 0; j < 8; j++)
            {
              if (sOver.getFile() == j)
              {
                cpTemp = cbCurrent.getPiece(i, j);
                if (cpTemp != null)
                {
                  if (cpTemp.whiteP() == bMove)
                  {
                    for (k = 0; k < cpTemp.numLegal(); k++)
                    {
                      if (bRotated)
                      {
                        bHighlightFlag[7 - cpTemp.getLegalMove(k).getRank()][7 - cpTemp.getLegalMove(k).getFile()] = true;
                      }
                      else
                      {
                        bHighlightFlag[cpTemp.getLegalMove(k).getRank()][cpTemp.getLegalMove(k).getFile()] = true;
                      }
                    }
                  }
                }
                break;
              }
            }
            break;
          }
        }
      }
      repaint();
    }
    return true;

  } // End of mouseMove().

  private void computeMoveAlphaBeta(GameNode gn, int iDepthCount, float sfAlpha, float sfBeta)
  {
    int i = 0;                             // Depth-first recursion to a set depth.
    int j = 0;
    int k = 0;
    float sfVal = 0;                       // For alpha-beta pruning.
    float sfMetric = 0;
    int iNumAvailableMoves = 0;            // The number of moves available to the computer (black).
    ChessPiece cpTemp0 = null;
    ChessBoard cbTemp0 = null;
    ChessPiece cpTemp1 = null;
    ChessBoard cbTemp1 = null;
    GameNode gnTemp = null;
    int iNumLegal = 0;
    int iFromRank = 0;
    int iFromFile = 0;
    Square sMoveTo = null;
    ChessBoard cbNext = null;              // To show move being considered when depth count = 2.
    Square sMove = null;                   //
    String strMove = null;                 //
    boolean bCapture = false;
    boolean bWhiteMove = false;
    boolean bCheck = false;
    int iQuiescentBonus = 0;

    iDepthCount++;                         // Increment the depth counter.
    lPositionsCounter++;                   // Increment the number of positions considered counter.
    cbTemp0 = gn.getBoard();
    bWhiteMove = cbTemp0.moveP();          // The person to move: white's move if true.
    bCapture = cbTemp0.captureP();         // Has a capture been made on the chessboard?
    bCheck = cbTemp0.checkP();             // Is somebody in check?
    sfMetric = gn.getMetric();

    iQuiescentBonus = iRecursionLimit / 2;

    if ((iRecursionLimit + iQuiescentBonus) % 2 != 0)
    {
      iQuiescentBonus -= 1;                // Changed from -+ to += in version 3.93.
    }

    if (bThinking &&                                              // Kill the thought processes if applet has restarted.
        ((iDepthCount <= iRecursionLimit) ||                      // n half-moves look-ahead or
         (bCapture && iDepthCount <= iRecursionLimit + iQuiescentBonus)) &&
        ((sfMetric < 100 && sfMetric > -100) || ((sfMetric > 100 || sfMetric < -100) && bCheck)))
    {
      // Not at max depth, keep recursion going...
      // Construct my children:      
      // For each legal move for each piece on the board, as appropriate, spawn a game node:
      for (i = 0; i < 8; i++)                                     // Ranks.
      {
        for (j = 0; j < 8; j++)                                   // Files.
        {
          cpTemp0 = cbTemp0.getPiece(i, j);
          if (cpTemp0 != null)
          {
            if ((!bWhiteMove && !cpTemp0.whiteP()) || (bWhiteMove && cpTemp0.whiteP()))
            {
              // It's the computer's move and the piece is black or it's the human's move and the piece is white:
              iNumLegal = cpTemp0.numLegal();
              for (k = 0; k < iNumLegal; k++)                     // For each move of the piece.
              {
                sMoveTo = cpTemp0.getLegalMove(k);                // Get the move square.

                if (cpTemp0.getType() == 'P' && sMoveTo.pawnPromoteP())    // Pawn promotion move:
                {
                  // Make four alternative moves: each for a type of promotion, N, B, R, or Q:

                  // Copy the parent ChessBoard and piece:
                  cbTemp1 = new ChessBoard(cbTemp0);
                  cbTemp1.setMove(!bWhiteMove);
                  cbTemp1.setCapture(false);
                  cbTemp1.setCheck(false);
                  cpTemp1 = new ChessPiece(cpTemp0.whiteP(), 'N', i, j);
                  cbTemp1.setPiece(cpTemp1, sMoveTo);
                  cbTemp1.refreshLegal();
                  gnTemp = new GameNode();
                  gnTemp.setBoard(cbTemp1);
                  gnTemp.setMetric(cbTemp1.evaluate());
                  gn.addChild(gnTemp);
                  
                  cbTemp1 = new ChessBoard(cbTemp0);
                  cbTemp1.setMove(!bWhiteMove);
                  cbTemp1.setCapture(false);
                  cbTemp1.setCheck(false);
                  cpTemp1 = new ChessPiece(cpTemp0.whiteP(), 'B', i, j);
                  cbTemp1.setPiece(cpTemp1, sMoveTo);
                  cbTemp1.refreshLegal();
                  gnTemp = new GameNode();
                  gnTemp.setBoard(cbTemp1);
                  gnTemp.setMetric(cbTemp1.evaluate());
                  gn.addChild(gnTemp);
                  
                  cbTemp1 = new ChessBoard(cbTemp0);
                  cbTemp1.setMove(!bWhiteMove);
                  cbTemp1.setCapture(false);
                  cbTemp1.setCheck(false);
                  cpTemp1 = new ChessPiece(cpTemp0.whiteP(), 'R', i, j);
                  cbTemp1.setPiece(cpTemp1, sMoveTo);
                  cbTemp1.refreshLegal();
                  gnTemp = new GameNode();
                  gnTemp.setBoard(cbTemp1);
                  gnTemp.setMetric(cbTemp1.evaluate());
                  gn.addChild(gnTemp);
                  
                  cbTemp1 = new ChessBoard(cbTemp0);
                  cbTemp1.setMove(!bWhiteMove);
                  cbTemp1.setCapture(false);
                  cbTemp1.setCheck(false);
                  cpTemp1 = new ChessPiece(cpTemp0.whiteP(), 'Q', i, j);
                  cbTemp1.setPiece(cpTemp1, sMoveTo);
                  cbTemp1.refreshLegal();
                  gnTemp = new GameNode();
                  gnTemp.setBoard(cbTemp1);
                  gnTemp.setMetric(cbTemp1.evaluate());
                  gn.addChild(gnTemp);
                }
                else                                                        // Ordinary move:
                {
                  // Copy the parent ChessBoard and set its attributes:
                  cbTemp1 = new ChessBoard(cbTemp0);
                  cbTemp1.setMove(!bWhiteMove);
                  cbTemp1.setCapture(false);
                  cbTemp1.setCheck(false);
                  cpTemp1 = new ChessPiece(cpTemp0);
                  cbTemp1.setPiece(cpTemp1, sMoveTo);                       // Make the move on the ChessBoard copy.
                  cbTemp1.refreshLegal();                                   // Refresh the legal moves.
                  gnTemp = new GameNode();                                  // Generate a new GameNode.
                  gnTemp.setBoard(cbTemp1);                                 // Attach the ChessBoard to the GameNode.
                  gnTemp.setMetric(cbTemp1.evaluate());                     // Evaluate the board.
                  gn.addChild(gnTemp);                                      // Add the new GameNode to the child vector.
                }
              } // End for k.
            }
          }
        }  // End for j.
      } // End for i.

      iNumAvailableMoves = gn.numChildren();

      if (iNumAvailableMoves == 0)                                // Always have at least one move
      {                                                           // so we can see a checkmate (king missing).
        // Make a dummy static move:
        cbTemp1 = new ChessBoard(cbTemp0);
        cbTemp1.setMove(!bWhiteMove);                             // Change the player to move, but don't move.
        cbTemp1.setCapture(false);
        cbTemp1.setCheck(false);
        cbTemp1.refreshLegal();                                   // Refresh the legal moves.
        gnTemp = new GameNode();                                  // Generate a new GameNode.
        gnTemp.setBoard(cbTemp1);                                 // Attach the ChessBoard to the GameNode.
        gnTemp.setMetric(cbTemp1.evaluate());                     // Evaluate the board.
        gn.addChild(gnTemp);                                      // Add the new GameNode to the child vector.
        iNumAvailableMoves = 1;
      }

      // Sort the children (capture moves get considered first):
      gn.sortChildrenQuick(!bWhiteMove);                          // True for ascending sort.

      if (iDepthCount == 1)
      {
        iNumImmediateChoices = iNumAvailableMoves;
        if (!bFixedDepth)
        {
          if (bVerbose)
          {
            System.out.println("Looking ahead " + iRecursionLimit + " tempi, quiescent bonus = " + iQuiescentBonus + ".");
          }
          else
          {
            System.out.println("Looking ahead " + iRecursionLimit + " tempi.");
          }
        }
      }
      else
      {
        if (iDepthCount == 2)
        {
          if  (iRecursionLimit > 2)
          {
            strMove = "";
            cbNext = gn.getBoard();
            if (cbNext != null)
            {
              sMove = cbNext.getMove();
              if (sMove != null)
              {
                strMove = algebraicNotation(gnGameRoot.getBoard(), cbNext, sMove, !bWhiteMove);
              }
            }
            iMoveCounter++;
            if (strBest == null) strBest = strMove;
            if (strBestMetric == null) strBestMetric = "";
            lblMessage.setText("Considering " + strMove + " (" + iMoveCounter + ") of " + iNumImmediateChoices +
                               " moves, " + iNumAvailableMoves + " replies, " +
                               lPositionsCounter + " positions: " + strBest + " " + strBestMetric);
          }
          else
          {
            lblMessage.setText("Thinking in speed chess mode.");
          }
        }
      }
      
      // Spawn by recursion:
      // Maximize on the computer's move, minimize on the human's move.
      if (bWhiteMove)                                                                   // Max node, white's move.
      {
        gn.setMetric(sfAlpha);
        for (i = 0; i < iNumAvailableMoves; i++)
        {
          computeMoveAlphaBeta(gn.getChild(i), iDepthCount, sfAlpha, sfBeta);           // Recursion.
          
          // It's white's move, search the children for the worst outcome (positive).
          sfVal = gn.getChild(i).getMetric();

          if (sfVal > sfAlpha)
          {
            sfAlpha = sfVal;
            gn.setBest(gn.getChild(i).getBoard());
            gn.setMetric(sfAlpha);
            if (iDepthCount == 1)
            {
              cbNext = gn.getBest();
              if (cbNext != null)
              {
                sMove = cbNext.getMove();
                if (sMove != null)
                {
                  strBest = algebraicNotation(gnGameRoot.getBoard(), cbNext, sMove, bWhiteMove);
                  if (sfVal < 0.001 && sfVal > -0.001)
                  {
                    strBestMetric = "(" + Float.toString(0) + ")";
                  }
                  else
                  {
                    try
                    {
                      strBestMetric = "(" + (Float.toString(sfVal)).substring(0, 7) + ")";
                    }
                    catch (StringIndexOutOfBoundsException e)
                    {
                      strBestMetric = "(" + Float.toString(sfVal) + ")";
                    }
                  }
                }
              }
            }
          }
          if (sfAlpha >= sfBeta) break;
        }
      }
      else                                                                               // Min node, black's move.
      {
        gn.setMetric(sfBeta);
        for (i = 0; i < iNumAvailableMoves; i++)
        {
          computeMoveAlphaBeta(gn.getChild(i), iDepthCount, sfAlpha, sfBeta);            // Recursion.

          // It's the black's move, search the children for the best outcome (negative).
          sfVal = gn.getChild(i).getMetric();
          
          if (sfVal < sfBeta)
          {
            sfBeta = sfVal;
            gn.setBest(gn.getChild(i).getBoard());
            gn.setMetric(sfBeta);
            if (iDepthCount == 1)
            {
              cbNext = gn.getBest();
              if (cbNext != null)
              {
                sMove = cbNext.getMove();
                if (sMove != null)
                {
                  strBest = algebraicNotation(gnGameRoot.getBoard(), cbNext, sMove, bWhiteMove);
                  if (sfVal < 0.001 && sfVal > -0.001)
                  {
                    strBestMetric = "(" + Float.toString(0) + ")";
                  }
                  else
                  {
                    try
                    {
                      strBestMetric = "(" + (Float.toString(sfVal)).substring(0, 7) + ")";
                    }
                    catch (StringIndexOutOfBoundsException e)
                    {
                      strBestMetric = "(" + Float.toString(sfVal) + ")";
                    }
                  }
                }
              }
            }
          }
          if (sfBeta <= sfAlpha) break;
        }
      }

      // Recursion is complete and we've gathered the information: destroy the children.
      for (i = 0; i < iNumAvailableMoves; i++)
      {
        gn.setChild(null, i);                                                             // This activates the JRE garbage collector.
      }
    }
  } // End of applet function computeMoveAlphaBeta().

  // Implements double buffering
  public void update(Graphics g)
  {
    if (imOffScreen == null)
    {
      // Make sure the offscreen and graphics exist:
      imOffScreen = this.createImage(dApplet.width, dApplet.height);
      grOffScreen = imOffScreen.getGraphics();
      grOffScreen.clearRect(0, 0, dApplet.width, dApplet.height);
    }
    this.paint(grOffScreen);
    g.drawImage(imOffScreen, 0, 0, null);

  } // End of update().

  // The applet frame painting function
  public void paint(Graphics g)
  {
    // Code for displaying images or drawing in the applet frame
    g.clearRect(0, 0, dApplet.width, dApplet.height);        // Needed for double buffering.
    this.setBackground(Color.lightGray);                     // Ditto.

    drawBoard(g);                                            // Draw the board.
    drawPieces(g);                                           // Draw the pieces.
    drawFrame(g);                                            // Draw the frame abound the applet.

  } // End of paint().

  private void drawBoard(Graphics g)
  {
    int i = 0;
    int j = 0;
    String strRankFile = null;
    char cRank = 0;

    // Draw the chess board:
    for (i = 0; i < 8; i++)                                  // Ranks.
    {
      // File letters:
      g.setColor(Color.black);
      if (bRotated)
      {
        cRank = (char) (97 + 7 - i);
      }
      else
      {
        cRank = (char) (97 + i);
      }
      strRankFile = (new Character(cRank)).toString();
      g.drawString(strRankFile,
                   iSquareWidth * (i + 1) + iSquareWidth / 2 - fmApplet.stringWidth(strRankFile) / 2,
                   iSquareHeight * 9 + iSquareHeight / 3 + fmApplet.getAscent() / 2);

      // Rank digits:
      if (bRotated)
      {
        strRankFile = Integer.toString(i + 1);
      }
      else
      {
        strRankFile = Integer.toString(8 - i);
      }
      g.drawString(strRankFile,
                   iSquareWidth - iSquareWidth / 3,
                   iSquareHeight * (i + 1) + iSquareHeight / 2 + fmApplet.getAscent() / 2);
      g.setColor(Color.white);

      for (j = 0; j < 8; j++)                                // Files.
      {
        if ((i % 2 == 0 && j %2 == 0) ||                     // Even row and even column
            (i % 2 != 0 && j %2 != 0))                       // or odd row and odd column.
        {
          g.fillRect(iSquareWidth * (j + 1), iSquareHeight * (i + 1), iSquareWidth, iSquareHeight);
        }

        // Highlight the legal moves of the piece the mouse is over:
        if (bHighlightFlag[i][j])
        {
          bHighlightFlag[i][j] = false;
          if ((i % 2 == 0 && j %2 == 0) ||                     // Even row and even column
              (i % 2 != 0 && j %2 != 0))                       // or odd row and odd column.
          {
            g.setColor(Color.yellow.brighter());
          }
          else
          {
            g.setColor(Color.yellow.darker());
          }
          g.fillRect(iSquareWidth * (j + 1), iSquareHeight * (i + 1), iSquareWidth, iSquareHeight);
          g.setColor(Color.white);
        }
      }
    }

    // Highlight the from square:
    if (sHumanFrom != null)
    {
      g.setColor(Color.green);
      if (bRotated)
      {
        g.fillRect(iSquareWidth * (7 - sHumanFrom.getFile() + 1), iSquareHeight * (7 - sHumanFrom.getRank() + 1),
                   iSquareWidth, iSquareHeight);
      }
      else
      {
        g.fillRect(iSquareWidth * (sHumanFrom.getFile() + 1), iSquareHeight * (sHumanFrom.getRank() + 1),
                   iSquareWidth, iSquareHeight);
      }
    }

    // Draw a border around the squares:
    g.setColor(Color.gray);
    g.drawLine(iSquareWidth, iSquareHeight, iSquareWidth * 9, iSquareHeight);
    g.drawLine(iSquareWidth, iSquareHeight * 9, iSquareWidth * 9, iSquareHeight * 9);
    g.drawLine(iSquareWidth, iSquareHeight, iSquareWidth, iSquareHeight * 9);
    g.drawLine(iSquareWidth * 9, iSquareHeight, iSquareWidth * 9, iSquareHeight * 9);

  } // End of drawBoard().

  private void drawPieces(Graphics g)                      // Draw all the chess pieces.
  {
    int i = 0;
    int j = 0;
    ChessPiece cbTemp = null;
    ChessBoard cbCurrent = gnGameRoot.getBoard();

    if (bRotated)
    {
      for (i = 0; i < 8; i++)        // Ranks
      {
        for (j = 0; j < 8; j++)        // Files
        {
          cbTemp = cbCurrent.getPiece(i, j);
          if (cbTemp != null)
          {
            switch (cbTemp.getType())
            {
              case 'P': drawPawn(iSquareWidth * (7 - j + 1), iSquareHeight * (7 - i + 1), cbTemp.whiteP(), g); break;
              case 'B': drawBishop(iSquareWidth * (7 - j + 1), iSquareHeight * (7 - i + 1), cbTemp.whiteP(), g); break;
              case 'N':
              {
                if (7 - j < 4)
                {
                  drawQKnight(iSquareWidth * (7 - j + 1), iSquareHeight * (7 - i + 1), cbTemp.whiteP(), g);
                }
                else
                {
                  drawKKnight(iSquareWidth * (7 - j + 1), iSquareHeight * (7 - i + 1), cbTemp.whiteP(), g);
                }
                break;
              }
              case 'R': drawRook(iSquareWidth * (7 - j + 1), iSquareHeight * (7 - i + 1), cbTemp.whiteP(), g); break;
              case 'Q': drawQueen(iSquareWidth * (7 - j + 1), iSquareHeight * (7 - i + 1), cbTemp.whiteP(), g); break;
              case 'K': drawKing(iSquareWidth * (7 - j + 1), iSquareHeight * (7 - i + 1), cbTemp.whiteP(), g); break;
            }
          }
        }
      }
    }
    else
    {
      for (i = 0; i < 8; i++)        // Ranks
      {
        for (j = 0; j < 8; j++)        // Files
        {
          cbTemp = cbCurrent.getPiece(i, j);
          if (cbTemp != null)
          {
            switch (cbTemp.getType())
            {
              case 'P': drawPawn(iSquareWidth * (j + 1), iSquareHeight * (i + 1), cbTemp.whiteP(), g); break;
              case 'B': drawBishop(iSquareWidth * (j + 1), iSquareHeight * (i + 1),cbTemp.whiteP(), g); break;
              case 'N':
              {
                if (j < 4)
                {
                  drawQKnight(iSquareWidth * (j + 1), iSquareHeight * (i + 1), cbTemp.whiteP(), g);
                }
                else
                {
                  drawKKnight(iSquareWidth * (j + 1), iSquareHeight * (i + 1), cbTemp.whiteP(), g);
                }
                break;
              }
              case 'R': drawRook(iSquareWidth * (j + 1), iSquareHeight * (i + 1), cbTemp.whiteP(), g); break;
              case 'Q': drawQueen(iSquareWidth * (j + 1), iSquareHeight * (i + 1), cbTemp.whiteP(), g); break;
              case 'K': drawKing(iSquareWidth * (j + 1), iSquareHeight * (i + 1), cbTemp.whiteP(), g); break;
            }
          }
        }
      }
    }
  } // End of drawPieces().

  private void drawPiece(Polygon pg, boolean bW, Graphics g)
  {
    if (bW)
    {
      g.setColor(cWhitePiece);
      g.fillPolygon(pg);
      g.setColor(Color.black);
      g.drawPolygon(pg);
    }
    else
    {
      g.setColor(cBlackPiece);
      g.fillPolygon(pg);
    }
  } // End of drawPiece().

  private void drawPawn(int iStartX, int iStartY, boolean bW, Graphics g)
  {
    int i = 0;
    Polygon pgTemp = new Polygon(pgPawn.xpoints, pgPawn.ypoints, pgPawn.npoints);

    for (i = 0; i < pgTemp.npoints; i++)
    {
      pgTemp.xpoints[i] += iStartX;
      pgTemp.ypoints[i] += iStartY;
    }
    drawPiece(pgTemp, bW, g);

    if (bW)
    {
      g.setColor(Color.black);
    }
    else
    {
      g.setColor(Color.white);
    }
    g.drawLine(pgTemp.xpoints[2], pgTemp.ypoints[2], pgTemp.xpoints[8], pgTemp.ypoints[8]);

  } // End of drawPawn().

  private void drawKKnight(int iStartX, int iStartY, boolean bW, Graphics g)
  {
    int i = 0;
    int x1 = 0;
    int x2 = 0;

    Polygon pgTemp = new Polygon(pgKKnight.xpoints, pgKKnight.ypoints, pgKKnight.npoints);

    for (i = 0; i < pgTemp.npoints; i++)
    {
      pgTemp.xpoints[i] += iStartX;
      pgTemp.ypoints[i] += iStartY;
    }
    drawPiece(pgTemp, bW, g);

    if (bW)
    {
      g.setColor(Color.black);
    }
    else
    {
      g.setColor(Color.white);
    }
    g.drawLine(pgTemp.xpoints[2], pgTemp.ypoints[2], pgTemp.xpoints[13], pgTemp.ypoints[13]);
    x1 = iStartX + (pgTemp.xpoints[4] - pgTemp.xpoints[9]) * 2 / 3;
    x2 = x1 - (pgTemp.xpoints[4] - pgTemp.xpoints[9]) / 12;

    g.drawLine(x1, pgTemp.ypoints[4], x2, pgTemp.ypoints[4]);

  } // End of drawKKnight().

  private void drawQKnight(int iStartX, int iStartY, boolean bW, Graphics g)
  {
    int i = 0;
    int x1 = 0;
    int x2 = 0;

    Polygon pgTemp = new Polygon(pgQKnight.xpoints, pgQKnight.ypoints, pgQKnight.npoints);

    for (i = 0; i < pgTemp.npoints; i++)
    {
      pgTemp.xpoints[i] += iStartX;
      pgTemp.ypoints[i] += iStartY;
    }
    drawPiece(pgTemp, bW, g);

    if (bW)
    {
      g.setColor(Color.black);
    }
    else
    {
      g.setColor(Color.white);
    }
    g.drawLine(pgTemp.xpoints[2], pgTemp.ypoints[2], pgTemp.xpoints[13], pgTemp.ypoints[13]);
    x1 = iStartX + (pgTemp.xpoints[6] - pgTemp.xpoints[11]) * 7 / 8;
    x2 = x1 + (pgTemp.xpoints[6] - pgTemp.xpoints[11]) / 12;

    g.drawLine(x1, pgTemp.ypoints[11], x2, pgTemp.ypoints[11]);

  } // End of drawQKnight().

  private void drawBishop(int iStartX, int iStartY, boolean bW, Graphics g)
  {
    int i = 0;
    Polygon pgTemp = new Polygon(pgBishop.xpoints, pgBishop.ypoints, pgBishop.npoints);

    for (i = 0; i < pgTemp.npoints; i++)
    {
      pgTemp.xpoints[i] += iStartX;
      pgTemp.ypoints[i] += iStartY;
    }
    drawPiece(pgTemp, bW, g);

    if (bW)
    {
      g.setColor(Color.black);
    }
    else
    {
      g.setColor(Color.white);
    }
    g.drawLine(pgTemp.xpoints[2], pgTemp.ypoints[2],
               pgTemp.xpoints[pgTemp.npoints - 1], pgTemp.ypoints[pgTemp.npoints - 1]);
    g.drawLine(pgTemp.xpoints[4], pgTemp.ypoints[4],
               pgTemp.xpoints[8], pgTemp.ypoints[8]);

  } // End of drawBishop().

  private void drawRook(int iStartX, int iStartY, boolean bW, Graphics g)
  {
    int i = 0;
    Polygon pgTemp = new Polygon(pgRook.xpoints, pgRook.ypoints, pgRook.npoints);

    for (i = 0; i < pgTemp.npoints; i++)
    {
      pgTemp.xpoints[i] += iStartX;
      pgTemp.ypoints[i] += iStartY;
    }
    drawPiece(pgTemp, bW, g);

    if (bW)
    {
      g.setColor(Color.black);
    }
    else
    {
      g.setColor(Color.white);
    }
    g.drawLine(pgTemp.xpoints[2], pgTemp.ypoints[2],
               pgTemp.xpoints[pgTemp.npoints - 1], pgTemp.ypoints[pgTemp.npoints - 1]);
    g.drawLine(pgTemp.xpoints[4], pgTemp.ypoints[4],
               pgTemp.xpoints[17], pgTemp.ypoints[17]);

  } // End of drawRook().

  private void drawQueen(int iStartX, int iStartY, boolean bW, Graphics g)
  {
    int i = 0;
    Polygon pgTemp = new Polygon(pgQueen.xpoints, pgQueen.ypoints, pgQueen.npoints);

    for (i = 0; i < pgTemp.npoints; i++)
    {
      pgTemp.xpoints[i] += iStartX;
      pgTemp.ypoints[i] += iStartY;
    }
    drawPiece(pgTemp, bW, g);

    if (bW)
    {
      g.setColor(Color.black);
    }
    else
    {
      g.setColor(Color.white);
    }
    g.drawLine(pgTemp.xpoints[2], pgTemp.ypoints[2],
               pgTemp.xpoints[pgTemp.npoints - 1], pgTemp.ypoints[pgTemp.npoints - 1]);
    g.drawLine(pgTemp.xpoints[4], pgTemp.ypoints[4],
               pgTemp.xpoints[12], pgTemp.ypoints[12]);

  } // End of drawQueen().

  private void drawKing(int iStartX, int iStartY, boolean bW, Graphics g)
  {
    int i = 0;
    Polygon pgTemp = new Polygon(pgKing.xpoints, pgKing.ypoints, pgKing.npoints);

    for (i = 0; i < pgTemp.npoints; i++)
    {
      pgTemp.xpoints[i] += iStartX;
      pgTemp.ypoints[i] += iStartY;
    }
    drawPiece(pgTemp, bW, g);

    if (bW)
    {
      g.setColor(Color.black);
    }
    else
    {
      g.setColor(Color.white);
    }
    g.drawLine(pgTemp.xpoints[2], pgTemp.ypoints[2],
               pgTemp.xpoints[pgTemp.npoints - 1], pgTemp.ypoints[pgTemp.npoints - 1]);
    g.drawLine(pgTemp.xpoints[5], pgTemp.ypoints[5],
               pgTemp.xpoints[16], pgTemp.ypoints[16]);

  } // End of drawKing().

  private void drawFrame(Graphics g)
  {
    // Draw a recessed frame around the applet border. Designed for gray-on-gray browser background.
    g.setColor(Color.black);
    g.drawLine(0, 0, dApplet.width - 1, 0);
    g.drawLine(0, 0, 0, dApplet.height - 1);
    g.setColor(Color.white);
    g.drawLine(0, dApplet.height - 1, dApplet.width - 1, dApplet.height - 1);
    g.drawLine(dApplet.width - 1, 1, dApplet.width - 1, dApplet.height - 1);

  } // End of drawFrame().

  private Square getSquare(int x, int y)                               // Turn mouse coordinates into
  {                                                                    // a board location.
    Square s = null;                                                   // The return object.

    if (x > iSquareWidth && x < iSquareWidth * 9)                      // Inbounds horizontally.
    {
      if (y > iSquareHeight && y < iSquareHeight * 9)
      {
        if (bRotated)
        {
          s = new Square(7 - (y - iSquareHeight) / iSquareHeight, 7 - (x - iSquareWidth) / iSquareWidth);
        }
        else
        {
          s = new Square((y - iSquareHeight) / iSquareHeight, (x - iSquareWidth) / iSquareWidth);
        }
      }
    }
    return s;

  } // End of applet function getSquare().

  public void onTimer(long lStartTime)
  {
    ChessBoard cbBest = null;
    ChessBoard cbCurrent = null;
    RememberedPosition rpTemp = null;

    iRecurrences++;

    if (iRecurrences > 50)
    {
      if (iRecurrences == 51)
      {
        if (bVerbose)
        {
          System.out.println("Timed out waiting for chess memory server response.");
        }
        bMoveSendInhibited = true;
      }
      // Need to begin the computation...
      cbCurrent = gnGameRoot.getBoard();
      localMove(cbCurrent, lStartTime);
      bWaitingForServer = false;
      tMoveGetter.stop();
      tMoveGetter = null;
    }
    else
    {
      if (sMoveFromServer != null)
      {
        if (sMoveFromServer.startsWith("not found"))
        {
          // "Server does not have the move." printed out from cmc.
          iRecurrences = 51;
        }
        else
        {
          // We have a remembered position from the server:
          rpTemp = new RememberedPosition(sMoveFromServer);

          if (rpTemp.getEvalDepth() >= iRecursionLimit)
          {
            cbCurrent = new ChessBoard(rpTemp.getBoard());
            cbCurrent.refreshLegal();

            // *****
            // Need to put EP capture information, if applicable, into cbBest, to fix the EP capture bug.
            cbBest = new ChessBoard(rpTemp.getBest());  // Modify the ChessBoard string constructor to reflect "was"
            // information so that we can see if an EP capture is now allowed.
            // Solution: create a new two string ChessBoard constructor that adds the EPCapture property to the pawn square.


            cbBest.refreshLegal();
            if (cbBest.captureP(cbCurrent))
            {
              cbBest.setCapture(true);
            }
            sfDeepMetric = cbBest.evaluate();
            iNumImmediateChoices = 0;
            computerMoveAfterActions(cbCurrent, cbBest, lStartTime, true);
            if (bVerbose)
            {
              if (rpTemp.getEvalDepth() > iRecursionLimit)
              {
                System.out.println("Using move remembered from server memory at a greater depth.");
              }
              else
              {
                System.out.println("Using move remembered from server memory at the same depth.");
              }
            }
            bWaitingForServer = false;
            tMoveGetter.stop();
            tMoveGetter = null;
          }
          else
          {
            if (bVerbose) System.out.println("Move remembered from server memory is at a shallower depth.");
            iRecurrences = 51;
          }
        }
      }
    }
  } // End of function onTimer().

} // End of Applet class Chess.

final class ChessBoard                                          // A ChessBoard is a 2D array of (null and
{                                                               // non-null) ChessPieces.
  private ChessPiece p[][];
  private boolean bCapture = false;                             // Has a capture been made? For heuristic
  private Square sMove = null;                                  // search in the minimax algorithm.
  private boolean bMove = false;                                // True if white to move. For draw testing.
  private boolean bCheck = false;                               // Is a king in check?

  ChessBoard()                                                  // Default constructor.
  {
    int i = 0;                                                  // Ranks.
    int j = 0;                                                  // Files.

    p = new ChessPiece[8][8];
    bCapture = false;
    sMove = null;
    bMove = false;
    bCheck = false;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        p[i][j] = null;                                         // Clear the board.
      }
    }
  }
  
  ChessBoard(ChessBoard cb)                                     // Copy constructor.
  {
    int i = 0;                                                  // Ranks.
    int j = 0;                                                  // Files.

    if (cb == null) cb = new ChessBoard();

    p = new ChessPiece[8][8];
    bCapture = cb.bCapture;
    sMove = new Square(cb.sMove);
    bMove = cb.bMove;
    bCheck = cb.bCheck;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (cb.p[i][j] != null)
        {
          p[i][j] = new ChessPiece(cb.p[i][j]);                 // Copy and set the piece on the board.
        }
        else
        {
          p[i][j] = null;                                       // Set the null piece on the board.
        }
      }
    }
  }

  ChessBoard(String s)                                          // String constructor.
  {
    int i = 0;                                                  // Ranks.
    int j = 0;                                                  // Files.
    int iStringIndex = 0;
    StringBuffer sbInput = null;
    char cIn = 0;
    Square sTemp = null;

    sbInput = new StringBuffer(s);
    p = new ChessPiece[8][8];
    bCapture = false;
    sMove = null;
    bMove = true;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        cIn = sbInput.charAt(iStringIndex);
        if (cIn == '_')
        {
          p[i][j] = null;                                          // Set the null piece on the board.
        }
        else
        {
          sTemp = new Square(i, j);
          if (Character.isLowerCase(cIn))
          {
            switch (cIn)
            {
              case 'j':
              {
                p[i][j] = new ChessPiece(false, 'B', sTemp, true);  // Moved is true.
                break;
              }
              case 's':
              {
                p[i][j] = new ChessPiece(false, 'K', sTemp, true);
                break;
              }
              case 'v':
              {
                p[i][j] = new ChessPiece(false, 'N', sTemp, true);
                break;
              }
              case 'x':
              {
                p[i][j] = new ChessPiece(false, 'P', sTemp, true);
                break;
              }
              case 'y':
              {
                p[i][j] = new ChessPiece(false, 'Q', sTemp, true);
                break;
              }
              case 'z':
              {
                p[i][j] = new ChessPiece(false, 'R', sTemp, true);
                break;
              }
              default:
              {
                p[i][j] = new ChessPiece(false, Character.toUpperCase(cIn), sTemp);  // Not moved is default.
              }
            }
          }
          else
          {
            switch (cIn)
            {
              case 'J':
              {
                p[i][j] = new ChessPiece(true, 'B', sTemp, true);  // Moved is true.
                break;
              }
              case 'S':
              {
                p[i][j] = new ChessPiece(true, 'K', sTemp, true);
                break;
              }
              case 'V':
              {
                p[i][j] = new ChessPiece(true, 'N', sTemp, true);
                break;
              }
              case 'X':
              {
                p[i][j] = new ChessPiece(true, 'P', sTemp, true);
                break;
              }
              case 'Y':
              {
                p[i][j] = new ChessPiece(true, 'Q', sTemp, true);
                break;
              }
              case 'Z':
              {
                p[i][j] = new ChessPiece(true, 'R', sTemp, true);
                break;
              }
              default:
              {
                p[i][j] = new ChessPiece(true, cIn, sTemp);  // Not moved is default.
              }
            }
          }
        }
        iStringIndex++;
      }
    }
    cIn = 0;
    try
    {
      cIn = sbInput.charAt(iStringIndex);
    }
    catch (StringIndexOutOfBoundsException e)
    {
    }
    if (cIn != 0)
    {
      if (cIn == '0')
      {
        bMove = false;
      }
    }
  } // End string constructor.

  public boolean captureP()                                     // Predicate naming convention (a Lisp standard).
  {
    return bCapture;                                            // Has a piece just been captured?
  }

  public boolean captureP(ChessBoard cbWas)
  {
    // Given the previous board and this one, has a capture been made?

    boolean bR = false;                                         // The return value.
    int i = 0;
    int j = 0;
    int iNumIs = 0;
    int iNumWas = 0;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (p[i][j] != null)
        {
          iNumIs++;
        }
        if (cbWas.p[i][j] != null)
        {
          iNumWas++;
        }
      }
    }
    if (iNumIs != iNumWas)
    {
      bR = true;
    }
    return bR;
  } // End of captureP().

  public boolean checkP()                                       // Predicate naming convention (a Lisp standard).
  {
    return bCheck;
  }

  public boolean moveP()                                        // True if white has the move.
  {
    return bMove;
  }
  
  public boolean equalsP(ChessBoard cb)                         // Predicate to test for equivalence of chessboards.
  {
    boolean bEquals = true;
    int i = 0;                                                  // Ranks.
    int j = 0;                                                  // Files.

    if (bMove != cb.bMove)                                      // Boards are not equal if a different player has the move.
    {
      bEquals = false;
    }
    else
    {
      for (i = 0; i < 8; i++)
      {
        for (j = 0; j < 8; j++)
        {
          if ((p[i][j] == null && cb.p[i][j] != null) || (p[i][j] != null && cb.p[i][j] == null))
          {
            bEquals = false;
            break;                                              // Break out of j.
          }
          if (p[i][j] != null)
          {
            if (cb.p[i][j] != null)
            {
              if (p[i][j].getType() != cb.p[i][j].getType())
              {
                bEquals = false;
                break;                                          // Break out of j.
              }
            }
          }
        }
        if (!bEquals) break;                                    // Break out of i.
      }
    }
    return bEquals;
  }

  public void setCapture(boolean b)
  {
    bCapture = b;
  }

  public void setMove(boolean b)                                // True if white to move.
  {
    bMove = b;
  }

  public void setCheck(boolean b)
  {
    bCheck = b;
  }

  public Square getMove()
  {
    return sMove;
  }

  public void setPiece(ChessPiece cpInput, Square s)        // Mutator
  {
    int i = s.getRank();
    int j = s.getFile();
    boolean bCastleKingSide = false;
    boolean bCastleQueenSide = false;
    ChessPiece cpTemp = null;                               // Temporary rook for castling.

    sMove = null;

    try
    {
      if (cpInput!= null)
      {
        // See if the (pawn) move will enable the enemy pawn to capture en passant:
        if (cpInput.getType() == 'P')
        {
          if (cpInput.whiteP())
          {
            // Only if a white pawn is standing on rank 6:
            if (cpInput.getSquare().getRank() == 6)
            {
              // Only if the move is two squares ahead (to rank 4):
              if (i == 4)
              {
                // Only if a black pawn is on rank 4 and standing to the left or right of it:
                if (cpInput.getSquare().getFile() > 0)  // Left
                {
                  if ((p[4][cpInput.getSquare().getFile() - 1] != null) &&
                      (p[4][cpInput.getSquare().getFile() - 1].getType() == 'P') &&
                      !(p[4][cpInput.getSquare().getFile() - 1].whiteP()))
                  {
                    p[4][cpInput.getSquare().getFile() - 1].enableEPCapture(cpInput.getSquare().getFile());
                  }
                }
                if (cpInput.getSquare().getFile() < 7)  // Right
                {
                  if ((p[4][cpInput.getSquare().getFile() + 1] != null) &&
                      (p[4][cpInput.getSquare().getFile() + 1].getType() == 'P') &&
                      !(p[4][cpInput.getSquare().getFile() + 1].whiteP()))
                  {
                    p[4][cpInput.getSquare().getFile() + 1].enableEPCapture(cpInput.getSquare().getFile());
                  }
                }
              }
            }
          }
          else
          {
            // Only if a black pawn is standing on rank 1:
            if (cpInput.getSquare().getRank() == 1)
            {
              // Only if the move is two squares ahead (to rank 3):
              if (i == 3)
              {
                // Only if a white pawn is on rank 3 and standing to the left or right of it:
                if (cpInput.getSquare().getFile() > 0)  // Left
                {
                  if ((p[3][cpInput.getSquare().getFile() - 1] != null) &&
                      (p[3][cpInput.getSquare().getFile() - 1].getType() == 'P') &&
                      (p[3][cpInput.getSquare().getFile() - 1].whiteP()))
                  {
                    p[3][cpInput.getSquare().getFile() - 1].enableEPCapture(cpInput.getSquare().getFile());
                  }
                }
                if (cpInput.getSquare().getFile() < 7)  // Right
                {
                  if ((p[3][cpInput.getSquare().getFile() + 1] != null) &&
                      (p[3][cpInput.getSquare().getFile() + 1].getType() == 'P') &&
                      (p[3][cpInput.getSquare().getFile() + 1].whiteP()))
                  {
                    p[3][cpInput.getSquare().getFile() + 1].enableEPCapture(cpInput.getSquare().getFile());
                  }
                }
              }
            }
          }
        } // End if EP capture.

        // Remove the piece from the square it stands on:
        if (cpInput.getSquare() != null)
        {
          p[cpInput.getSquare().getRank()][cpInput.getSquare().getFile()] = null;
        }

        // Castling detection:
        if (cpInput.getType() == 'K')
        {
          if (j - cpInput.getSquare().getFile() == 2) bCastleKingSide = true;
          if (cpInput.getSquare().getFile() - j == 2) bCastleQueenSide = true;
        }

        // Let the piece know its new square:
        cpInput.moveTo(i, j);

        // Let the chessboard store the move:
        sMove = new Square(i, j);
      }

      // Put the piece on the new square:
      p[i][j] = cpInput;

      if (cpInput != null)
      {
        bCapture = s.captureP();
        // If the move is an EP capture, remove the captured piece:
        if (s.EPCP())
        {
          if (cpInput.whiteP())
          {
            p[3][j] = null;
          }
          else
          {
            p[4][j] = null;
          }
        }

        // If the move is castling, put the rook on the other side:
        if (bCastleKingSide)
        {
          cpTemp = p[cpInput.getSquare().getRank()][7];
          p[cpInput.getSquare().getRank()][5] = cpTemp;
          p[cpInput.getSquare().getRank()][7] = null;
          if (cpTemp != null)
          {
            cpTemp.moveTo(cpInput.getSquare().getRank(), 5);
          }
        }
        if (bCastleQueenSide)
        {
          cpTemp = p[cpInput.getSquare().getRank()][0];
          p[cpInput.getSquare().getRank()][3] = cpTemp;
          p[cpInput.getSquare().getRank()][0] = null;
          if (cpTemp != null)
          {
            cpTemp.moveTo(cpInput.getSquare().getRank(), 3);
          }
        }
      }
    }
    catch(ArrayIndexOutOfBoundsException e)
    {
      System.out.println("setPiece(): ArrayIndexOutOfBoundsException i = " + i + " j = " + j);
    }
  } // End of setPiece().

  public void deletePiece(int iRank, int iFile)
  {
    p[iRank][iFile] = null;
  }

  public ChessPiece getPiece(int i, int j)                                   // Accessor
  {
    if (i >= 0 && i <= 7 && j >= 0 & j <= 7)
    {
      return p[i][j];
    }
    else
    {
      return null;
    }
  }

  public ChessPiece getPiece(Square s)                                       // Accessor
  {
    int i = s.getRank();
    int j = s.getFile();
    if (i >= 0 && i <= 7 && j >= 0 & j <= 7)
    {
      return p[i][j];
    }
    else
    {
      return null;
    }
  }

  public void setUpBoard()
  {
    int i = 0;                                                         // Ranks.
    int j = 0;                                                         // Files.
    Square sTemp = null;

    for (j = 0; j < 8; j++)
    {
      sTemp = new Square(1, j);
      setPiece(new ChessPiece(false, 'P', sTemp), sTemp);              // The black pawns.
      sTemp = new Square(6, j);
      setPiece(new ChessPiece(true, 'P', sTemp), sTemp);               // The white pawns.
    }
    
    sTemp = new Square(0, 0);
    setPiece(new ChessPiece(false, 'R', sTemp), sTemp);                // The black rooks.
    sTemp = new Square(0, 7);
    setPiece(new ChessPiece(false, 'R', sTemp), sTemp);
    sTemp = new Square(7, 0);
    setPiece(new ChessPiece(true, 'R', sTemp), sTemp);                 // The white rooks.
    sTemp = new Square(7, 7);
    setPiece(new ChessPiece(true, 'R', sTemp), sTemp);
    
    sTemp = new Square(0, 1);
    setPiece(new ChessPiece(false, 'N', sTemp), sTemp);                // The black knights.
    sTemp = new Square(0, 6);
    setPiece(new ChessPiece(false, 'N', sTemp), sTemp);
    sTemp = new Square(7, 1);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);                 // The white knights.
    sTemp = new Square(7, 6);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);

    sTemp = new Square(0, 2);
    setPiece(new ChessPiece(false, 'B', sTemp), sTemp);                // The black bishops.
    sTemp = new Square(0, 5);
    setPiece(new ChessPiece(false, 'B', sTemp), sTemp);
    sTemp = new Square(7, 2);
    setPiece(new ChessPiece(true, 'B', sTemp), sTemp);                 // The white bishops.
    sTemp = new Square(7, 5);
    setPiece(new ChessPiece(true, 'B', sTemp), sTemp);

    sTemp = new Square(0, 3);
    setPiece(new ChessPiece(false, 'Q', sTemp), sTemp);                // The black queen.
    sTemp = new Square(7, 3);
    setPiece(new ChessPiece(true, 'Q', sTemp), sTemp);                 // The white queen. 
    sTemp = new Square(0, 4);
    setPiece(new ChessPiece(false, 'K', sTemp), sTemp);                // The black king.
    sTemp = new Square(7, 4);
    setPiece(new ChessPiece(true, 'K', sTemp), sTemp);                 // The white king.

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (p[i][j] != null)
        {
          p[i][j].setMoved(false);                                     // For evaluation penalty.
        }
      }
    }
    refreshLegal();                                                    // All pieces.

  } // End of setUpBoard().

  public void TestSetUpBoard()
  {
    int i = 0;                                                         // Ranks.
    int j = 0;                                                         // Files.
    Square sTemp = null;

    //sTemp = new Square(1, 6);
    //setPiece(new ChessPiece(false, 'P', sTemp), sTemp);                // A black pawn.
    //sTemp = new Square(4, 5);
    //setPiece(new ChessPiece(true, 'P', sTemp), sTemp);                 // A white pawn.

    
    //sTemp = new Square(1, 6);
    //setPiece(new ChessPiece(false, 'P', sTemp), sTemp);                // A black pawn.
    //sTemp = new Square(4, 5);
    //setPiece(new ChessPiece(true, 'P', sTemp), sTemp);                 // A white pawn.

    
    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'R', sTemp), sTemp);                // The black rooks.
    //sTemp = new Square(0, 7);
    //setPiece(new ChessPiece(false, 'R', sTemp), sTemp);
    //sTemp = new Square(1, 1);
    //setPiece(new ChessPiece(true, 'R', sTemp), sTemp);                 // The white rooks.
    //sTemp = new Square(5, 6);
    //setPiece(new ChessPiece(true, 'R', sTemp), sTemp);
    
    //sTemp = new Square(0, 1);
    //setPiece(new ChessPiece(false, 'N', sTemp), sTemp);                // The black knights.
    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'N', sTemp), sTemp);
    sTemp = new Square(4, 5);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);                 // The white knights.
    sTemp = new Square(5, 6);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);

    sTemp = new Square(6, 7);
    setPiece(new ChessPiece(false, 'B', sTemp), sTemp);                // The black bishops.
    //sTemp = new Square(0, 5);
    //setPiece(new ChessPiece(false, 'B', sTemp), sTemp);
    //sTemp = new Square(7, 2);
    //setPiece(new ChessPiece(true, 'B', sTemp), sTemp);                 // The white bishops.
    //sTemp = new Square(5, 6);
    //setPiece(new ChessPiece(true, 'B', sTemp), sTemp);

    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'Q', sTemp), sTemp);                // The black queen.
    //sTemp = new Square(7, 3);
    //setPiece(new ChessPiece(true, 'Q', sTemp), sTemp);                 // The white queen. 
    sTemp = new Square(0, 4);
    setPiece(new ChessPiece(false, 'K', sTemp), sTemp);                // The black king.
    sTemp = new Square(3, 4);
    setPiece(new ChessPiece(true, 'K', sTemp), sTemp);                 // The white king.

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (p[i][j] != null)
        {
          p[i][j].setMoved(false);                                     // For evaluation penalty.
        }
      }
    }
    refreshLegal();                                                    // All pieces.

  } // End of TestSetUpBoard().

  public void TestSetUpBoard2()
  {
    int i = 0;                                                         // Ranks.
    int j = 0;                                                         // Files.
    Square sTemp = null;

    //sTemp = new Square(1, 6);
    //setPiece(new ChessPiece(false, 'P', sTemp), sTemp);                // A black pawn.
    //sTemp = new Square(4, 5);
    //setPiece(new ChessPiece(true, 'P', sTemp), sTemp);                 // A white pawn.

    
    //sTemp = new Square(1, 6);
    //setPiece(new ChessPiece(false, 'P', sTemp), sTemp);                // A black pawn.
    //sTemp = new Square(4, 5);
    //setPiece(new ChessPiece(true, 'P', sTemp), sTemp);                 // A white pawn.

    
    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'R', sTemp), sTemp);                // The black rooks.
    //sTemp = new Square(0, 7);
    //setPiece(new ChessPiece(false, 'R', sTemp), sTemp);
    //sTemp = new Square(1, 1);
    //setPiece(new ChessPiece(true, 'R', sTemp), sTemp);                 // The white rooks.
    //sTemp = new Square(5, 6);
    //setPiece(new ChessPiece(true, 'R', sTemp), sTemp);
    
    //sTemp = new Square(0, 1);
    //setPiece(new ChessPiece(false, 'N', sTemp), sTemp);                // The black knights.
    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'N', sTemp), sTemp);
    sTemp = new Square(2, 5);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);                 // The white knights.
    sTemp = new Square(1, 6);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);

    sTemp = new Square(0, 7);
    setPiece(new ChessPiece(false, 'B', sTemp), sTemp);                // The black bishops.
    //sTemp = new Square(0, 5);
    //setPiece(new ChessPiece(false, 'B', sTemp), sTemp);
    //sTemp = new Square(7, 2);
    //setPiece(new ChessPiece(true, 'B', sTemp), sTemp);                 // The white bishops.
    //sTemp = new Square(5, 6);
    //setPiece(new ChessPiece(true, 'B', sTemp), sTemp);

    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'Q', sTemp), sTemp);                // The black queen.
    //sTemp = new Square(7, 3);
    //setPiece(new ChessPiece(true, 'Q', sTemp), sTemp);                 // The white queen. 
    sTemp = new Square(0, 4);
    setPiece(new ChessPiece(false, 'K', sTemp), sTemp);                // The black king.
    sTemp = new Square(3, 4);
    setPiece(new ChessPiece(true, 'K', sTemp), sTemp);                 // The white king.

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (p[i][j] != null)
        {
          p[i][j].setMoved(false);                                     // For evaluation penalty.
        }
      }
    }
    refreshLegal();                                                    // All pieces.

  } // End of TestSetUpBoard2().

  public void TestSetUpBoard3()
  {
    int i = 0;                                                         // Ranks.
    int j = 0;                                                         // Files.
    Square sTemp = null;

    //sTemp = new Square(1, 6);
    //setPiece(new ChessPiece(false, 'P', sTemp), sTemp);                // A black pawn.
    //sTemp = new Square(4, 5);
    //setPiece(new ChessPiece(true, 'P', sTemp), sTemp);                 // A white pawn.

    
    //sTemp = new Square(1, 6);
    //setPiece(new ChessPiece(false, 'P', sTemp), sTemp);                // A black pawn.
    //sTemp = new Square(4, 5);
    //setPiece(new ChessPiece(true, 'P', sTemp), sTemp);                 // A white pawn.

    
    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'R', sTemp), sTemp);                // The black rooks.
    //sTemp = new Square(0, 7);
    //setPiece(new ChessPiece(false, 'R', sTemp), sTemp);
    //sTemp = new Square(1, 1);
    //setPiece(new ChessPiece(true, 'R', sTemp), sTemp);                 // The white rooks.
    //sTemp = new Square(5, 6);
    //setPiece(new ChessPiece(true, 'R', sTemp), sTemp);
    
    //sTemp = new Square(0, 1);
    //setPiece(new ChessPiece(false, 'N', sTemp), sTemp);                // The black knights.
    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'N', sTemp), sTemp);
    sTemp = new Square(2, 3);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);                 // The white knights.
    sTemp = new Square(1, 2);
    setPiece(new ChessPiece(true, 'N', sTemp), sTemp);

    sTemp = new Square(0, 1);
    setPiece(new ChessPiece(false, 'B', sTemp), sTemp);                // The black bishops.
    //sTemp = new Square(0, 5);
    //setPiece(new ChessPiece(false, 'B', sTemp), sTemp);
    //sTemp = new Square(7, 2);
    //setPiece(new ChessPiece(true, 'B', sTemp), sTemp);                 // The white bishops.
    //sTemp = new Square(5, 6);
    //setPiece(new ChessPiece(true, 'B', sTemp), sTemp);

    //sTemp = new Square(0, 6);
    //setPiece(new ChessPiece(false, 'Q', sTemp), sTemp);                // The black queen.
    //sTemp = new Square(7, 3);
    //setPiece(new ChessPiece(true, 'Q', sTemp), sTemp);                 // The white queen. 
    sTemp = new Square(0, 4);
    setPiece(new ChessPiece(false, 'K', sTemp), sTemp);                // The black king.
    sTemp = new Square(3, 4);
    setPiece(new ChessPiece(true, 'K', sTemp), sTemp);                 // The white king.

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (p[i][j] != null)
        {
          p[i][j].setMoved(false);                                     // For evaluation penalty.
        }
      }
    }
    refreshLegal();                                                    // All pieces.

  } // End of TestSetUpBoard3().

  public void refreshLegal()
  {
    // Generate the legal moves (need whole knowledge of board piece positions to do this):
    int i = 0;                                                         // Ranks.
    int j = 0;                                                         // Files.

    Square sWhiteKing = null;
    Square sBlackKing = null;

    Square sKnight[] = null;
    int iNumKnights = 0;

    Square sBishop[] = null;
    int iNumBishops = 0;

    Square sRook[] = null;
    int iNumRooks = 0;

    sKnight = new Square[10];  // It's possible to have 20 knights, 10 for white and 10 for black, if they cooperate.
    sBishop = new Square[10];  // Not likely, but possible.
    sRook = new Square[10];

    boolean bRemoveFlag = false;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (p[i][j] != null)
        {
          // Do the kings last (we need to know the enemy's attacking squares to avoid moving into check):
          if (p[i][j].getType() == 'K')
          {
            if (p[i][j].whiteP())
            {
              sWhiteKing = new Square(i, j);
            }
            else
            {
              sBlackKing = new Square(i, j);
            }
          }
          else
          {
            if (p[i][j].getType() == 'B')
            {
              // Remember where the bishops are for later pin testing.
              sBishop[iNumBishops] = new Square(i, j);
              iNumBishops++;
            }
            if (p[i][j].getType() == 'R')
            {
              // Remember where the rooks are for later pin testing.
              sRook[iNumRooks] = new Square(i, j);
              iNumRooks++;
            }
            if (p[i][j].getType() == 'N')
            {
              // Remember where the knights are for later pin testing.
              sKnight[iNumKnights] = new Square(i, j);
              iNumKnights++;
            }                                                    // Is there some way around passing the chessboard to the
            p[i][j].generateLegalMoves(this);                    // piece? This illustrates the non-triviality of chess 
          }                                                      // modeling, and the potential non-optimality of this model.
        }
      }
    }

    if (sWhiteKing != null && sBlackKing != null)
    {
      // Generate the king moves:
      p[sWhiteKing.getRank()][sWhiteKing.getFile()].generateLegalMoves(this);
      p[sBlackKing.getRank()][sBlackKing.getFile()].generateLegalMoves(this);

      // Then see if the knights, bishops, or rooks are pinned:
      for (i = 0; i < iNumKnights; i++)
      {
        if (pinnedBy(p[sKnight[i].getRank()][sKnight[i].getFile()],
                     p[sKnight[i].getRank()][sKnight[i].getFile()].whiteP()) != 0)
        {
          // Remove the pinned knight's legal moves only if it's not attacking the king of opposite color:
          bRemoveFlag = true;
          if (p[sKnight[i].getRank()][sKnight[i].getFile()].whiteP())
          {
            for (j = 0; j < p[sKnight[i].getRank()][sKnight[i].getFile()].numLegal(); j++)
            {
              if (p[sKnight[i].getRank()][sKnight[i].getFile()].getLegalMove(j).equals(sBlackKing))
              {
                bRemoveFlag = false;
              }
            }
          }
          else
          {
            for (j = 0; j < p[sKnight[i].getRank()][sKnight[i].getFile()].numLegal(); j++)
            {
              if (p[sKnight[i].getRank()][sKnight[i].getFile()].getLegalMove(j).equals(sWhiteKing))
              {
                bRemoveFlag = false;
              }
            }
          }
          if (bRemoveFlag)
          {
            p[sKnight[i].getRank()][sKnight[i].getFile()].removeLegalMoves();
          }
        }
      }
      for (i = 0; i < iNumBishops; i++)
      {
        if (pinnedBy(p[sBishop[i].getRank()][sBishop[i].getFile()],
                     p[sBishop[i].getRank()][sBishop[i].getFile()].whiteP()) == 'R')
        {
          // Remove the pinned bishop's legal moves only if it's not attacking the king of opposite color:
          bRemoveFlag = true;
          if (p[sBishop[i].getRank()][sBishop[i].getFile()].whiteP())
          {
            for (j = 0; j < p[sBishop[i].getRank()][sBishop[i].getFile()].numLegal(); j++)
            {
              if (p[sBishop[i].getRank()][sBishop[i].getFile()].getLegalMove(j).equals(sBlackKing))
              {
                bRemoveFlag = false;
              }
            }
          }
          else
          {
            for (j = 0; j < p[sBishop[i].getRank()][sBishop[i].getFile()].numLegal(); j++)
            {
              if (p[sBishop[i].getRank()][sBishop[i].getFile()].getLegalMove(j).equals(sWhiteKing))
              {
                bRemoveFlag = false;
              }
            }
          }
          if (bRemoveFlag)
          {
            p[sBishop[i].getRank()][sBishop[i].getFile()].removeLegalMoves();
          }
        }
      }
      for (i = 0; i < iNumRooks; i++)
      {
        if (pinnedBy(p[sRook[i].getRank()][sRook[i].getFile()],
                     p[sRook[i].getRank()][sRook[i].getFile()].whiteP()) == 'B')
        {
          // Remove the pinned rook's legal moves only if it's not attacking the king of opposite color:
          bRemoveFlag = true;
          if (p[sRook[i].getRank()][sRook[i].getFile()].whiteP())
          {
            for (j = 0; j < p[sRook[i].getRank()][sRook[i].getFile()].numLegal(); j++)
            {
              if (p[sRook[i].getRank()][sRook[i].getFile()].getLegalMove(j).equals(sBlackKing))
              {
                bRemoveFlag = false;
              }
            }
          }
          else
          {
            for (j = 0; j < p[sRook[i].getRank()][sRook[i].getFile()].numLegal(); j++)
            {
              if (p[sRook[i].getRank()][sRook[i].getFile()].getLegalMove(j).equals(sWhiteKing))
              {
                bRemoveFlag = false;
              }
            }
          }
          if (bRemoveFlag)
          {
            p[sRook[i].getRank()][sRook[i].getFile()].removeLegalMoves();
          }
        }
      }
    }
    else
    {
      if (sWhiteKing != null)
      {
        p[sWhiteKing.getRank()][sWhiteKing.getFile()].generateLegalMoves(this);
      }
      if (sBlackKing != null)
      {
        p[sBlackKing.getRank()][sBlackKing.getFile()].generateLegalMoves(this);
      }
    }
  } // End of refreshLegal().

  public float evaluate()
  {
    float sfMetric = 0;                                                  // Return value.

    int i = 0;                                                           // Ranks.
    int j = 0;                                                           // Files.
    ChessPiece cpTemp = null;
    int iWhiteLegalMoveCounter = 0;
    int iBlackLegalMoveCounter = 0;
    int iLegalMoveBalance = 0;
    int iWhiteMaxPawnDist = 0;
    int iBlackMaxPawnDist = 0;
    int iPawnDistBalance = 0;
    float sfFreedom = 0;                                                  // Freedom factor.
    float sfPawnDist = 0;
    char cT = 0;
    float sfRand = 0;
    final float sfMoveReward = (float) 0.008;                             // Reward for moving unmoved pieces (opening strategy),
    final float sfNoiseLevel = (float) 0.018;                             // and noise, adjusted by trial and error.
    final float sfFreedomFactor = (float) 800;                            // Was 1000 in v. 1.96, 900 in v. 2.03.
    final float sfPawnDistFactor = (float) 800;                           // Introduced in v. 3.93 (at 1600, 800 in v. 3.99).
    boolean bWhiteKing = false;
    boolean bBlackKing = false;
    sfMetric = 0;                                                         // Initialize the metric.

    for (i = 0; i < 8; i++)                                               // Ranks. Black backrank is zero.
    {
      for (j = 0; j < 8; j++)                                             // Files.
      {
        // Sum up the material (black is negative):
        cpTemp = getPiece(i, j);
        if (cpTemp != null)
        {
          cT = cpTemp.getType();
          if (cpTemp.whiteP())
          {
            sfMetric += cpTemp.getValue();
            iWhiteLegalMoveCounter += cpTemp.numLegal();

            //Test for missing white king:
            if (cT == 'K')
            {
              bWhiteKing = true;
            }
            
            if (cT == 'P')
            {
              if (6 - i > iWhiteMaxPawnDist)
              {
                iWhiteMaxPawnDist = 6 - i;
              }
            }
            else
            {
              if (!cpTemp.movedP())                                         // Penalize not moving new pieces:
              {
                if (cT == 'N')
                {
                  sfMetric -= ((float) 1.2) * sfMoveReward;                 // Encourage knights more.  v. 1.96
                }
                if (cT == 'B')
                {
                  sfMetric -= ((float) 1.1) * sfMoveReward;                 // Encourage bishops more.  v. 1.96
                }
                if (cT == 'Q')
                {
                  sfMetric -= ((float) -0.2) * sfMoveReward;                // Encourage queen less.  v. 2.77
                }                                                           // Adjusted from 0.0 to -0.2 v. 3.03
                if (cT == 'K')
                {
                  sfMetric -= ((float) 0.8) * sfMoveReward;                 // Encourage king less.  v. 2.77
                }                                                           // Adjusted from 0.7 to 0.8 v. 3.03
                if (cT == 'R')
                {
                  sfMetric -= sfMoveReward;                                 // Rewards are relative to rook reward.
                }
              }
            }
          }
          else                                                              // It's a black piece.
          {
            sfMetric -= cpTemp.getValue();
            iBlackLegalMoveCounter += cpTemp.numLegal();

            // Test for missing black king:
            if (cT == 'K')
            {
              bBlackKing = true;
            }

            if (cT == 'P')
            {
              if (i - 1 > iBlackMaxPawnDist)
              {
                iBlackMaxPawnDist = i - 1;
              }
            }
            else
            {
              if (!cpTemp.movedP())                                         // Penalize not moving new pieces:
              {
                if (cT == 'N')
                {
                  sfMetric += ((float) 1.2) * sfMoveReward;                 // Encourage knights more.  v. 1.96
                }
                if (cT == 'B')
                {
                  sfMetric += ((float) 1.1) * sfMoveReward;                 // Encourage bishops more.  v. 1.96
                }
                if (cT == 'Q')
                {
                  sfMetric += ((float) -0.2) * sfMoveReward;                // Encourage queen less.  v. 2.77
                }                                                           // Adjusted from 0.0 to -0.2 v. 3.03
                if (cT == 'K')
                {
                  sfMetric += ((float) 0.8) * sfMoveReward;                 // Encourage kings less.  v. 2.77
                }                                                           // Adjusted from 0.7 to 0.8 v. 3.03
                if (cT == 'R')
                {
                  sfMetric += sfMoveReward;                                 // Rewards are relative to rook reward.
                }
              }
            }
          }
        }
      }
    }

    iLegalMoveBalance = iWhiteLegalMoveCounter - iBlackLegalMoveCounter;
    sfFreedom = ((float) iLegalMoveBalance) / sfFreedomFactor;            // Was 1000 in v. 1.96, 900 in v. 2.03.
    sfMetric += sfFreedom;                                                // 

    iPawnDistBalance = iWhiteMaxPawnDist - iBlackMaxPawnDist;             // Introduced in v. 3.93. This should encourage
    sfPawnDist = ((float) iPawnDistBalance) / sfPawnDistFactor;           // pushing the most advanced pawn in the endgame.
    sfMetric += sfPawnDist;                                               //

    // Avoid stalemate of the white king:
    if (sfMetric < 0 && bWhiteKing && iWhiteLegalMoveCounter == 0)        // Black is winning and white
    {                                                                     // king exists and has no moves:
      sfMetric += 500;                                                    // Stalemate is good for white.
    }

    // Avoid stalemate of the black king:
    if (sfMetric > 0 && bBlackKing && iBlackLegalMoveCounter == 0)        // White is winning and black
    {                                                                     // king exists and has no moves:
      sfMetric -= 500;                                                    // Stalemate is good for black.
    }

    // Avoid countercheck:
    if (!bWhiteKing && !bBlackKing)                                       // Both kings are gone.
    {
      if (bMove)                                                          // White to move.
      {
        sfMetric += 3000;                                                 // Bad for black.
      }
      else
      {
        sfMetric -= 3000;                                                 // Bad for white.
      }
    }

    sfRand = (float) ((Math.random() - 0.5) * sfNoiseLevel);              // Noisify the decision.
    sfMetric += sfRand;

    return sfMetric;

  } // End of evaluate().

  public ChessPiece getKing(boolean bW)
  {
    int i = 0;
    int j = 0;
    ChessPiece cpTemp = null;
    ChessPiece cpKing = null;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        cpTemp  = p[i][j];
        if (cpTemp  != null)
        {
          if (cpTemp.getType() == 'K' && cpTemp.whiteP() == bW)
          {
            cpKing = cpTemp;
            break;                                                       // Break out of j.
          }
        }
      }
      if (cpKing != null)
      {
        break;                                                           // Break out of i.
      }
    }
    return cpKing;

  } // End of getKing().

  public char pinnedBy(ChessPiece cp, boolean bW)
  {
    char cR = 0;                                                    // Character return value.

    boolean iContinue = true;                                       // Continuation flag, default true.
    char cContinue = 0;
    int iRank = 0;
    int iFile = 0;
    int iRank2 = 0;
    int iFile2 = 0;
    ChessPiece cpKing = null;
    ChessPiece cpTemp = null;
    Square sPiece = null;
    Square sKing = null;
    int iKingRank = 0;
    int iKingFile = 0;
    int iPieceRank = 0;
    int iPieceFile = 0;
    int iProgress = 0;

    cpKing = getKing(bW);
    sPiece = cp.getSquare();
    sKing = cpKing.getSquare();
    iKingRank = sKing.getRank();
    iKingFile = sKing.getFile();
    iPieceRank = sPiece.getRank();
    iPieceFile = sPiece.getFile();

    // Is the piece on a rook attack line to the king?
    if (iKingRank == iPieceRank || iKingFile == iPieceFile)
    {
      // Is there nothing between the piece and the king?
      if (iPieceFile < iKingFile || iPieceRank < iKingRank)
      {
        // We can scan upward from the piece to the king for an interposing piece:
        if (iKingRank == iPieceRank)
        {
          for (iFile = iPieceFile + 1; iFile < iKingFile; iFile++)
          {
            if (p[iKingRank][iFile] != null)
            {
              iContinue = false;                                     // There's a piece blocking the pin.
              break;                                                 // Break out of iFile.
            }
          }
        }
        else
        {
          for (iRank = iPieceRank + 1; iRank < iKingRank; iRank++)
          {
            if (p[iRank][iKingFile] != null)
            {
              iContinue = false;                                     // There's a piece blocking the pin.
              break;                                                 // Break out of iRank.
            }
          }
        }
      }
      else
      {
        // We can scan downward from the piece to the king for an interposing piece:
        if (iKingRank == iPieceRank)
        {
          for (iFile = iPieceFile - 1; iFile > iKingFile; iFile--)
          {
            if (p[iKingRank][iFile] != null)
            {
              iContinue = false;                                     // There's a piece blocking the pin.
              break;                                                 // Break out of iFile.
            }
          }
        }
        else
        {
          for (iRank = iPieceRank - 1; iRank > iKingRank; iRank--)
          {
            if (p[iRank][iKingFile] != null)
            {
              iContinue = false;                                     // There's a piece blocking the pin.
              break;                                                 // Break out of iRank.
            }
          }
        }
      }
      if (iContinue)
      {
        // For each enemy rook or queen...
        for (iRank = 0; iRank < 8; iRank++)
        {
          for (iFile = 0; iFile < 8; iFile++)
          {
            cpTemp = p[iRank][iFile];
            if (cpTemp != null && cpTemp.whiteP() != bW && (cpTemp.getType() == 'R' || cpTemp.getType() == 'Q'))
            {
              cContinue = 'R';                                      // The rook mode of attack.
              // Is the rook or queen on an attack line?
              if ((iKingRank == iRank && iKingRank == iPieceRank) ||
                  (iKingFile == iFile && iKingFile == iPieceFile))              {
                // Is the piece between the attacker and the king?
                if ((iPieceFile < iKingFile && iFile < iPieceFile) ||
                    (iPieceRank < iKingRank && iRank < iPieceRank) ||
                    (iPieceFile > iKingFile && iFile > iPieceFile) ||
                    (iPieceRank > iKingRank && iRank > iPieceRank))
                {  
                  // Is there nothing between the piece and the attacker?
                  if (iPieceFile < iFile || iPieceRank < iRank)
                  {
                    // We can scan upward from the piece to the attacker for an interposing piece:
                    if (iRank == iPieceRank)
                    {
                      for (iFile2 = iPieceFile + 1; iFile2 < iFile; iFile2++)
                      {
                        if (p[iRank][iFile2] != null)
                        {
                          cContinue = 0;
                          break;                                                 // Break out of iFile2.
                        }
                      }
                    }
                    else
                    {
                      for (iRank2 = iPieceRank + 1; iRank2 < iRank; iRank2++)
                      {
                        if (p[iRank2][iFile] != null)
                        {
                          cContinue = 0;
                          break;                                                 // Break out of iRank2.
                        }
                      }
                    }
                  }
                  else
                  {
                    // We can scan downward from the piece to the attacker for an interposing piece:
                    if (iRank == iPieceRank)
                    {
                      for (iFile2 = iPieceFile - 1; iFile2 > iFile; iFile2--)
                      {
                        if (p[iRank][iFile2] != null)
                        {
                          cContinue = 0;
                          break;                                                 // Break out of iFile2.
                        }
                      }
                    }
                    else
                    {
                      for (iRank2 = iPieceRank - 1; iRank2 > iRank; iRank2--)
                      {
                        if (p[iRank2][iFile] != null)
                        {
                          cContinue = 0;
                          break;                                                 // Break out of iRank2.
                        }
                      }
                    }
                  }
                  if (cContinue != 0)
                  {
                    // The piece is pinned by the attacker.
                    cR = cContinue;
                  }
                }
              }
            }
          }
        }
      } // End if iContinue.
    } // End if rook attack on line to the king.
    else
    {
      // iContinue is still true.
      // Not on a rook line, is the piece on a bishop attack line to the king?
      if (iKingRank - iPieceRank == iKingFile - iPieceFile || iKingRank - iPieceRank == iPieceFile - iKingFile)
      {
        // Piece is on an upward or downward sloping line in relation to the king.
        // Is there nothing between the piece and the king?
        if (iPieceFile < iKingFile)
        {
          // We can scan rightward from the piece to the king for an interposing piece:
          if (iKingRank > iPieceRank)
          {
            // Piece is on the upward sloping line.
            for (iFile = 1; iFile < iKingFile - iPieceFile; iFile++)
            {
              if (p[iPieceRank + iFile][iPieceFile + iFile] != null)
              {
                iContinue = false;                                     // There's a piece blocking the pin.
                break;                                                 // Break out of iFile.
              }
            }
          }
          else
          {
            // Piece is on the downward sloping line.
            for (iFile = 1; iFile < iKingFile - iPieceFile ; iFile++)
            //for (iFile = 1; iFile < iPieceFile - iKingFile; iFile++)
            {
              if (p[iPieceRank - iFile][iPieceFile + iFile] != null)
              {
                iContinue = false;                                     // There's a piece blocking the pin.
                break;                                                 // Break out of iFile.
              }
            }
          }
        }
        else
        {
          // We can scan leftward from the piece to the king for an interposing piece:
          if (iKingRank < iPieceRank)
          {
            // Piece is on the upward sloping line.
            for (iFile = iPieceFile - iKingFile - 1; iFile > 0; iFile--)
            {
              if (p[iPieceRank - iFile][iPieceFile - iFile] != null)
              {
                iContinue = false;                                     // There's a piece blocking the pin.
                break;                                                 // Break out of iFile.
              }
            }
          }
          else
          {
            // Piece is on the downward sloping line.
            for (iFile = iPieceFile - iKingFile - 1; iFile > 0; iFile--)
            {
              if (p[iPieceRank + iFile][iPieceFile - iFile] != null)
              {
                iContinue = false;                                     // There's a piece blocking the pin.
                break;                                                 // Break out of iFile.
              }
            }
          }
        }
        if (iContinue)
        {
          // There is nothing between the piece and the king:
          // For each enemy bishop or queen...
          for (iRank = 0; iRank < 8; iRank++)
          {
            for (iFile = 0; iFile < 8; iFile++)
            {
              cpTemp = p[iRank][iFile];
              if (cpTemp != null && cpTemp.whiteP() != bW && (cpTemp.getType() == 'B' || cpTemp.getType() == 'Q'))
              {
                cContinue = 'B';
                // It's an enemy bishop or queen:
                // Is the bishop or queen on an attack line?
                //if ((iKingRank - iRank == iKingFile - iFile || iKingRank - iRank == iFile - iKingFile))
                if ((iKingRank - iRank == iKingFile - iFile && iKingRank - iPieceRank == iKingFile - iPieceFile) ||
                    (iKingRank - iRank == iFile - iKingFile && iKingRank - iPieceRank == iPieceFile - iKingFile))
                {
                  // The attacker is on a diagonal line to the king.
                  // Is the piece between the attacker and the king?
                  if ((iPieceFile < iKingFile && iFile < iPieceFile) ||
                      (iPieceFile > iKingFile && iFile > iPieceFile))
                  {  
                    // Is there nothing between the piece and the attacker?
                    if (iPieceFile < iFile) // Attacker is on the right
                    {
                      // We can scan upward from the piece to the attacker for an interposing piece:
                      if (iRank > iPieceRank)
                      {
                        // Attacker is on the upward sloping line to the king:
                        //cContinue = 0;
                        for (iFile2 = 1; iFile2 <  iFile - iPieceFile; iFile2++)
                        {
                          if (p[iPieceRank + iFile2][iPieceFile + iFile2] != null)
                          {
                            cContinue = 0;
                            break;                                                 // Break out of iFile2.
                          }
                        }
                      }
                      else
                      {
                        for (iFile2 = 1; iFile2 < iFile - iPieceFile; iFile2++)
                        {
                          if (p[iPieceRank - iFile2][iPieceFile + iFile2] != null)
                          {
                            cContinue = 0;
                            break;                                                 // Break out of iRank2.
                          }
                        }
                      }
                    }
                    else
                    {
                      // We can scan downward from the piece to the attacker for an interposing piece:
                      if (iPieceRank < iRank)
                      {
                        for (iFile2 = iPieceFile - iFile - 1 ; iFile2 > 0; iFile2--)
                        {
                          if (p[iPieceRank + iFile2][iPieceFile - iFile2] != null)
                          {
                            cContinue = 0;
                            break;                                                 // Break out of iFile2.
                          }
                        }
                      }
                      else
                      {
                        for (iFile2 = iPieceFile - iFile - 1 ; iFile2 > 0; iFile2--)
                        {
                          if (p[iPieceRank - iFile2][iPieceFile - iFile2] != null)
                          {
                            cContinue = 0;
                            break;                                                 // Break out of iRank2.
                          }
                        }
                      }
                    }
                    if (cContinue != 0)
                    {
                      // The piece is pinned by the attacker.
                      cR = cContinue;
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    return cR;

  } // End of pinnedBy().

  public int numChecking(ChessPiece cpKing, ChessPiece[] cpAttacker, boolean bW)
  {
    int i = 0;
    int j = 0;
    int k = 0;
    ChessPiece cpTemp = null;
    int iNumChecking = 0;

    if (cpKing != null)
    {
      for (i = 0; i < 8; i++)
      {
        for (j = 0; j < 8; j++)
        {
          cpTemp = getPiece(i, j);
          if (cpTemp != null && cpTemp.whiteP() != bW)  
          {
            for (k = 0; k < cpTemp.numLegal(); k++)
            {
              if (cpTemp.getLegalMove(k).equals(cpKing.getSquare()))
              {
                cpAttacker[iNumChecking] = new ChessPiece(cpTemp);
                iNumChecking++;
                break;                                                      // Break out of k.
              }                                                             // A piece can check only once.
            }
          }
        }
      }
    }
    return iNumChecking;

  } // End of numChecking().

  // See if there is a checkmate (predicate naming convention, a Lisp standard):
  public boolean checkmateP(boolean bW)                                  // bW is true if the king is white.
  {
    boolean bR = false;                                                  // Return value.
    int i = 0;
    int j = 0;
    int k = 0;
    int l = 0;
    boolean bOK = false;
    boolean bSafe = false;
    ChessPiece cpTemp = null;
    ChessPiece cpKing = null;
    int iNumChecking = 0;
    int iNumLegal = 0;
    ChessPiece cpAttacker[] = new ChessPiece[2];
    cpAttacker[0] = null;
    cpAttacker[1] = null;
    Square sTemp = null;
    int iNumInterpose = 0;

    boolean bQueen1 = false;
    boolean bQueen2 = false;

    ChessBoard cbTestCheck = null;

    // Get the king:
    cpKing = getKing(bW);

    // Is the king in check?
    iNumChecking = numChecking(cpKing, cpAttacker, bW);

    if (iNumChecking > 0)
    {      
      // The king is in check. Does the king have any legal moves?
      iNumLegal = cpKing.numLegal();
      if (iNumLegal > 0)
      {
        // Are any of the legal move squares not under attack?
        for (i = 0; i < iNumLegal; i++)
        {
          // For each legal move of the king, see if it is OK to move there.
          
          // Make a copy of the chessboard:
          cbTestCheck = new ChessBoard(this);
          
          // Make the move on the chessboard:
          cbTestCheck.setPiece(new ChessPiece(cpKing), cpKing.getLegalMove(i));
          
          // Refresh the legal moves:
          cbTestCheck.refreshLegal();
          
          bOK = true;                                                      // Assume the square is not in check.
          for (j = 0; j < 8; j++)
          {
            for (k = 0; k < 8; k++)
            {
              cpTemp = cbTestCheck.getPiece(j, k);
              if (cpTemp != null && cpTemp.whiteP() != bW)
              {
                for (l = 0; l < cpTemp.numLegal(); l++)
                {
                  if (cpKing.getLegalMove(i).equals(cpTemp.getLegalMove(l)))
                  {
                    // A king's move square is attacked.
                    bOK = false;
                    break;                                                  // Break out of piece's move loop.
                  }
                } // End l.
              }
            } // End k.
          } // End j.
          if (bOK)
          {
            // The king has a square to move to, he's safe!
            bSafe = true;
            break;
          }
        } // End i.
      }
      if (!bSafe)
      {
        // The king has no moves. If it's double check it's also checkmate:
        if (iNumChecking > 1)
        {
          bR = true;                                                      // Checkmate.
        }
        else
        {
          // See if we can capture the single attacker:
          bSafe = captureP(this, bW, cpAttacker[0]);
          if (!bSafe)
          {
            // Last resort: we might interpose a piece if it's not a knight or pawn giving check:
            if (cpAttacker[0].getType() == 'P' || cpAttacker[0].getType() == 'N')
            {
              bR = true;                                                  // Checkmate.
            }
            else
            {
              // Can we interpose?
              if (cpAttacker[0].getType() == 'R')
              {
                bSafe = rookInterposeP(this, bW, cpKing, cpAttacker[0]);
              }
              else
              {
                if (cpAttacker[0].getType() == 'B')
                {
                  bSafe = bishopInterposeP(this, bW, cpKing, cpAttacker[0]);
                }
                else
                {
                  if (cpAttacker[0].getType() == 'Q')
                  {
                    bSafe = rookInterposeP(this, bW, cpKing, cpAttacker[0]) || 
                            bishopInterposeP(this, bW, cpKing, cpAttacker[0]);
                  }
                }
              }
            }
          }
          if (!bSafe) bR = true;                                      // Checkmate.
        }
      }
    }
    return bR;

  } // End of checkmateP().

  private boolean captureP(ChessBoard cb, boolean bW, ChessPiece cpAttacker)
  {
    boolean bSafe = false;
    int i = 0;
    int j = 0;
    int k = 0;
    ChessPiece cpTemp = null;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        cpTemp = cb.getPiece(i, j);
        if (cpTemp != null && cpTemp.whiteP() == bW && cpTemp.getType() != 'K')
        {
          // It's one of my guys and it's not the king:
          for (k = 0; k < cpTemp.numLegal(); k++)
          {
            if (cpTemp.getLegalMove(k).equals(cpAttacker.getSquare()))
            {
              // May capture only if the capturing piece is not pinned:
              if (cb.pinnedBy(cpTemp, bW) == 0)
              {
                bSafe = true;
                break;                                                // Break out of k.
              }
            }
          } // End k.
        }
        if (bSafe) break;                                             // Break out of j.
      } // End j.
      if (bSafe) break;                                               // Break out of i.
    } // End i.

    return bSafe;

  } // End of captureP().

  private boolean rookInterposeP(ChessBoard cb, boolean bW, ChessPiece cpKing, ChessPiece cpAttacker)
  {
    boolean bSafe = false;
    int i = 0;
    int j = 0;
    int k = 0;
    int l = 0;
    Square sTemp = null;
    ChessPiece cpTemp = null;

    // First see if the rook is adjacent to the king:
    // Same rank:
    if (cpKing.getSquare().getRank() == cpAttacker.getSquare().getRank() &&
        Math.abs(cpKing.getSquare().getFile() - cpAttacker.getSquare().getFile()) == 1)
    {
      // They're adjacent, no possibility of interposing.
      return bSafe;
    }
    // Same file:
    if (cpKing.getSquare().getFile() == cpAttacker.getSquare().getFile() &&
        Math.abs(cpKing.getSquare().getRank() - cpAttacker.getSquare().getRank()) == 1)
    {
      // They're adjacent, no possibility of interposing.
      return bSafe;
    }
       
    // See if there is an interposable square for the attacking rook:
    for (i = 0; i < cpAttacker.numLegal(); i++)
    {
      sTemp = cpAttacker.getLegalMove(i);
      
      // See if the attack is along a rank or file:
      if (cpAttacker.getSquare().getRank() == cpKing.getSquare().getRank())
      {
        // We're on the same rank. See if the attacker is to my left or right.
        if (sTemp.getRank() == cpAttacker.getSquare().getRank())
        {
          // Attacker legal move square is also on the same rank.
          if (cpAttacker.getSquare().getFile() < cpKing.getSquare().getFile())
          {
            // Attacker is on my left, interposing squares are between us.
            if (sTemp.getFile() < cpKing.getSquare().getFile() &&
                sTemp.getFile() > cpAttacker.getSquare().getFile())
            {
              // The square is between us, see if I can move a piece onto it:
              for (j = 0; j < 8; j++)
              {
                for (k = 0; k < 8; k++)
                {
                  cpTemp = cb.getPiece(j, k);
                  if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                  {
                    for (l = 0; l < cpTemp.numLegal(); l++)
                    {
                      if (sTemp.equals(cpTemp.getLegalMove(l)))
                      {
                        // May interpose only if the interposing piece is not pinned:
                        if (cb.pinnedBy(cpTemp, bW) == 0)
                        {
                          bSafe = true;
                          break;                                       // Break out of l;
                        }
                      }
                    } // End l.
                  }
                  if (bSafe) break;                                  // Break out of k.
                } // End k.
                if (bSafe) break;                                    // Break out of j.
              } // End j.
            }
          }
          else
          {
            // Attacker is on my right.
            if (sTemp.getFile() > cpKing.getSquare().getFile() &&
                sTemp.getFile() < cpAttacker.getSquare().getFile())
            {
              // The square is between us, see if I can move a piece onto it:
              for (j = 0; j < 8; j++)
              {
                for (k = 0; k < 8; k++)
                {
                  cpTemp = cb.getPiece(j, k);
                  if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                  {
                    for (l = 0; l < cpTemp.numLegal(); l++)
                    {
                      if (sTemp.equals(cpTemp.getLegalMove(l)))
                      {
                        // May interpose only if the interposing piece is not pinned:
                        if (cb.pinnedBy(cpTemp, bW) == 0)
                        {
                          bSafe = true;
                          break;                                       // Break out of l;
                        }
                      }
                    } // End l.
                  }
                  if (bSafe) break;                                   // Break out of k.
                } // End k.
                if (bSafe) break;                                     // Break out of j.
              } // End j.
            }
          }
        }
      }
      else
      { 
        if (sTemp.getFile() == cpAttacker.getSquare().getFile())
        {
          // Attacker legal move square is also on the same file.
          if (cpAttacker.getSquare().getFile() == cpKing.getSquare().getFile())
          {
            // We're on the same file. See if the attacker is above or below me.
            if (cpAttacker.getSquare().getRank() < cpKing.getSquare().getRank())
            {
              // Attacker is above me, interposing squares between us.
              if (sTemp.getRank() < cpKing.getSquare().getRank() &&
                  sTemp.getRank() > cpAttacker.getSquare().getRank())
              {
                // The square is between us, see if I can move a piece onto it:
                for (j = 0; j < 8; j++)
                {
                  for (k = 0; k < 8; k++)
                  {
                    cpTemp = cb.getPiece(j, k);
                    if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                    {
                      for (l = 0; l < cpTemp.numLegal(); l++)
                      {
                        if (sTemp.equals(cpTemp.getLegalMove(l)))
                        {
                          // May interpose only if the interposing piece is not pinned:
                          if (cb.pinnedBy(cpTemp, bW) == 0)
                          {
                            bSafe = true;
                            break;                                       // Break out of l;
                          }
                        }
                      } // End l.
                    }
                    if (bSafe) break;                                  // Break out of k.
                  } // End k.
                  if (bSafe) break;                                    // Break out of j.
                } // End j.
              }
            }
            else
            {
              // Attacker is below me.
              if (sTemp.getRank() > cpKing.getSquare().getRank() &&
                  sTemp.getRank() < cpAttacker.getSquare().getRank())
              {
                // The square is between us, see if I can move a piece onto it:
                for (j = 0; j < 8; j++)
                {
                  for (k = 0; k < 8; k++)
                  {
                    cpTemp = cb.getPiece(j, k);
                    if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                    {
                      for (l = 0; l < cpTemp.numLegal(); l++)
                      {
                        if (sTemp.equals(cpTemp.getLegalMove(l)))
                        {
                          // May interpose only if the interposing piece is not pinned:
                          if (cb.pinnedBy(cpTemp, bW) == 0)
                          {
                            bSafe = true;
                            break;                                       // Break out of l;
                          }
                        }
                      } // End l.
                    }
                    if (bSafe) break;                                    // Break out of k.
                  } // End k.
                  if (bSafe) break;                                      // Break out of j.
                } // End j.
              }
            }
          }
        }
      }
      if (bSafe) break;                                                  // Bug fix, June 3, 2002.
    } // End i.
    return bSafe;
    
  } // End of rookInterposeP().

  private boolean bishopInterposeP(ChessBoard cb, boolean bW, ChessPiece cpKing, ChessPiece cpAttacker)
  {
    boolean bSafe = false;
    int i = 0;
    int j = 0;
    int k = 0;
    int l = 0;
    Square sTemp = null;
    ChessPiece cpTemp = null;

    // First see if the bishop is adjacent to the king:
    // They are adjacent if delta rank == 1 and delta file == 1:
    if (Math.abs(cpKing.getSquare().getRank() - cpAttacker.getSquare().getRank()) == 1 &&
        Math.abs(cpKing.getSquare().getFile() - cpAttacker.getSquare().getFile()) == 1)
    {
      // They're adjacent, no possibility of interposing.
      return bSafe;
    }

    // See if there is an interposable square for the attacking rook:
    for (i = 0; i < cpAttacker.numLegal(); i++)
    {
      sTemp = cpAttacker.getLegalMove(i);

      // See if the attack is along a positive or negative slope diagonal:
      if (cpAttacker.getSquare().getRank() - cpKing.getSquare().getRank() ==
          cpAttacker.getSquare().getFile() - cpKing.getSquare().getFile())
      {
        if (sTemp.getRank() - cpAttacker.getSquare().getRank() ==
            sTemp.getFile() - cpAttacker.getSquare().getFile())
        {
          // Attacker legal move square is also on the same diagonal.
          // We're on the negative slope. See if the attacker is to my left or right.
          if (cpAttacker.getSquare().getFile() < cpKing.getSquare().getFile())
          {
            // Attacker is on my left, interposing squares are between us.
            if (sTemp.getFile() < cpKing.getSquare().getFile() &&
                sTemp.getFile() > cpAttacker.getSquare().getFile())
            {
              // The square is between us, see if I can move a piece onto it:
              for (j = 0; j < 8; j++)
              {
                for (k = 0; k < 8; k++)
                {
                  cpTemp = cb.getPiece(j, k);
                  if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                  {
                    for (l = 0; l < cpTemp.numLegal(); l++)
                    {
                      if (sTemp.equals(cpTemp.getLegalMove(l)))
                      {
                        // May interpose only if the interposing piece is not pinned:
                        if (cb.pinnedBy(cpTemp, bW) == 0)
                        {
                          bSafe = true;
                          break;                                       // Break out of l;
                        }
                      }
                    } // End l.
                  }
                  if (bSafe) break;                                  // Break out of k.
                } // End k.
                if (bSafe) break;                                    // Break out of j.
              } // End j.
            }
          }
          else
          {
            // Attacker is on my right, interposing squares are between us.
            if (sTemp.getFile() > cpKing.getSquare().getFile() &&
                sTemp.getFile() < cpAttacker.getSquare().getFile())
            {
              // The square is between us, see if I can move a piece onto it:
              for (j = 0; j < 8; j++)
              {
                for (k = 0; k < 8; k++)
                {
                  cpTemp = cb.getPiece(j, k);
                  if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                  {
                    for (l = 0; l < cpTemp.numLegal(); l++)
                    {
                      if (sTemp.equals(cpTemp.getLegalMove(l)))
                      {
                        // May interpose only if the interposing piece is not pinned:
                        if (cb.pinnedBy(cpTemp, bW) == 0)
                        {
                          bSafe = true;
                          break;                                       // Break out of l;
                        }
                      }
                    } // End l.
                  }
                  if (bSafe) break;                                  // Break out of k.
                } // End k.
                if (bSafe) break;                                    // Break out of j.
              } // End j.
            }
          }
        }
      }
      else
      {
        if (cpAttacker.getSquare().getRank() - cpKing.getSquare().getRank() ==
            cpKing.getSquare().getFile() - cpAttacker.getSquare().getFile())
        {
          // We're on the positive slope. See if the attacker is to my left or right.
          if (sTemp.getRank() - cpAttacker.getSquare().getRank() ==
              cpAttacker.getSquare().getFile() - sTemp.getFile())
          {
            // Attacker legal move square is also on the same diagonal.
            if (cpAttacker.getSquare().getFile() < cpKing.getSquare().getFile())
            {
              // Attacker is on my left, interposing squares are between us.
              if (sTemp.getFile() < cpKing.getSquare().getFile() &&
                  sTemp.getFile() > cpAttacker.getSquare().getFile())
              {
                // The square is between us, see if I can move a piece onto it:
                for (j = 0; j < 8; j++)
                {
                  for (k = 0; k < 8; k++)
                  {
                    cpTemp = cb.getPiece(j, k);
                    if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                    {
                      for (l = 0; l < cpTemp.numLegal(); l++)
                      {
                        if (sTemp.equals(cpTemp.getLegalMove(l)))
                        {
                          // May interpose only if the interposing piece is not pinned:
                          if (cb.pinnedBy(cpTemp, bW) == 0)
                          {
                            bSafe = true;
                            break;                                       // Break out of l;
                          }
                        }
                      } // End l.
                    }
                    if (bSafe) break;                                  // Break out of k.
                  } // End k.
                  if (bSafe) break;                                    // Break out of j.
                } // End j.
              }
            }
            else
            {
              // Attacker is on my right, interposing squares are between us.
              if (sTemp.getFile() > cpKing.getSquare().getFile() &&
                  sTemp.getFile() < cpAttacker.getSquare().getFile())
              {
                // The square is between us, see if I can move a piece onto it:
                for (j = 0; j < 8; j++)
                {
                  for (k = 0; k < 8; k++)
                  {
                    cpTemp = cb.getPiece(j, k);
                    if (cpTemp != null && cpTemp.getType() != 'K' && cpTemp.whiteP() == bW)
                    {
                      for (l = 0; l < cpTemp.numLegal(); l++)
                      {
                        if (sTemp.equals(cpTemp.getLegalMove(l)))
                        {
                          // May interpose only if the interposing piece is not pinned:
                          if (cb.pinnedBy(cpTemp, bW) == 0)
                          {
                            bSafe = true;
                            break;                                       // Break out of l;
                          }
                        }
                      } // End l.
                    }
                    if (bSafe) break;                                    // Break out of k.
                  } // End k.
                  if (bSafe) break;                                      // Break out of j.
                } // End j.
              }
            }
          }
        }
        if (bSafe) break;                                                // Bug fix, June 3, 2002.
      } // End of i;
    }
    return bSafe;
  } // End of bishopInterposeP().

  // See if there is a stalemate (predicate naming convention, a Lisp standard):
  public boolean stalemateP(boolean bW)                                  // bW is true if it's white's turn.
  {
    boolean bR = false;                                                  // Return value.
    int i = 0;
    int j = 0;
    int iNumLegal = 0;
    ChessPiece cpTemp0 = null;

    // It's a stalemate if the one to move's number of legal moves is zero.
    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        cpTemp0 = getPiece(i, j);
        if (cpTemp0 != null)
        {
          if (cpTemp0.whiteP())
          {
            // The piece is white.
            if (bW)
            {
              // It's white's move.
              iNumLegal += cpTemp0.numLegal();
            }
          }
          else
          {
            // The piece is black.
            if (!bW)
            {
              // It's black's move.
              iNumLegal += cpTemp0.numLegal();
            }
          }
        }
      }
    }
    if (iNumLegal == 0) bR = true;
    return bR;

  } // End of stalemateP().

  public String toString()
  {
    String sR = null;
    StringBuffer sbTemp = null;
    int i = 0;
    int j = 0;
    char cTemp = 0;

    sbTemp = new StringBuffer();

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        if (p[i][j] == null)
        {
          cTemp = '_';                                           // Empty square, underscore character.
        }
        else
        {
          if (p[i][j].whiteP())
          {
            if (!p[i][j].movedP())
            {
              cTemp = p[i][j].getType();                         // Not moved is the default.
            }
            else
            {
              cTemp = (char) (p[i][j].getType() + 8);            // Encoding for not-moved penalty pieces.
            }
          }
          else
          {
            if (!p[i][j].movedP())
            {
              cTemp = Character.toLowerCase(p[i][j].getType());  // Not moved is the default.
            }
            else
            {
              cTemp = (char) (Character.toLowerCase(p[i][j].getType()) + 8);  // Encoding for not-moved penalty pieces.
            }
          }
        }
        sbTemp.append(cTemp);
      }
    }
    if (bMove)
    {
      sbTemp.append('1');
    }
    else
    {
      sbTemp.append('0');
    }
    sR = new String(sbTemp);
    return sR;
  }

} // End of class ChessBoard.

final class GameNode                                                   // A node in the game tree.
{
  private float sfMetric = 0;                                    // Each node has a goodness value.
  private Vector gnChild = null;                                 // Vector of Game tree children array.
  private ChessBoard cb = null;                                  // Holds array of chess pieces.
  private ChessBoard cbBest = null;                              // The best move represented by a ChessBoard.

  GameNode()                                                     // Constructor.
  {
    gnChild = new Vector();                                      // Must set boards and data externally.
  }

  public ChessBoard getBoard()                                   // Accessor
  {
    return cb;
  }

  public float getMetric()                                       // Accessor
  {
    return sfMetric;
  }

  public GameNode getChild(int i)
  {
    GameNode r = null;

    if (i < gnChild.size())
    {
      r = (GameNode) gnChild.elementAt(i);
    }
    return r;
  }

  public ChessBoard getBest()
  {
    return cbBest;
  }

  public void setBoard(ChessBoard cb)                            // Mutator
  {
    this.cb = cb;
  }

  public void setBoard(String s)                                 // Mutator
  {
    this.cb = new ChessBoard(s);                                 // Call the string chessboard constructor.
  }

  public void setMetric(float sf)
  {
    sfMetric = sf;
  }

  public void setBest(ChessBoard cb)
  {
    cbBest = cb;
  }

  public void setChild(GameNode gn, int i)
  {
    gnChild.setElementAt(gn, i);
  }

  public void addChild(GameNode gn)
  {
    gnChild.addElement(gn);
  }

  public int numChildren()
  {
    return gnChild.size();
  }

  public void sortChildrenQuick(boolean bAscending)
  {
    int i = 0;
    int iSize = gnChild.size();
    GameNode in[] = new GameNode[iSize];
    GameNode out[] = new GameNode[iSize];

    for (i = 0; i < iSize; i++)
    {
      in[i] = (GameNode) gnChild.elementAt(i);
    }
    quickSort(in, 0, iSize - 1, out, bAscending);
    for (i = 0; i < iSize; i++)
    {
      gnChild.setElementAt(out[i], i);
    }
  }

  /* This function is adapted from Structured C for Engineering and Technology, p. 403. */
  private void quickSort(GameNode in[], int a, int b, GameNode out[], boolean bAscending)
  {
    int iSize = b - a + 1;
    GameNode pivot;                             /* Holder for array value.              */
    int i = 0;                                  /* i, j, k, and z are array indices.    */
    int j = 0;
    int k = 1;                                  /* k starts at one because our pivot is */
    int z = 0;                                  /* always at the zeroeth index in this version. */
    GameNode in1[] = new GameNode[iSize];       /* Arrays for copying.                   */
    GameNode in2[] = new GameNode[iSize];
    GameNode out1[] = new GameNode[iSize];
    GameNode out2[] = new GameNode[iSize];

    if(b != -1)                                 /* Just one element? Do nothing,        */
    {                                           /* ending the recursion.                */
      if(a == b)
      {
        out[0] = in[a];
      }
      else
      {
        /* Only two elements? They may need to be swappped. */
        if(1 == (b - a))
        {
          if (bAscending)
          {
            if(in[a].sfMetric <= in[b].sfMetric)  /* Do not swap. */
            {
              out[0] = in[a];
              out[1] = in[b];
            }
            else                                  /* Swap them. */
            {
              out[0] = in[b];
              out[1] = in[a];
            }
          }
          else
          {
            if(in[a].sfMetric >= in[b].sfMetric)  /* Do not swap. */
            {
              out[0] = in[a];
              out[1] = in[b];
            }
            else                                  /* Swap them. */
            {
              out[0] = in[b];
              out[1] = in[a];
            }
          }
        }
        else                                    /* Handle 3-or-more elements. */
        {
          pivot = in[0];                        /* Pick first element for pivot. */
          while(k <= b)
          {
            /* Choose an output array to write to. */
            if (bAscending)
            {
              if(pivot.sfMetric > in[k].sfMetric)                   /* Compare pivot. */
              {
                /* Write smaller output array. */
                in1[i] = in[k];
                i++;
              }
              else                                /* Pivot is smaller. */
              {
                /* Write larger output array. */
                in2[j] = in[k];
                j++;
              }
            }
            else
            {
              if(pivot.sfMetric < in[k].sfMetric)                   /* Compare pivot. */
              {
                /* Write smaller output array. */
                in1[i] = in[k];
                i++;
              }
              else                                /* Pivot is smaller. */
              {
                /* Write larger output array. */
                in2[j] = in[k];
                j++;
              }
            }
            k++;                                /* Adjust input pointer. */
          }

          /* Sort smaller partition. */
          quickSort(in1, 0, i-1, out1, bAscending);

          /* Sort larger partition. */
          quickSort(in2, 0, j-1, out2, bAscending);

          /* Write smaller array to output. */
          for(k = 0; k < i; k++)
          {
            out[z] = out1[k];
            z++;
          }
          out[z] = pivot;                       /* Write pivot to output. */
          z++;

          /* Write larger array to output. */
          for(k = 0; k < j; k++)
          {
            out[z] = out2[k];
            z++;
          }
        }
      }
    }
  } // End of function quickSort().

} // End of class GameNode.

final class ChessPiece                                   // Class for chess pieces.
{
  // Inheritance is not used for chesspieces. In this model, if inheritance were used, typecasting from
  // an array of pieces (the ChessBoard) would still be required in order to make a copy of the ChessBoard,
  // so type information will still be needed as a piece attribute. 
  // The problem comes when trying to make a copy of a ChessPiece (say a Pawn) to
  // propagate in the chess game tree: the copy is a ChessPiece, not a Pawn.

  // A piece has a color, type, and value.
  private boolean bWhite = false;                        // Flag for whiteness of piece.
  private boolean bMoved = false;                        // Flag for having been moved.
  private char cType = 0;                                // P, N, B, R, Q, or K
  private float sfValue = 0;                             // pawn = 1.0, knight = 3; bishop = 3.5
                                                         // rook = 5; queen = 9; king = 1000
  private Square sOn = null;                             // Square I'm standing on.
  private Vector vLegal = null;                          // Vector of squares to move to legally (legal moves).

  private int EPCaptureFile = -1;                        // File on which EP capture is enabled.
 
  ChessPiece(boolean bW, char cT, int iRank, int iFile)  // Color, type, and position constructor
  {                                                      //  (used in pawn promotion in min-max).
    bWhite = bW;
    bMoved = false;
    cType = cT;
    sOn = new Square(iRank, iFile);
    vLegal = new Vector();

    switch (cType)
    {
      case 'P': sfValue = 1; break;
      case 'N': sfValue = 3; break;
      case 'B': sfValue = (float) 3.5; break;
      case 'R': sfValue = 5; break;
      case 'Q': sfValue = 9; break;
      case 'K': sfValue = 1000; break;
    }

    EPCaptureFile = -1;
  }

  ChessPiece(boolean bW, char cT, Square s)              // Square constructor (used in ChessBoard string constructor)
  {
    bWhite = bW;
    bMoved = false;
    cType = cT;
    sOn = s;
    vLegal = new Vector();

    switch (cType)
    {
      case 'P': sfValue = 1; break;
      case 'N': sfValue = 3; break;
      case 'B': sfValue = (float) 3.5; break;
      case 'R': sfValue = 5; break;
      case 'Q': sfValue = 9; break;
      case 'K': sfValue = 1000; break;
    }

    EPCaptureFile = -1;
  }

  ChessPiece(boolean bW, char cT, Square s, boolean b)    // Moved constructor (used in setupBoard())
  {
    bWhite = bW;
    bMoved = b;
    cType = cT;
    sOn = s;
    vLegal = new Vector();

    switch (cType)
    {
      case 'P': sfValue = 1; break;
      case 'N': sfValue = 3; break;
      case 'B': sfValue = (float) 3.5; break;
      case 'R': sfValue = 5; break;
      case 'Q': sfValue = 9; break;
      case 'K': sfValue = 1000; break;
    }

    EPCaptureFile = -1;
  }

  ChessPiece(ChessPiece cp)                               // Copy constructor (used in the min-max method)
  {
    bWhite = cp.bWhite;
    bMoved = cp.bMoved;
    cType = cp.cType;                                     // Copy as the correct type of piece.
    sfValue = cp.sfValue;
    sOn = new Square(cp.sOn);
    vLegal = (Vector) cp.vLegal.clone();
    EPCaptureFile = cp.EPCaptureFile;
  }

  public boolean whiteP()                                 // Accessor.
  {
    return bWhite;
  }

  public boolean movedP()
  {
    return bMoved;
  }

  public char getType()                                   // Accessor.
  {
    return cType;
  }

  public float getValue()                                 // Accessor.
  {
    return sfValue;
  }

  public Square getSquare()
  {
    return sOn;
  }

  public void setType(char c)                             // Mutator.
  {
    cType = c;
    switch (cType)
    {
      case 'P': sfValue = 1; break;
      case 'N': sfValue = 3; break;
      case 'B': sfValue = (float) 3.5; break;
      case 'R': sfValue = 5; break;
      case 'Q': sfValue = 9; break;
      case 'K': sfValue = 1000; break;
    }
  }

  public void moveTo(Square s)                            // Mutator.
  {
    sOn = s;
    bMoved = true;
  }

  public void moveTo(int iRank, int iFile)
  {
    sOn.setRank(iRank);
    sOn.setFile(iFile);
    bMoved = true;
  }

  public void setMoved(boolean b)
  {
    bMoved = b;
  }

  public void enableEPCapture(int iFile)                  // The pawn may be captured EP on file iFile.
  {                                                       // Default iFile = -1 (off the board).
    EPCaptureFile = iFile;
  }

  public int getEPCaptureFile()
  {
    return EPCaptureFile;
  }

  public void removeLegalMoves()
  {
    vLegal.removeAllElements();
  }

  public void generateLegalMoves(ChessBoard cb)
  {
    vLegal.removeAllElements();                           // Start fresh.

    switch (cType)                                        // The switch on cType is required because Java doesn't
    {                                                     // handle generic copying of ChessPiece subclasses.
      case 'P':                                           // Tried a version of this program with inheritance, NFG.
      {
        pawnGenerate(cb);
        break;
      }
      case 'N':
      {
        knightGenerate(cb);
        break;
      }
      case 'B':
      {
        bishopGenerate(cb);
        break;
      }
      case 'R':
      {
        rookGenerate(cb);
        break;
      }
      case 'Q':
      {
        bishopGenerate(cb);
        rookGenerate(cb);
        break;
      }
      case 'K':
      {
        kingGenerate(cb);
        break;
      }
    }
  }

  public int numLegal()
  {
    return vLegal.size();
  }

  public Square getLegalMove(int i)                  // Legal move "to" square accessor.
  {
    Square r = null;

    if (i < vLegal.size())
    {
      r = ((Square) vLegal.elementAt(i));
    }
    return r;
  }

  public boolean legalP(Square sTo)                  // Is the "to" square a member of the set of legal moves?
  {
    boolean r = false;                               // Return value.
    int i = 0;

    for (i = 0; i < vLegal.size(); i ++)
    {
      if (sTo.equals((Square) vLegal.elementAt(i)))  // Cast Object to Square.
      {
        r = true;
        break;                                       // Break out of for.
      }
    }
    return r;
  }

  public boolean EPCP(Square sTo)                    // Is the "to" square EPC enabled?
  {
    boolean r = false;                               // Return value.
    int i = 0;

    for (i = 0; i < vLegal.size(); i ++)
    {
      if (sTo.equals((Square) vLegal.elementAt(i)))  // Cast Object to Square.
      {
        r = ((Square) vLegal.elementAt(i)).EPCP();
        break;                                       // Break out of for.
      }
    }
    return r;
  }

  private void pawnGenerate(ChessBoard cb)                                         // Generate legal moves for a pawn.
  {
    Square sTemp = null;
    int iRank = 0;
    int iFile = 0;

    iRank = sOn.getRank();
    iFile = sOn.getFile();

    if (bWhite)
    {
      if (iRank > 0)                                                               // Not at the end of the line for a pawn.
      {
        // Non-capturing moves:
        if (cb.getPiece(iRank - 1, iFile) == null)                                 // Pawn is not blocked.
        {
          sTemp = new Square(iRank - 1, iFile);
          if (iRank - 1 == 0) sTemp.setPawnPromote(true);
          vLegal.addElement(sTemp);                                                // Pawn moves forward one square.
        }
        if (iRank == 6)
        {
          if (cb.getPiece(iRank - 2, iFile) == null)                               // Pawn is not blocked two up.
          {
            if (cb.getPiece(iRank - 1, iFile) == null)                             // Pawn is not blocked one up.
            {
              vLegal.addElement(new Square(iRank - 2, iFile));                     // Pawn moves forward two squares.
            }
          }
        }
        
        // Capturing moves:
        if (iFile > 0)                                                             // We're not on the far left file.
        {
          if (cb.getPiece(iRank - 1, iFile - 1) != null)                           // Some piece is there.
          {
            if (!cb.getPiece(iRank - 1, iFile - 1).whiteP())                       // Left diagonal piece is not white.
            {
              sTemp = new Square(iRank - 1, iFile - 1);
              sTemp.setCapture(true);
              if (iRank - 1 == 0) sTemp.setPawnPromote(true);
              vLegal.addElement(sTemp);                                            // Pawn captures to the left diagonal.
            }
          }
        }
        if (iFile < 7)                                                             // We're not on the far right file.
        {
          if (cb.getPiece(iRank - 1, iFile + 1) != null)                           // Some piece is there.
          {
            if (!cb.getPiece(iRank - 1, iFile + 1).whiteP())                       // Right diagonal piece is not white.
            {
              sTemp = new Square(iRank - 1, iFile + 1);
              sTemp.setCapture(true);
              if (iRank - 1 == 0) sTemp.setPawnPromote(true);
              vLegal.addElement(sTemp);                                            // Pawn captures to the right diagonal.
            }
          }
        }
        // En passant capture:
        if (EPCaptureFile > -1)
        {
          // The white pawn, if enabled, will be on the 3rd rank:
          sTemp = new Square(iRank - 1, EPCaptureFile);
          sTemp.setEPC(true);
          sTemp.setCapture(true);
          vLegal.addElement(sTemp);                                                // Pawn captures 2nd rank.
        }
      }
    }
    else // Pawn is black
    {
      if (iRank < 7)                                                               // Not at the end of the line for a pawn.
      {
        // Non-capturing moves:
        if (cb.getPiece(iRank + 1, iFile) == null)                                 // Pawn is not blocked.
        {
          sTemp = new Square(iRank + 1, iFile);
          if (iRank + 1 == 7) sTemp.setPawnPromote(true);
          vLegal.addElement(sTemp);                                                // Pawn moves forward one square.
        }
        if (iRank == 1)                                                            // Sitting on original black pawn rank.
        {
          if (cb.getPiece(iRank + 1, iFile) == null)                               // Pawn is not blocked one up.
          {
            if (cb.getPiece(iRank + 2, iFile) == null)                             // Pawn is not blocked two up.
            {
              vLegal.addElement(new Square(iRank + 2, iFile));                     // Pawn moves forward two squares.
            }
          }
        }
        
        // Capturing moves:
        if(iFile > 0)                                                               // We're not on the far left file.
        {
          if (cb.getPiece(iRank + 1, iFile - 1) != null)                            // Some piece is there.
          {
            if (cb.getPiece(iRank + 1, iFile - 1).whiteP())                         // Left diagonal piece is not black.
            {
              sTemp = new Square(iRank + 1, iFile - 1);
              sTemp.setCapture(true);
              if (iRank + 1 == 7) sTemp.setPawnPromote(true);
              vLegal.addElement(sTemp);                                            // Pawn captures to the left diagonal.
            }
          }
        }
        if(iFile < 7)                                                              // We're not on the far right file.
        {
          if (cb.getPiece(iRank + 1, iFile + 1) != null)                           // Some piece is there.
          {
            if (cb.getPiece(iRank + 1, iFile + 1).whiteP())                        // Right diagonal piece is not black.
            {
              sTemp = new Square(iRank + 1, iFile + 1);
              sTemp.setCapture(true);
              if (iRank + 1 == 7) sTemp.setPawnPromote(true);
              vLegal.addElement(sTemp);                                            // Pawn captures to the right diagonal.
            }
          }
        }
        // En passant capture:
        if (EPCaptureFile > -1)
        {
          // The black pawn, if enabled, will be on the 4th rank:
          sTemp = new Square(iRank + 1, EPCaptureFile);
          sTemp.setCapture(true);
          sTemp.setEPC(true);
          vLegal.addElement(sTemp);                                                // Pawn captures 5th rank.
        }
      }
    }
    EPCaptureFile = -1;                                                            // Disable EP capture.
  } // End of pawnGenerate()

  public void knightGenerate(ChessBoard cb)
  {    
    Square sTemp = null;
    int iRank = 0;
    int iFile = 0;
    int i = 0;
    int j = 0;
    int k = 0;
    int iMoveToRank = 0;
    int iMoveToFile = 0;
    boolean bWhitePiece = false;

    iRank = sOn.getRank();
    iFile = sOn.getFile();

    for (i = -1; i <= 1; i += 2)
    {
      for (j = -1; j <= 1; j += 2)
      {
        for (k = 0; k <= 1; k++)
        {
          if (k == 0)
          {
            iMoveToRank = iRank + 2 * i;
            iMoveToFile = iFile + j;
          }
          else
          {
            iMoveToRank = iRank + i;
            iMoveToFile = iFile + 2 * j;
          }
          if (iMoveToRank >= 0 && iMoveToRank < 8 &&
              iMoveToFile >= 0 && iMoveToFile < 8)                                    // Move is not off the board.
          {
            if (cb.getPiece(iMoveToRank, iMoveToFile) == null)
            {  
              sTemp = new Square(iMoveToRank, iMoveToFile);
              vLegal.addElement(sTemp);
            }
            else
            {
              bWhitePiece = cb.getPiece(iMoveToRank, iMoveToFile).bWhite;
              if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))              // Capturing opposite color.
              {
                sTemp = new Square(iMoveToRank, iMoveToFile);
                sTemp.setCapture(true);
                vLegal.addElement(sTemp);
              }
            }
          }
        }
      }
    }
  } // End of knightGenerate()

  public void bishopGenerate(ChessBoard cb)
  {
    int i = 0;
    Square sTemp = null;
    int iRank = 0;
    int iFile = 0;
    boolean bWhitePiece = false;

    iRank = sOn.getRank();
    iFile = sOn.getFile();

    // Moves to the south-east:
    for (i = 1; i <= 7; i++)
    {
      if (((iRank + i) <= 7) && ((iFile + i) <= 7))                               // Move is not off the board.
      {
        if (cb.getPiece(iRank + i, iFile + i) == null)
        {  
            sTemp = new Square(iRank + i, iFile + i);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank + i, iFile + i).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))               // Capturing opposite color.
          {
            sTemp = new Square(iRank + i, iFile + i);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                     // Break out of for().
      }
    }

    // Moves to the south-west:
    for (i = 1; i <= 7; i++)
    {
      if (((iRank + i) <= 7) && ((iFile - i) >= 0))                                // Move is not off the board.
      {
        if (cb.getPiece(iRank + i, iFile - i) == null)
        {  
            sTemp = new Square(iRank + i, iFile - i);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank + i, iFile - i).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))               // Capturing opposite color.
          {
            sTemp = new Square(iRank + i, iFile - i);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                      // Break out of for().
      }
    }

    // Moves to the north-west:
    for (i = 1; i <= 7; i++)
    {
      if (((iRank - i) >= 0) && ((iFile - i) >= 0))                                 // Move is not off the board.
      {
        if (cb.getPiece(iRank - i, iFile - i) == null)
        {  
            sTemp = new Square(iRank - i, iFile - i);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank - i, iFile - i).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))                // Capturing opposite color.
          {
            sTemp = new Square(iRank - i, iFile - i);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                      // Break out of for().
      }
    }

    // Moves to the north-east:
    for (i = 1; i <= 7; i++)
    {
      if (((iRank - i) >= 0) && ((iFile + i) <= 7))                                 // Move is not off the board.
      {
        if (cb.getPiece(iRank - i, iFile + i) == null)
        {  
            sTemp = new Square(iRank - i, iFile + i);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank - i, iFile + i).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))                 // Capturing opposite color.
          {
            sTemp = new Square(iRank - i, iFile + i);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                       // Break out of for().
      }
    }
  } // End of bishopGenerate

  public void rookGenerate(ChessBoard cb)
  {
    int i = 0;
    Square sTemp = null;
    int iRank = 0;
    int iFile = 0;
    boolean bWhitePiece = false;

    iRank = sOn.getRank();
    iFile = sOn.getFile();

    // Moves to the east:
    for (i = 1; i <= 7; i++)
    {
      if ((iFile + i) <= 7)                                                          // Move is not off the board.
      {
        if (cb.getPiece(iRank, iFile + i) == null)
        {  
            sTemp = new Square(iRank, iFile + i);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank, iFile + i).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))                  // Capturing opposite color.
          {
            sTemp = new Square(iRank, iFile + i);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                       // Break out of for().
      }
    }

    // Moves to the west:
    for (i = 1; i <= 7; i++)
    {
      if ((iFile - i) >= 0)                                                          // Move is not off the board.
      {
        if (cb.getPiece(iRank, iFile - i) == null)
        {  
            sTemp = new Square(iRank, iFile - i);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank, iFile - i).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))                  // Capturing opposite color.
          {
            sTemp = new Square(iRank, iFile - i);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                       // Break out of for().
      }
    }

    // Moves to the south:
    for (i = 1; i <= 7; i++)
    {
      if ((iRank + i) <= 7)                                                          // Move is not off the board.
      {
        if (cb.getPiece(iRank + i, iFile ) == null)
        {  
            sTemp = new Square(iRank + i, iFile);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank + i, iFile).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))                  // Capturing opposite color.
          {
            sTemp = new Square(iRank + i, iFile);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                       // Break out of for().
      }
    }

    // Moves to the north:
    for (i = 1; i <= 7; i++)
    {
      if ((iRank - i) >= 0)                                                          // Move is not off the board.
      {
        if (cb.getPiece(iRank - i, iFile ) == null)
        {  
            sTemp = new Square(iRank - i, iFile);
            vLegal.addElement(sTemp);
        }
        else
        {
          bWhitePiece = cb.getPiece(iRank - i, iFile).bWhite;
          if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))                  // Capturing opposite color.
          {
            sTemp = new Square(iRank - i, iFile);
            sTemp.setCapture(true);
            vLegal.addElement(sTemp);
            break;
          }
          else
          {
            break;
          }
        }
      }
      else
      {
        break;                                                                        // Break out of for().
      }
    }
  } // End of rookGenerate()

  public void kingGenerate(ChessBoard cb)
  {
    int i = 0;
    int j = 0;
    ChessPiece cpTemp = null;
    ChessBoard cbCopy = null;
    Square sTemp = null;
    int iRank = 0;
    int iFile = 0;
    int iMoveToRank = 0;
    int iMoveToFile = 0;
    boolean bInCheck = false;
    boolean bWhitePiece = false;

    iRank = sOn.getRank();
    iFile = sOn.getFile();
    bInCheck = underAttackP(cb, iRank, iFile);
    if (bInCheck)
    {
      cb.setCheck(true);
    }

    for (i = -1; i <= 1; i++)
    {
      for (j = -1; j <= 1; j++)
      {
        if (!(i == 0 && j == 0))
        {
          iMoveToRank = iRank + i;
          iMoveToFile = iFile + j;
          if (iMoveToRank >= 0 && iMoveToRank < 8 &&
              iMoveToFile >= 0 && iMoveToFile < 8 &&
              !underAttackP(cb, iMoveToRank, iMoveToFile))
          {
            if (cb.getPiece(iMoveToRank, iMoveToFile) == null)
            {
              // Rare bug fix (occurs only in a mating attack):
              // If the king is in check, see if it's still in check (it backs away from the checking piece):
              if (bInCheck)
              {
                cbCopy = new ChessBoard(cb);                                       // Copy constructor.
                cbCopy.deletePiece(iRank, iFile);                                  // Delete this king.
                cbCopy.refreshLegal();
                if (!underAttackP(cbCopy, iMoveToRank, iMoveToFile))
                {
                  sTemp = new Square(iMoveToRank, iMoveToFile);
                  vLegal.addElement(sTemp);
                }
              }
              else
              {
                sTemp = new Square(iMoveToRank, iMoveToFile);
                vLegal.addElement(sTemp);
              }
            }
            else
            {
              bWhitePiece = cb.getPiece(iMoveToRank, iMoveToFile).bWhite;
              if ((bWhitePiece && !bWhite) || (!bWhitePiece && bWhite))            // Capturing opposite color.
              {
                try
                {
                  // See if the square is under attack when the piece is removed:
                  cbCopy = new ChessBoard(cb);                                     // Copy constructor.
                  cbCopy.deletePiece(iMoveToRank, iMoveToFile);                    // Delete the potential capture piece.
                  cbCopy.deletePiece(iRank, iFile);                                // Delete this king to avoid recursion.
                  cbCopy.refreshLegal();
                  if (!underAttackP(cbCopy, iMoveToRank, iMoveToFile))
                  {
                    sTemp = new Square(iMoveToRank, iMoveToFile);
                    sTemp.setCapture(true);
                    vLegal.addElement(sTemp);
                  }
                }
                catch (NullPointerException e1)
                {
                  // The copy constructor seems occasionally to trigger the null pointer exception even while
                  // cb is not null.
                  System.out.println("Caught error in kingGenerate() " + e1);
                }
              }
            }
          }
        }
      }
    }

    // See about castling:
    if (!bMoved && !bInCheck)                                                      // Can't castle if the king's been moved
    {                                                                              // or is in check.
      // King side:
      // If the squares to the king side are empty:
      if ((cb.getPiece(iRank, iFile + 1) == null) &&
          (cb.getPiece(iRank, iFile + 2) == null))                                 // Might castle king-side.
      {
        // If the king's rook exists and hasn't moved...
        if ((cb.getPiece(iRank, iFile + 3) != null) &&
            (cb.getPiece(iRank, iFile + 3).whiteP() == bWhite) &&                  // Same color as me.
            (cb.getPiece(iRank, iFile + 3).getType() =='R') &&     
            (!cb.getPiece(iRank, iFile + 3).movedP()))
        {
          // If the castling squares are not under attack...
          if (!underAttackP(cb, iRank, iFile + 1) && !underAttackP(cb, iRank, iFile + 2))
          {
            sTemp = new Square(iRank, iFile + 2);
            vLegal.addElement(sTemp);
          }
        }
      }

      // Queen side:
      // If the squares to the queen side are empty:
      if ((cb.getPiece(iRank, iFile - 1) == null) &&
          (cb.getPiece(iRank, iFile - 2) == null) &&
          (cb.getPiece(iRank, iFile - 3) == null))                                 // Might castle queen-side.
      {
        // If the king's rook exists and hasn't moved...
        if ((cb.getPiece(iRank, iFile - 4) != null) &&
            (cb.getPiece(iRank, iFile - 4).whiteP() == bWhite) &&                  // Same color as me.
            (cb.getPiece(iRank, iFile - 4).getType() =='R') &&     
            (!cb.getPiece(iRank, iFile - 4).movedP()))
        {
          // If the castling squares are not under attack...
          if (!underAttackP(cb, iRank, iFile - 1) && !underAttackP(cb, iRank, iFile - 2))
          {
            sTemp = new Square(iRank, iFile - 2);
            vLegal.addElement(sTemp);
          }
        }
      }
    }
  } // End of kingGenerate()

  public boolean underAttackP(ChessBoard cb)                                        // Is the piece under attack?
  {
    return underAttackP(cb, sOn.getRank(), sOn.getFile());

  } // End underAttackP().


  private boolean underAttackP(ChessBoard cb, int iRank, int iFile)                 // Is the square under attack?
  {
    boolean bR = false;
    int i = 0;
    int j = 0;
    int k = 0;
    ChessPiece cpTemp = null;
    Square sTemp = null;

    for (i = 0; i < 8; i++)
    {
      for (j = 0; j < 8; j++)
      {
        cpTemp = cb.getPiece(i, j);
        if ((cpTemp != null) && (cpTemp.whiteP() == !bWhite))                            // Enemy color.
        {
          if (cpTemp.getType() == 'P')
          {
            if (cpTemp.sOn.getFile() + 1 == iFile || cpTemp.sOn.getFile() - 1 == iFile)  // Pawns only attack diagonally.
            {
              if (cpTemp.whiteP() && !bWhite)
              {
                if (cpTemp.sOn.getRank() == iRank + 1)
                {
                  bR = true;                                                             // The square is under attack.
                }
              }
              else
              {
                if (!cpTemp.whiteP() && bWhite)
                {
                  if (cpTemp.sOn.getRank() == iRank - 1)
                  {
                    bR = true;                                                           // The square is under attack.
                  }
                }
              }
            }
          }
          else
          {
            if (cpTemp.getType() == 'K')
            {
              if (cpTemp.distanceTo(iRank, iFile) == 1)
              {
                bR = true;                                                               // The square is under attack by a king.
              }
            }
            else
            {
              for (k = 0; k < cpTemp.numLegal(); k ++)
              {
                sTemp = cpTemp.getLegalMove(k);
                if (sTemp.getRank() == iRank && sTemp.getFile() == iFile)
                {
                  bR = true;                                                              // The square is under attack.
                  break;                                                                  // Break out of k.
                }
              }
            }
          }
        }
        if (bR) break;                     // Break out of j.
      }
      if (bR) break;                       // Break out of i.
    }
    return bR;

  } // End underAttackP().

  int distanceTo(int iRank, int iFile)
  {
    int iDistance = 0;
    int iRankDistance = 0;
    int iFileDistance = 0;

    iRankDistance = Math.abs(sOn.getRank() - iRank);
    iFileDistance = Math.abs(sOn.getFile() - iFile);

    if (iRankDistance > iFileDistance)
    {
      iDistance = iRankDistance;
    }
    else
    {
      iDistance = iFileDistance;
    }

    return iDistance;
  }

} // End of class ChessPiece.

final class Square
{
  private int iRank = 0;                    // Row (i index to ChessBoard)
  private int iFile = 0;                    // Column (j index to ChessBoard)
  private boolean bEPCapture = false;       // Flag for EP capture.
  private boolean bPawnPromote = false;     
  private boolean bCapture = false;

  Square()                                  // Default constructor;
  {
    iRank = 0;
    iFile = 0;
    bEPCapture = false;
    bPawnPromote = false;
    bCapture = false;
  }

  Square(int r, int f)                      // Constructor;
  {
    iRank = r;
    iFile = f;
    bEPCapture = false;
    bPawnPromote = false;
    bCapture = false;
  }

  Square(Square s)                          // Copy constructor;
  {
    if (s == null) s = new Square();
    iRank = s.iRank;
    iFile = s.iFile;
    bEPCapture = s.bEPCapture;
    bPawnPromote = s.bPawnPromote;
    bCapture = s.bCapture;
  }

  public int getRank()                      // Accessors
  {
    return iRank;
  }

  public int getFile()
  {
    return iFile;
  }

  public void setRank(int r)                // Mutators
  {
    iRank = r;
  }

  public void setFile(int f)
  {
    iFile = f;
  }

  public boolean equals(Square s)           // Predicate
  {
    boolean bR = false;

    if (s != null)
    {
      bR = (s.iRank == iRank && s.iFile == iFile);
    }
    return bR;
  }

  public void setEPC(boolean b)
  {
    bEPCapture = b;
  }

  public boolean EPCP()
  {
    return bEPCapture;
  }

  public void setPawnPromote(boolean b)
  {
    bPawnPromote = b;
  }

  public boolean pawnPromoteP()
  {
    return bPawnPromote;
  }

  public void setCapture(boolean b)
  {
    bCapture = b;
  }

  public boolean captureP()
  {
    return bCapture;
  }
} // End of class Square.

final class Timer implements Runnable
{
  // Calls the onTimer() function in the calling applet every iInterval milliseconds.
  // Copyright 1998-2005 by Rick Wagner, all rights reserved. No use without attribution.

  private Thread tTimer;
  private int iInterval;
  private Chess app;                                // The calling applet.
  private long lTime;

  public Timer(Chess a, int i)                      // Interval constructor.
  {
    iInterval = i;
    app = a;
    lTime = 0;
  }

  public Timer(Chess a, int i, long l)              // Time constructor.
  {
    iInterval = i;
    app = a;
    lTime = l;
  }

  public void start()
  {
    tTimer = new Thread(this);
    tTimer.start();
  }

  public void stop()
  {
    if (tTimer != null) tTimer.stop();
    tTimer = null;
  }

  public void run()
  {
    long lStartTime = System.currentTimeMillis();
    long lTemp = 0;
    int iSleepTime = iInterval / 10;
    if (iSleepTime == 0) iSleepTime = 1;
    while (true)
    {
      try
      {
        tTimer.sleep(iSleepTime);                    // Wake up every 1/10 interval and see.
      }                                              // if it's time yet.
      catch(InterruptedException e1)
      {
      }
      lTemp = System.currentTimeMillis();
      if (lTemp - lStartTime >= iInterval)
      {
        app.onTimer(lTime);                          // Trigger the desired events.
        lStartTime = lTemp;
      }
    }
  }
} // End of Timer class


// End of chess applet code.
//
// Version history:
//
// 1.00  February 2, 2001: Development started, Chess.java file created, the board was drawn
//       with character piece representations. Only pawn moves were under development in this
//       first version.
//
// 1.38  February 26, 2001: All basic features operational for complete and correct chess
//       play. The chess applet was Web published for testing on this day.
//
// 1.50  March 15, 2001: Sound feature added to the applet.
//
// 1.60  March 23, 2001: This version adds the game list text box feature.
//
// 1.61  March 27, 2001: This version adds the feature to disable the long playing sounds.
//
// 1.62  March 30, 2001: Adds the rank and file digits and letters to the chessboard.
//
// 1.65  May 23, 2001: Changes the elapsed time limits to 30 < et < 600. Sets a minimum
//       recursion depth to 3 half-moves, preventing a counter-check tactic. Also fixes
//       a bug in interposing a piece in detection of checkmate by a queen.
//
// 1.67  February 14, 2002: Adds the move being considered to the message label.
//
// 1.69  February 15, 2002: Allows the setting of the thinking time boundaries in the applet
//       parameters.
//
// 1.70  February 17, 2002: Adds the number of the move being considered to the message label.
//
// 1.72  February 20, 2002: Implements testing for a draw by stalemate.
//
// 1.75  February 25, 2002: Implements testing for a draw by repetition and testing for a draw
//       by lack of progress (no irreversible moves for 50 moves). Fixes a bug in checking for
//       off-board moves in the generation of king moves.
//
// 1.76  February 26, 2002: Fixes a bug in the detection of a draw by repetition.
//
// 1.78  February 27, 2002: Fixes a bug in en passant capture by white.
//
// 1.80  February 27, 2002: Fixes a bug in the detection of a draw by stalemate.
//
// 1.82  March 1, 2002: Fixes a bug in the detection of checkmate and also fixes a bug
//       that allowed a counter-checking tactic.
//
// 1.84  March 1, 2002: Simplifies the function checkmateP().
//
// 1.85  March 4, 2002: Improves user messages in draw situations.
//
// 1.87  March 5, 2002: Fixes a bug added in versin 1.84 in the simplification of the checkmate
//       detection function checkmateP().
//
// 1.90  March 6, 2001: Fixes a bug in the detection of checkmate that would erroneously allow
//       the capture or interposition of a pinned piece.
//
// 1.93  March 7, 2002: Fixes a bug, found by Harald Schifferl, in the determination of the legal
//       moves of a king. Improves search efficiency slightly by not considering the moves of
//       knights pinned to kings.
//
// 2.00  March 8, 2002: Improves performance slightly by fixing a minor bug in the search depth
//       computation and by removing some error trapping code. Removes some unused functions from
//       the source code. Adjusts evaluation parameters slightly for stronger play in the opening.
//
// 2.08  March 12, 2002: Fixes a minor bug in refreshing legal moves. Improves performance slightly
//       by not considering moves of bishops pinned to kings by rook mode nor moves of rooks pinned
//       to kings by bishop mode. Disambiguates notation in the game listing when more than one of
//       a piece type might make the move.
//
// 2.11  March 14, 2002: Fixes three minor bugs in generating the kings' moves: it would occasionally
//       allow a king to check the other king, it would allow the king to capture a supported
//       piece, and it inappropriately allowed kings' moves into check by pawns. 4874 lines of code.
//
// 2.14  March 18, 2002: Reduces code size by loop rolling the generation of knight and king moves,
//       making the source easier to maintain. 4463 lines of code.
//
// 2.24  March 27, 2002: Adds the save and load game features. 4751 lines of code. Adds the two
//       files ChessDataClient2.java and ChessDataServer.java.
//
// 2.29  March 29, 2002: Allows the user to force the computer to make it's move with the control-M
//       key combination. 4848 lines of code.
//
// 2.42  April 10, 2002: Adds the game file list feature for opening games. 4803 lines of code.
//
// 2.50  May 22, 2002: Fixes a rare bug in which the black king while in check under a mating attack
//       might retreat to the square in line with the checking piece. 4877 lines of code.
//
// 2.50  June 3, 2002: Fixes bugs a in rookInterpose() and bishopInterpose(). 4962 lines of code.
//
// 2.56  June 11, 2002: Simplifies the computeMove() function and makes quiescence search depth a
//       function of the general search depth for stronger play. 4962 lines of code.
//
// 2.61  June 14, 2002: Fixes a bug in setPiece() in which extra EP capture indications were generated.
//       Fixing the bug results in slightly  increased efficiency.
//
// 2.63  June 17, 2002: Implements computer resignation: the computer will resign if it sees a forced
//       mate coming. 4900 lines of code.
//
// 2.69  June 29, 2002: Correctness first, then performance. Implements alpha-beta pruning
//       for deeper thinking. 5083 lines of code, including comments.
//
// 2.82  July 10, 2002: Implements the auto-play feature, toggled on and off by control-A. Now the
//       user can watch the computer play against itself. 5467 lines of code, including comments.
//
// 2.95  July 22, 2002: Implements chess server memory. In this preliminary version, the entire memory
//        database is downloaded for each applet instance. Computed moves are sent to the server for
//        retention. 5467 lines of code, including comments.
//
// 3.00  July 24, 2002: Implements chess server memory. In this version, the applet requests a position
//       query and waits for the server callback. If the position is not found for the current or higher
//       search depth, or if the server doesn't reply in five seconds, the applet computes the move and
//       sends the result to the server for storage. 6000 lines of code, including comments.
//
// 3.12  August 8, 2002: Implements the control-R key combination for board rotation.
//       6196 lines of code, including comments.
//
// 3.18  August 6, 2002: Implements the control-I and control-D key combinations for board search
//       depth incrementing and decrementing. 6248 lines of code, including comments.
//
// 3.22  August 23, 2002: Implements performance boost computation based on hardware perforamance.
//       The base six log of one over the test time duration is used to increment the default recursion
//       limit. 6283 lines of code, including comments.
//
// 3.56  October 7, 2002: Adds the best move found so far to the thinking message.
//
// 4.00  October 31, 2002: Adds the chess piece color contrast feature.
//
// 4.07  November 6, 2002: Compiled with JDK v. 1.3.1. Compiled using the -O (optimize) switch for 10%
//       move speed in move computation. Refines the alebraic notation function to fully accord with
//       convention.
//
// 4.17  February 18, 2003: Fixes a bug in disabling server memory use in duplicate positions.
//
// 4.21  June 10, 2003: Fixes a bug in dealing with pinned pieces.
//
// 4.22  December 16, 2003: Fixes another bug in dealing with pinned pieces.
//
// 4.26  January 29, 2004: Changes the client default socket port to 443 and adds code for setting
//       the port with applet parameters.
//
// 4.27  March 22, 2004: Adds recursion limit incrementing in duplicate positions to try to avoid
//       draws by repetition when the opponent is losing. Changes the default port back to 1997.
//
// 5.04  June 11, 2004: Changes chess memory and data service to MySQL server for hosting at
//       chess.captain.at. Major upgrade by Hannes "Captain" Mayer, compiled for publication
//       by Rick Wagner with JDK on Windows NT, in order to back off from the "bleeding edge."
//
//  6.03  November 24, 2012: Fixes a bug in the width of the message label. Now uses the Label function
//        setMinimumSize() so that occasionally message text during computation isn't clipped. 7197 lines
//        of code in the main file, chess.java.
//
//  6.05  May 18, 2016: Fixes a bug in which a checking pawn could not be captured en passant.
//
//  6.06  May 18, 2016: Adds requestFocus() to the mousemove event to allow reception of keystroke events.
//
// End of file Chess.java, line 7197.